// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.casetool;

import java.io.File;
import java.util.Enumeration;
import java.util.Vector;


import de.uka.ilkd.key.logic.op.Modality;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.ocl.tudresden.TUDresdenParserAdapter;

import de.uka.ilkd.key.ocl.ocldltranslator.detailedocl.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.gui.ProofSettings;
import de.uka.ilkd.key.proof.init.*;
import de.uka.ilkd.key.proof.mgt.PrePostPair;
import de.uka.ilkd.key.proof.mgt.ContractUtil;

public class FunctionalityOnModel {

    private FunctionalityOnModel() {}


    /**
     * parse the invariant of a class if available.
     * @param aReprModelClass the class whose invariant should be parsed
     * @return error message to the user
     */
    public static String parseOneInvariant(ReprModelClass aReprModelClass) {
	String invText = aReprModelClass.getMyInv();
	String returnvalue;
	if (invText==null || "".equals(invText)){
	    returnvalue = "No invariant available";
	}else {
	    returnvalue = 
		parseInvariantCore(aReprModelClass, 
				   invText).getErrorMessageToUser();
	}
	return returnvalue;
    }

    /**
     * parse the throughout property of a class if available.
     * @param aReprModelClass the class whose throughout should be parsed
     * @return error message to the user
     */
    public static String parseOneThroughout(ReprModelClass aReprModelClass) {
	String toutText = aReprModelClass.getMyThroughout();
	String returnvalue;
	if (toutText==null || "".equals(toutText)){
	    returnvalue = "No throughout property available";
	}else {
	    returnvalue = 
		parseInvariantCore(aReprModelClass, 
				   toutText).getErrorMessageToUser();
	}
	return returnvalue;
    }

    /**
     *  calls the parsing routine of the OCL parser
     *  @return TUDresdenParserAdapter contains the result of the parsing
     *          process.
     **/

    private static  TUDresdenParserAdapter parseInvariantCore
	(ReprModelClass aReprModelClass, String invText){
		String className = aReprModelClass.getClassName();
		String xmifile = aReprModelClass.getActXmifile();
	    TUDresdenParserAdapter aAdapter = new TUDresdenParserAdapter();
	    aAdapter.doParse1(className + " inv:",  
			      OCLProofOblInput.getModelManager(),
			      xmifile, "context "
			      +className + " inv: " + invText,
			      aReprModelClass.getContainingPackage());
	    return aAdapter;
    }



    /**
     * parse the specification of a method if available.
     * @param aMethRepr the method whose pre/post specification should be parsed
     * @return error message to the user
     */
    public static String parseOneMethSpec(ReprModelMethod aMethRepr) {
	String preText = aMethRepr.getMyPreCond();
	String postText = aMethRepr.getMyPostCond();
	String returnvalue;
	if ((preText==null || preText.equals("")) && 
	    (postText==null || postText.equals(""))) {
	    returnvalue = "No specification of method available";
	}else {
	    String xmifile = aMethRepr.getActXmifile();
	    String contextString = aMethRepr.getContainingClassName() +"::"+ 
		aMethRepr.getCallSignature();

	    TUDresdenParserAdapter aAdapter1 = new TUDresdenParserAdapter();
	    TUDresdenParserAdapter aAdapter2 = new TUDresdenParserAdapter();

	    aAdapter1.doParse1(contextString + " pre:",  
			      OCLProofOblInput.getModelManager(), 
			       xmifile, "context "
			       + contextString + " pre: " + normalizeConstraint(preText),
			       aMethRepr.getContainingPackage());

	    returnvalue = aAdapter1.getErrorMessageToUser();
	    aAdapter2.doParse1(contextString + " post:",  
			      OCLProofOblInput.getModelManager(), 
			       xmifile, "context "
			       + contextString + " post: " 
			       + normalizeConstraint(postText),
			       aMethRepr.getContainingPackage());
	    String msgSep = "";
	    if (!returnvalue.equals("")){msgSep="\n\n";}
	    returnvalue=returnvalue+msgSep+aAdapter2.getErrorMessageToUser();
	}
	return returnvalue;
    }

    /**
    * Purpose: parses both the pre- and postcondition
    *          The error-messages do not distinguish between errors occurred
    *          in pre- and those occurred in post-condition.
    *          Should rather be used to test whether there occurred an error.
    * @return TUDresdenParserAdapter contains all error messages
    * */

    private static TUDresdenParserAdapter parseMethSpecCore
	(ReprModelMethod aMethRepr, String preText, 
	 String postText, String xmifile) {
	    String contextString = aMethRepr.getContainingClassName() 
		+"::"+ aMethRepr.getCallSignature();
	    TUDresdenParserAdapter aAdapter = new TUDresdenParserAdapter();
	    String parseString = "context "+ 
		contextString + " pre: " + normalizeConstraint(preText) + 
		" post: " + normalizeConstraint(postText);
	    aAdapter.doParse1(contextString+" prepost",  
			      OCLProofOblInput.getModelManager(), 
			      xmifile, 
			      parseString, aMethRepr.getContainingPackage());
	    return aAdapter;
    }
    

    private static String startProver(final ProofOblInput po) {
	return ProblemInitializer.startProver(po);
    }


    /**
     * Check whether the implementation of the method 
     * meets the specification of the post-condition.
     * @param aMethRepr the method whose implementation should be checked.
     * @param totality the Modality to use 
     * @return error message to the user
     */
    public static String methEnsuresPostCondition(ReprModelMethod aMethRepr,
						  Modality totality) {

	String preText = aMethRepr.getMyPreCond();
	String postText = aMethRepr.getMyPostCond();
	String invText = aMethRepr.getContainingClass().getMyInv();

	if (normalizeConstraint(postText).equals("true")){
	    return "no or trivial postcondition available";
	}

	// test iff method is constructor
	String poPreText = "";
	if (aMethRepr.getContainingClassName().equals(aMethRepr.getName())){
	    // preText and invText are ignored for po-generation
	    poPreText="";
	} else {
	    poPreText= "("+ normalizeConstraint(preText) + ")" + " and " 
		+  "(" + normalizeConstraint(invText) + ")";
	}
	OCLMod[] mods=ProofSettings.DEFAULT_SETTINGS
	    .getOCLRefinementSettings().getOCLMods();
	String modSource=ProofSettings.DEFAULT_SETTINGS
	    .getModelSourceSettings().getModelSource();	
	String xmifile=prepareXMI(aMethRepr.getActXmifile(), mods, modSource);
	poPreText=prepareConstraint(poPreText, mods, modSource);
	postText=prepareConstraint(postText, mods, modSource);       		
	// test for syntactical correctness
	if (parseMethSpecCore(aMethRepr, poPreText, 
			      normalizeConstraint(postText),
			      xmifile).getErrorOccurred()){
	    return "Invariant or Pre-/Post-Condition syntactically incorrect";
	}

	PrePostPair ppp = ContractUtil.getPreInvPost(aMethRepr);
	OCLProofOblInput po = 
            new OCLPostPO("EnsuresPostCondition"+totality, aMethRepr, 
			  ppp.getPre(), ppp.getPost(), totality, xmifile);
	return startProver(po);
    }

    /**
     * Check whether the implementation of the method 
     * meets the specification of the post-condition and the invariant.
     * @param aMethRepr the method whose implementation should be checked.
     * @param totality the Modality to use 
     * @return error message to the user
     */
    public static String methEnsuresInvAndPostCondition
	(ReprModelMethod aMethRepr, Modality totality) {
	String preText = aMethRepr.getMyPreCond();
	String postText = aMethRepr.getMyPostCond();
	String invText = aMethRepr.getContainingClass().getMyInv();

	if (normalizeConstraint(postText).equals("true") 
	    && normalizeConstraint(invText).equals("true")){
	    return "no or trivial invariant/postcondition available";
	}
	
	// test iff method is constructor
	String poPreText = "";
	if (aMethRepr.getContainingClassName().equals(aMethRepr.getName())){
	    // preText and invText are ignored for po-generation
	    poPreText="";
	} else {
	    poPreText= "(" + normalizeConstraint(preText) + ")" + " and " 
		+  "(" + normalizeConstraint(invText) + ")";
	}

	OCLMod[] mods=ProofSettings.DEFAULT_SETTINGS.getOCLRefinementSettings()
	                       .getOCLMods();
	String modSource=ProofSettings.DEFAULT_SETTINGS
	    .getModelSourceSettings().getModelSource();	
	String xmifile=prepareXMI(aMethRepr.getActXmifile(), mods, modSource);
	poPreText=prepareConstraint(poPreText, mods, modSource);
	postText=prepareConstraint(postText, mods, modSource);
	
	if (parseMethSpecCore(aMethRepr, poPreText, 
			      "("+ normalizeConstraint(postText) + ")" +" and "
			      +"("+ normalizeConstraint(invText)+")", 
			      xmifile).getErrorOccurred()){
	    return "Invariant or Pre-/Post-Condition syntactically incorrect";
	}

	PrePostPair ppp = ContractUtil.getPreInvPostInv(aMethRepr);
	OCLProofOblInput po = 
            new OCLPostPO("EnsuresPostConditionAndInvariant"+totality, 
			  aMethRepr, ppp.getPre(), ppp.getPost(), totality, 
			  xmifile);
	return startProver(po);
    }


    /**
     * 
     */
    public static String methCorrectModifies(ReprModelMethod aMethRepr) {	
	ModifiesCheckProofOblInput po = new ModifiesCheckProofOblInput(
					    "ModifiesCheck",
					    aMethRepr);
	return startProver(po);
    }

    /**
     * Computes and displays the specification of the method.
     * Tries to compute the strongest specification (pre and postcondition)
     * of <code>aMethRepr</code>.
     * @param aMethRepr the method whose specification to compute.
     * @return any error messages to pass to the user.
     * @see de.uka.ilkd.key.cspec.ComputeSpecification
     * @see de.uka.ilkd.key.proof.init.ComputeSpecificationPO
     */
    public static String methComputeSpecification(ReprModelMethod aMethRepr) {
	String preText = aMethRepr.getMyPreCond();
	String postText = aMethRepr.getMyPostCond();
	String invText = aMethRepr.getContainingClass().getMyInv();

	// test iff method is constructor
	String poPreText = "";
	if (aMethRepr.getContainingClassName().equals(aMethRepr.getName())){
	    // invText is ignored for po-generation because it still has to be assured in the postcondition
	    poPreText = normalizeConstraint(preText);
	} else {
	    poPreText= "(" + normalizeConstraint(preText) + " ) and (" 
		+  normalizeConstraint(invText) + ")";
	}

	OCLMod[] mods=ProofSettings.DEFAULT_SETTINGS.getOCLRefinementSettings()
	                       .getOCLMods();
	String modSource=ProofSettings.DEFAULT_SETTINGS
	    .getModelSourceSettings().getModelSource();	
	String xmifile=prepareXMI(aMethRepr.getActXmifile(), mods, modSource);

	poPreText=prepareConstraint(poPreText, mods, modSource);
	postText=prepareConstraint(postText, mods, modSource);
	
	try {	
	    if (parseMethSpecCore(aMethRepr, poPreText, 
				  "("+normalizeConstraint(postText)+")"+" and "
				  +"("+normalizeConstraint(invText)+")", 
				  xmifile).getErrorOccurred()){
		return "Invariant or Pre-/Post-Condition syntactically incorrect";
	    }
	} catch (Exception e) {
	    return e.getMessage();
	}

	OCLProofOblInput po = new ComputeSpecificationPO("ComputeSpecification",
					    aMethRepr, poPreText, 
					    "("+normalizeConstraint(postText)+")"
					    + " and "
					    + "("+normalizeConstraint(invText)
					    +")", xmifile);
	return startProver(po);
    }

    /**
     * Check whether the implementation of the method 
     * meets the specification of the invariant.
     * @param aMethRepr the method whose implementation should be checked.
     * @param totality the Modality to use 
     * @return error message to the user if prover was not invoked
     */
    public static String methPreservesInvariant(ReprModelMethod aMethRepr, 
						Modality totality) {
	String preText = aMethRepr.getMyPreCond();
	String invText = aMethRepr.getContainingClass().getMyInv();

	if (normalizeConstraint(invText).equals("true")){
	    return "no or trivial invariant available";
	}

	// test iff method is constructor
	String poPreText = "";
	if (aMethRepr.getContainingClassName().equals(aMethRepr.getName())){
	    // preText and invText are ignored for po-generation
	    poPreText="";
	}else{
	    poPreText= "("+normalizeConstraint(preText)+")" 
		+ " and " +  "("+normalizeConstraint(invText)+")";
	}

	OCLMod[] mods=ProofSettings.DEFAULT_SETTINGS
	    .getOCLRefinementSettings().getOCLMods();

	String modSource=ProofSettings.DEFAULT_SETTINGS
	    .getModelSourceSettings().getModelSource();
	
	String xmifile=prepareXMI(aMethRepr.getActXmifile(), mods, modSource);
	poPreText=prepareConstraint(poPreText, mods, modSource);
	
	// test for syntactical correctness
	if (parseMethSpecCore(aMethRepr, "("+poPreText+")" +" and "
			      +"("+normalizeConstraint(invText)+")", 
			      " true ", xmifile).getErrorOccurred()){
	    return "Invariant or Pre-/Post-Condition syntactically incorrect";
	}
	PrePostPair ppp = ContractUtil.getPreInvInv(aMethRepr);
	OCLProofOblInput po = 
            new OCLPostPO("PreservesInvariant"+totality,
			  aMethRepr, ppp.getPre(), ppp.getPost(), totality, 
			  xmifile);
	return startProver(po);
    }


    /**
     * Check whether the implementation of the method 
     * meets the specification of the throughout property.
     * @param aMethRepr the method whose implementation should be checked.
     * @return error message to the user if prover was not invoked
     */
    public static String methPreservesThroughout(ReprModelMethod aMethRepr) {
	String preText = aMethRepr.getMyPreCond();
	String toutText = aMethRepr.getContainingClass().getMyThroughout();
	String invText = aMethRepr.getContainingClass().getMyInv();

	if (normalizeConstraint(toutText).equals("true")){
	    return "no or trivial throughout property available";
	}

	// test iff method is constructor
	String poToutText = "";
	if (aMethRepr.getContainingClassName().equals(aMethRepr.getName())){
	    // preText and toutText are ignored for po-generation
	    poToutText="";
	}else{
	    //	    poToutText= normalizeConstraint(preText) 
	    //		+ " and " +  normalizeConstraint(toutText);
	    poToutText= "("+normalizeConstraint(preText)+")" 
                + " and " + "(" + normalizeConstraint(invText) + ")" 
		+ " and " + "("+normalizeConstraint(toutText)+")";
	}

	OCLMod[] mods=ProofSettings.DEFAULT_SETTINGS
	    .getOCLRefinementSettings().getOCLMods();
	String modSource=ProofSettings.DEFAULT_SETTINGS
	    .getModelSourceSettings().getModelSource();	
	String xmifile=prepareXMI(aMethRepr.getActXmifile(), mods, modSource);

	poToutText=prepareConstraint(poToutText, mods, modSource);

	// test for syntactical correctness
	if (parseMethSpecCore(aMethRepr, poToutText, 
			      " true ", xmifile).getErrorOccurred()){
	    return "Throughout/Invariant/Precondition syntactically incorrect";
	}
	PrePostPair ppp = ContractUtil.getPreInvTout(aMethRepr);
	OCLProofOblInput po = new OCLPostPO("PreservesThroughout",
					    aMethRepr, ppp.getPre(), 
					    ppp.getPost(), 
					    Op.TOUT,
					    xmifile);
	return startProver(po);
    }


    /**
     * Check whether the specification of the method
     * meets the co-variance properties (formulated by B.Liskov).
     * @param aMethRepr the method whose pre/post specification should be checked.
     * @return error message to the user
     */
    public static String checkLiskovOnMethSpec(ReprModelMethod aMethRepr) {
   	if (!aMethRepr.hasOrigParent()){
	    return noSuperClassAvailableText();
	}

	String preText = aMethRepr.getMyPreCond();
	String postText = aMethRepr.getMyPostCond();
	String preTextParent = aMethRepr.getParentPreCond();
	String postTextParent = aMethRepr.getParentPostCond();

	String xmifile = aMethRepr.getActXmifile();
	String contextString = aMethRepr.getContainingClassName() + "::"+ aMethRepr.getCallSignature();
	String parentContextString = aMethRepr.getContainingClass().getParentClassName()+
	    "::"+ aMethRepr.getParentMethod().getCallSignature();

	// generating proofobligation
	//	pre-condition    preTextParent -> preText
	//	post-condition   preTextParent and preText -> rename_nonrigid(postText -> postTextParent)

	
	OCLBehavSubPO po = new OCLBehavSubPO(aMethRepr, 
					     aMethRepr.getParentMethod(), 
					     xmifile, contextString, 
					     parentContextString, 
					     preTextParent, 
					     preText, postTextParent, postText);
	return startProver(po);	
    }


    /**
     * Check whether the specification of the invariants
     * meets the co-variance properties (formulated by B.Liskov).
     * @param aClRepr the class whose invariant specification should be checked.     
     * @return error message to the user
     */
    public static String checkLiskovOnInvariant(ReprModelClass aClRepr) {
	if (!aClRepr.hasOrigParent()){
	    return noSuperClassAvailableText();
	}

	String invText = aClRepr.getMyInv();
	String invTextParent = aClRepr.getParentInv();

	String xmifile = aClRepr.getActXmifile();

	OCLStructSubPO po = new OCLStructSubPO("StructuralSubtyping", 
					       xmifile, 
					       aClRepr,
					       invTextParent, invText);
	return startProver(po);
    }



    public static String transformXMIFile(ReprModel aReprModel){
	return "";
    }


    /**
     * Check whether the specification of all methods which are
     * inherited from parent class but not redefined obey class invariant
     * of current class
     * @param aClRepr class whose invariant and methods are checked
     * @return error message to the user
     */
    
    public static String inheritedMethodsPreserveInvariant
	(ReprModelClass aClRepr) {

	if (!aClRepr.hasOrigParent()){
	    return noSuperClassAvailableText();
	}

    	Enumeration opNameEnum;
	Enumeration parentOpNameEnum = (aClRepr.getInheritedOpNames()).elements();
	Enumeration parentOpEnum = (aClRepr.getInheritedOps()).elements();
	Vector selectedOpVector = new Vector();
	boolean found;
        // we compare each op-name of the parent class with each op-name
        // of the current class
	while (parentOpNameEnum.hasMoreElements()){
	    String currParentOpName = (String) parentOpNameEnum.nextElement();
	    ReprModelMethod currParentOp 
		= (ReprModelMethod) parentOpEnum.nextElement();
	    found = false;
            opNameEnum = (aClRepr.getOpNames()).elements();
	    while (opNameEnum.hasMoreElements()){
		String currOpName = (String) opNameEnum.nextElement();
		if (currOpName.equals(currParentOpName)){found=true;}
	    }
	    if (!found 
		&& !(currParentOpName.equals(aClRepr.getParentClassName()))){
            	// we add the op-name from parent class iff there is no such
            	// op-name in current class and if it is not the constructor of
            	// parent class
		selectedOpVector.add(currParentOp);
	    }
	}
	
	// now we have in selectedOpVector all operations for which
	// proof obligations must be generated

	if (selectedOpVector.isEmpty()){
	    return "Inherited operation not found";
	}

	Enumeration selectedOpEnum = selectedOpVector.elements();
	String invText = aClRepr.getMyInv();
	String invParentText = aClRepr.getParentInv();

	OCLMod[] mods=ProofSettings.DEFAULT_SETTINGS
	    .getOCLRefinementSettings().getOCLMods();
	String modSource=ProofSettings.DEFAULT_SETTINGS
	    .getModelSourceSettings().getModelSource();	
	String xmifile=prepareXMI(aClRepr.getActXmifile(), mods, modSource);

	//%%%%%%% 
	// there should be only one PO generated
	// instead of one for each while loop iteration -AR
	while (selectedOpEnum.hasMoreElements()){
	    ReprModelMethod currOpRepr 
		= (ReprModelMethod) selectedOpEnum.nextElement();

	    // since the selected methods are not redefined in the subclass
	    // there is also no ReprModelMethod and we have to create one
	    ReprModelMethod current = 
		new DefaultReprModelMethodImpl(aClRepr, currOpRepr, currOpRepr,
					       currOpRepr.getName(), 
					       currOpRepr.isVoid(),
					       currOpRepr.isConstructor(),
					       currOpRepr.isPrivate());


	    String preText = current.getParentPreCond();
	    String postText = current.getParentPostCond();
	    preText=prepareConstraint(preText, mods, modSource);
	    postText=prepareConstraint(postText, mods, modSource);


	    // test for syntactical correctness
	    TUDresdenParserAdapter parseResult =
		parseMethSpecCore(current,  normalizeConstraint(preText)
				  +" and "+normalizeConstraint(invText)
				  +" and "
				  +normalizeConstraint(invParentText), 
				  normalizeConstraint(postText), xmifile);

	    if (parseResult.getErrorOccurred()){
		return "Invariant or Pre-/Post-Condition "
		    +"syntactically incorrect: " + 
		    parseResult.getErrorMessageToUser();
	    }

	    OCLInvInhMethPO po = new OCLInvInhMethPO(current, xmifile, 
						     preText, invText, 
						     invParentText, 
						     postText);
	    String s = startProver(po);
            if (!"".equals(s)) return s;
	}
	return "";
    }

    public static String proveDLFormula(ReprModelClass aReprModelClass,
					File file) {

	CasetoolDLPO dlPOFromCasetool =  new CasetoolDLPO(aReprModelClass,
							  file);
	return startProver(dlPOFromCasetool);
    }


 // ****************   Utilities   ********************************

    private static String noSuperClassAvailableText(){
        return " No superclass in current diagram available ";
    }


     /**
     * Converts an empty string to "true" and keeps it unchanged otherwise.
     */
    public static String normalizeConstraint(String aText) {
        if (aText.equals("")){
            return "true";
        }else{
            return aText;
        }
    }


    protected static String prepareXMI(String xmiFileName, OCLMod[] mods,
				       String modelSource) {
	if (! "0".equals(modelSource)) return null;
	if (mods.length==0) return xmiFileName;
	Debug.out("prepare XMI");
	DetailedPrep prep = new DetailedPrep(xmiFileName);
	prep.addListeners(mods);
	prep.transform();
	prep.writeToFile(xmiFileName+".det");
	Debug.out("prepare XMI finished");
	return xmiFileName+".det";
    }

    protected static String prepareConstraint(String orig, OCLMod[] mods,
					      String modelSource) {
	if (! "0".equals(modelSource)) return orig;
	if (mods.length==0) return orig;
	Debug.out("prepare OCL-Constraint");
	DetailedConstraintPrep prep = new DetailedConstraintPrep(orig);
	prep.addListeners(mods);
	prep.transform();
	return prep.result();
    }

    /** 
     * Used for OCL Simplification.
     * @param aClass The representation of the class whose invariant
     * needs to be simplified.
     */
    public static String simplifyInvariant(ReprModelClass aClass) {
	OCLProofOblInput po = new OCLInvSimplPO(aClass);
	return startProver(po);
    }
    
    public static void main(String[] argv) {
       String xmifile = "/home/klebanov/keydemo/paycardInteractive/paycard.xml";
//       String xmifile = "/home/klebanov/XMIAdapter/xml.txt";
       String constraint = "context PayCard inv: (self.balance > 0)";
       TUDresdenParserAdapter aAdapter = new TUDresdenParserAdapter();
       aAdapter.doParse1("Debug constraint",  
			 OCLProofOblInput.getModelManager(),  
			 xmifile, constraint, null);
    }


}  // end of class
