// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.cspec;

import java.util.Vector;

import de.uka.ilkd.key.ProverFacade;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.logic.JavaBlock;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.ocl.ocldltranslator.KeyOclConstraintTranslator;
import de.uka.ilkd.key.ocl.ocldltranslator.KeyOclExpressionTranslator;
import de.uka.ilkd.key.ocl.ocldltranslator.detailedocl.ExcThrownMod;
import de.uka.ilkd.key.ocl.ocldltranslator.detailedocl.OCLMod;

/**
 * Trick class for excerpting the method call block.  This is not a
 * true OCLMod, but only abuses its notification mechanism in order to
 * detect the method call within an OCL constraint (or more precisely,
 * its transformation to DL).
 * @author Andr&eacute; Platzer
 * @version 0.7, 2003-02-28
 * @attribute stateful
 * @see de.uka.ilkd.key.cspec.ComputeSpecification
 * @see de.uka.ilkd.key.proof.init.ComputeSpecificationPO
 */
public class ExcerptMethodCallMod extends OCLMod {

    /**
     * Remembers the last method call that occured (due to a call to
     * {@link #getMethodCallJavaBlock}).
     * @internal The volatile modifier would only be strictly
     * necessary in case of concurrent usage of this object.
     */
    private volatile JavaBlock methodCall = null;

    /**
     * Get the last method call that occured (due to a call to
     * {@link #getMethodCallJavaBlock}).
     */
    public final JavaBlock getMethodCall() {
	return methodCall;
    }

    private final void setMethodCall(JavaBlock argOccuringMethodCall) {
	this.methodCall = argOccuringMethodCall;
    }

    /**
     * Called when a method call block is constructed.
     * Remembers the method call block in {@link #getMethodCall()}.
     * @see ExcThrownMod#getMethodCallJavaBlock(KeyOclConstraintTranslator, String, ProgramVariable[], String, String)
     * @see KeyOclConstraintTranslator#getMethodCallJavaBlock(ProgramVariable[])
     */
    public JavaBlock getMethodCallJavaBlock(KeyOclConstraintTranslator 
					    consTransl,
					    String methodReturnType,
					    Vector methodParameterNames, 
					    Vector methodParameterTypes, 
					    String constrainedOperation, 
					    String constrainedType) {

	// First create the program variables result, p1, ..., pn and the
	// corresponding JavaBlock ... 

	KeyOclExpressionTranslator exprTransl=consTransl.exprTransl();
	ProverFacade proverFacade = consTransl.getProverFacade();
	boolean voidMethodCall = (methodReturnType == null);
	ProgramVariable[] argProgVars 
	    = new ProgramVariable[methodParameterNames.size()];
	ProgramVariable currentProgVar;
	String currentParamName, currentParamTypeName;
	KeYJavaType currentParamType;
	for(int i=0;i<methodParameterNames.size();i++){
	    currentParamName = (String) methodParameterNames.elementAt(i);
	    currentParamTypeName=(String) methodParameterTypes.elementAt(i);
	    currentParamType = exprTransl
		.getKeYJavaTypeForNonCollectionType(currentParamTypeName);
	    currentProgVar = consTransl.getProverFacade()
		.getProgramVariable(currentParamName,currentParamType);
	    argProgVars[i] = currentProgVar;
	}
		
	JavaBlock methodCallJavaBlock = null; 
	KeYJavaType calledObjectVarType = exprTransl
	    .getKeYJavaTypeForNonCollectionType(constrainedType);
		
	ProgramVariable calledObjectVar = consTransl.getProverFacade()
	    .getProgramVariable("self",calledObjectVarType);

	if (!voidMethodCall) {
	    // we do have a specified return type for the method ...
		    
	    // Create a program variable for the result of the method ...
	    KeYJavaType methodResultType 
		= exprTransl.getKeYJavaTypeForNonCollectionType
		(methodReturnType);
	    ProgramVariable resultProgVar 
		= proverFacade.getProgramVariable("result",
						    methodResultType);
		    
	    methodCallJavaBlock 
		= proverFacade.getMethodCall(calledObjectVar,
					       constrainedOperation,
					       resultProgVar,
					       argProgVars);
	} else {
	    // we do not have a specified return type for the method ...
	    methodCallJavaBlock 
		= proverFacade.getVoidMethodCall(calledObjectVar,
						   constrainedOperation,
						   argProgVars);
	}
	if (getMethodCall() == null)
	    this.setMethodCall(methodCallJavaBlock);
	else
	    throw new IllegalStateException("Only one single method call expected within the translation of a single method's OCL specification");
	return methodCallJavaBlock;
    }

}
