// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.gui;

import java.awt.Point;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.dnd.DropTargetAdapter;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;

import javax.swing.JPopupMenu;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.Constraint;
import de.uka.ilkd.key.logic.PosInOccurrence;
import de.uka.ilkd.key.logic.PosInTerm;
import de.uka.ilkd.key.logic.Sequent;
import de.uka.ilkd.key.pp.PosInSequent;
import de.uka.ilkd.key.rule.*;

/**
 * <p>
 * The DragNDropInstantiator interpretes drag'n drop actions as taclet
 * instantiations. The behaviour is described below (excluding some
 * "optimisation" details)
 * </p>
 * <p>
 * Let "source" denote the formula/term which is dragged and dropped on another
 * term called "target". The DragNDropInstantiation of a taclet "t" is now
 * defined as follows:
 * <ol>
 * <li>t must have exactly one formula/term in the <tt>find</tt> part and
 * <tt>if</tt> sequent (in particular it must have an <tt>if</tt> -part)
 * </li>
 * <li>and
 * <ol>
 * <li>if t's goal descriptions do not contain any <tt>replacewith</tt> then
 * "source" is matched against the <tt>find</tt> part of the taclet and
 * "target" has to match the "if" part.</li>
 * <li>if t's goal descriptions contain at least <em>one</em> replacewith
 * then "target" is matched against the <tt>find</tt> part and "source"
 * against the <tt>if</tt> part</li>*
 * </ol>
 * or
 * <li>t must have a <tt>find</tt> part, no <tt>if</tt> and at least one
 * addrule. In this case source is merged against <tt>find</tt> part and
 * target has to be the complete sequent. Dropping on the sequent arrow is
 * interpreted as applying an addrule(treats hide rules)</li>
 * </ol>
 * The DragNDropInstantiator now determines all taclets, which have a valid
 * drag'n drop instantiation for a given (source, target) position pair. If
 * there is only one taclet with a valid dnd-instantiation this one is executed
 * otherwise the user is presented a list of possible taclets from which she/he
 * can select one.
 * </p>
 */
public class DragNDropInstantiator extends DropTargetAdapter {

    /** the sequentview where dnd has been initiated */
    private SequentView seqView;
    
    DragNDropInstantiator(SequentView seqView) {
        this.seqView = seqView;
    }

    public void drop(DropTargetDropEvent event) {


        Point dropLocation = event.getLocation();

        try {
            Transferable transferable = event.getTransferable();
            if (transferable
                    .isDataFlavorSupported(PosInSequentTransferable.POS_IN_SEQUENT_TRANSFER)) {
                interpreteDragAndDropInstantiation(event, dropLocation,
                        transferable);
            } else {
                event.rejectDrop();
            }
        } catch (IOException exception) {
            // just reject drop do not bother the user
            event.rejectDrop();
        } catch (UnsupportedFlavorException ufException) {
            // just reject drop do not bother the user
            event.rejectDrop();
        }

        seqView.requestFocus();
    }

    public void dragOver(DropTargetDragEvent dtde) {                                                  
        seqView.paintHighlights(dtde.getLocation());
    }

    /**
     * Interpretes the drag and drop gesture. Checks which taclets could be
     * meant by the drag'n drop and applies if the app can be uniquely
     * determined, otherwise a selection menu is presented to the user
     */
    private void interpreteDragAndDropInstantiation(DropTargetDropEvent event,
            Point dropLocation, Transferable transferable)
            throws UnsupportedFlavorException, IOException {

        final PosInSequent sourcePos = (PosInSequent) transferable
                .getTransferData(PosInSequentTransferable.POS_IN_SEQUENT_TRANSFER);

        final PosInSequent targetPos = seqView.getPosInSequent(dropLocation);

        
        final Services services = seqView.mediator().getServices();

        // no action is taken in cases of equal source and target position
        // (shall avoid to accidently apply replace_knwon_* rules and entering
        // the hell of non-confluence...)
        if (targetPos == null
                || sourcePos == null
                || sourcePos.isSequent()
                || sourcePos.getPosInOccurrence().equals(
                        targetPos.getPosInOccurrence())) {
            event.rejectDrop();
            return;
        }

        ListOfPosTacletApp applicableApps = 
            getAllApplicableApps(sourcePos, targetPos, services);        
               
        
        if (applicableApps.isEmpty() && !targetPos.isSequent() &&
                targetPos.getPosInOccurrence().posInTerm() != PosInTerm.TOP_LEVEL) {
            // if no applicable taclet is found we relax the target position a bit
            applicableApps = 
                getAllApplicableApps(sourcePos, 
                        PosInSequent.createCfmaPos(targetPos.getPosInOccurrence().up()), 
                        services);
        } 
        
        if (applicableApps.size() == 1) {
            execute(applicableApps.head());      
        } else if (applicableApps.size() > 1) {
            // open a pop up menu for user selection
            SimpleTacletSelectionMenu menu = new SimpleTacletSelectionMenu(
                    applicableApps, seqView.mediator().getNotationInfo(),
                    new PopupListener());

            JPopupMenu pm = menu.getPopupMenu();
            pm.show(seqView, (int) dropLocation.getX(), (int) dropLocation
                    .getY());
        }
        event.getDropTargetContext().dropComplete(true);
    }

    /**
     * retrieves all drag'n drop instantiable taclet applications
     * @param sourcePos the PosInSequent where the drag started
     * @param targetPos the PosInSequent where the drop occured     
     * @param services theServices providing access to the program model  
     * @return all drag'n drop instantiable taclet applications
     */
    private ListOfPosTacletApp getAllApplicableApps(final PosInSequent sourcePos, 
            final PosInSequent targetPos, final Services services) {
        final Sequent sequent = seqView.mediator().getSelectedGoal().sequent();
        final Constraint userConstraint = seqView.mediator()
                .getUserConstraint().getConstraint();
        ListOfPosTacletApp applicableApps = SLListOfPosTacletApp.EMPTY_LIST;
        if (targetPos.isSequent()) {
            // collects all applicable taclets at the source position
            // which have an addrule section
            applicableApps = applicableApps.prepend(completeInstantiations(
                    getApplicableTaclets(sourcePos,
                            TacletFilter.TACLET_WITH_NO_IF_FIND_AND_ADDRULE),
                    sequent, targetPos.getPosInOccurrence(), userConstraint,
                    services));
        } else {
            // all applicable taclets where the drag source has been interpreted
            // as
            // the find part and the drop position as the one of the
            // in this case only taclets with no replacewith part are considered
            applicableApps = applicableApps
                    .prepend(completeInstantiations(
                            getApplicableTaclets(
                                    sourcePos,
                                    TacletFilter.TACLET_WITH_IF_FIND_AND_NO_REPLACEWITH),
                            sequent, targetPos.getPosInOccurrence(),
                            userConstraint, services));

            // switch source and target interpretation
            // source is now the "If" instantiation and target the one of the
            // find part in this case only the taclets with at least one
            // replacewith part are considered
            applicableApps = applicableApps.prepend(completeInstantiations(
                    getApplicableTaclets(targetPos,
                            TacletFilter.TACLET_WITH_IF_FIND_AND_REPLACEWITH),
                    sequent, sourcePos.getPosInOccurrence(), userConstraint,
                    services));
        }
        return applicableApps;
    }

    /**
     * retrieves all taclets applicable on the given position and select those
     * that fulfill the given filter condition. The filter selects taclets
     * satisfying certain syntactic categories (for example, taclets with at
     * least one replacewith in their goal description or those without any
     * replacewith).
     * 
     * @param findPos
     *            the PosInSequent specifying the formula/term that has to be
     *            matched by the find part of a taclet
     * @param filter
     *            the TacletFilter specifying syntactic restrictions on the
     *            taclets to be returned
     * @return the list of taclets which match the term at the given position
     *         and satisfy the filter condition
     */
    private ListOfPosTacletApp getApplicableTaclets(PosInSequent findPos,
            TacletFilter filter) {

        if (findPos == null || findPos.isSequent()) {
            return SLListOfPosTacletApp.EMPTY_LIST;
        }

        ListOfTacletApp allTacletsAtFindPosition = SLListOfTacletApp.EMPTY_LIST;

        final IteratorOfTacletApp it = seqView.mediator()
                .getFindTaclet(findPos).iterator();
        // if in replaceWithMode only apps that contain at least one replacewith
        // are collected. Otherwise only those without a replacewith.
        while (it.hasNext()) {
            final TacletApp app = it.next();
            if (filter.satisfiesFilterCondition(app.taclet())) {
                allTacletsAtFindPosition = allTacletsAtFindPosition
                        .prepend(app);
            }
        }

        return addPositionInformation(allTacletsAtFindPosition, findPos
                .getPosInOccurrence());
    }

    /**
     * the taclet applications is given the correct position information where
     * their "find" has been matched
     * 
     * @param tacletApps
     *            the ListOfTacletApp with taclet applications to be enriched by
     *            position information
     * @param findPos
     *            the PosInOccurrence against which the find part has been
     *            matched
     * @return the taclet apps as given in <tt>tacletApps</tt> but with
     *         position information
     */
    private ListOfPosTacletApp addPositionInformation(
            ListOfTacletApp tacletApps, PosInOccurrence findPos) {

        ListOfPosTacletApp applicableApps = SLListOfPosTacletApp.EMPTY_LIST;
        IteratorOfTacletApp it = tacletApps.iterator();
        while (it.hasNext()) {
            TacletApp app = it.next();
            if (app instanceof NoPosTacletApp) {
                app = PosTacletApp.createPosTacletApp(
                        (FindTaclet) app.taclet(), app.matchConditions(),
                        findPos);
            }
            applicableApps = applicableApps.prepend((PosTacletApp) app);
        }
        return applicableApps;
    }

    /**
     * tries to complete the (partial) taclet instantantiation of the
     * applications given in <tt>apps</tt>. The resulting applications are
     * returned. The given apps must have either all an if part or none of them.
     * 
     * @param apps
     *            the ListOfPosTacletApp with all apps whose if sequent has to
     *            be matched against the formula specified by the pair
     *            <tt>seq</tt> and <tt>ifPIO</tt>
     * @param seq
     *            the Sequent to which the position information in <tt>ifPIO<tt>
     * is relative to
     * @param ifPIO the PosInOccurrence describing the position of the term to 
     * be matched against the if sequent of the taclets
     * @param userConstraint the Constraint describing user instantiations of
     * metavariables
     * @param services the Services 
     * @return the ListOfPosTacletApp that have been matched successfully
     */
    private ListOfPosTacletApp completeInstantiations(ListOfPosTacletApp apps,
            Sequent seq, PosInOccurrence ifPIO, Constraint userConstraint,
            Services services) {

        ListOfPosTacletApp result = SLListOfPosTacletApp.EMPTY_LIST;

        final ListOfIfFormulaInstantiation ifFmlInst;

        if (ifPIO == null || !ifPIO.isTopLevel()) {
            // if formula have to be top level formulas
            // TODO: should update prefix be allowed?
            ifFmlInst = SLListOfIfFormulaInstantiation.EMPTY_LIST;
        } else {
            final IfFormulaInstSeq ifInst = new IfFormulaInstSeq(seq,
                    ifPIO.isInAntec(), ifPIO.constrainedFormula());
            ifFmlInst = SLListOfIfFormulaInstantiation.EMPTY_LIST
                    .prepend(ifInst);
        }

        final IteratorOfPosTacletApp it = apps.iterator();
        while (it.hasNext()) {
            PosTacletApp app = it.next();

            final Sequent ifSequent = app.taclet().ifSequent();
            if (ifSequent != null && !ifSequent.empty()) {
                if (ifFmlInst == SLListOfIfFormulaInstantiation.EMPTY_LIST) {
                    // as either all taclets have an if sequent or none
                    // we can exit here
                    return SLListOfPosTacletApp.EMPTY_LIST;
                } else {    
                    // the right side is not checked in tacletapp
                    // not sure where to incorporate the check...
                    if (((IfFormulaInstSeq)ifFmlInst.head()).inAntec() == 
                         (ifSequent.succedent().size() == 0)) {        	    
                        app = (PosTacletApp) app.setIfFormulaInstantiations(
                                ifFmlInst, services, userConstraint);
                    }
                }
            }
            if (app != null && app.complete()) {
                //allow use of sufficientlyComplete here?
                result = result.prepend(app);
            }
        }
        return result;
    }

    /**
     * applies the given app
     * 
     * @param app
     *            the PosTacletApp to be applied
     */
    private void execute(PosTacletApp app) {
        if (app == null) {
            return;
        }
        seqView.mediator().applyInteractive(app,
                seqView.mediator().getSelectedGoal());
    }

    /**
     * This popup listener listens to the taclet app selection listener and
     * executes the selected app if necessary
     */
    private class PopupListener implements ActionListener {

        /*
         * (non-Javadoc)
         * 
         * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
         */
        public void actionPerformed(ActionEvent e) {
            if (e.getSource() instanceof DefaultTacletMenuItem) {
                final TacletMenuItem item = (TacletMenuItem) e
                        .getSource();
                DragNDropInstantiator.this.execute((PosTacletApp) item
                        .connectedTo());
            }
        }

    }

    /**
     * This interface is impemented by filters selecting certain kinds of
     * taclets depending on their syntactic structure
     */
    protected interface TacletFilter {

        /**
         * This filter selects all Taclet which have an <tt>if</tt>,
         * <tt>find</tt> and at least one <tt>replacewith</tt> part.
         */
        TacletFilter TACLET_WITH_IF_FIND_AND_REPLACEWITH = new TacletWithIfFindAndReplacewith();

        /**
         * This filter selects all Taclet which have an <tt>if</tt>,
         * <tt>find</tt> and <em>no</em> <tt>replacewith</tt> part.
         */
        TacletFilter TACLET_WITH_IF_FIND_AND_NO_REPLACEWITH = new TacletWithIfFindAndNoReplacewith();

        /**
         * This filter selects all Taclet which have an <tt>if</tt>,
         * <tt>find</tt> and <em>no</em> <tt>replacewith</tt> part.
         */
        TacletFilter TACLET_WITH_NO_IF_FIND_AND_ADDRULE = new TacletWithNoIfFindAndAddrule();

        /**
         * checks if the taclet satisfies certain syntactic criterias
         * 
         * @param taclet
         *            the Taclet to be tested
         * @return true if filter condition is fulfilled
         */
        boolean satisfiesFilterCondition(Taclet taclet);

        /**
         * This filter selects all Taclet which have an <tt>if</tt>,
         * <tt>find</tt> and at least one <tt>replacewith</tt> part.
         */
        class TacletWithIfFindAndReplacewith implements TacletFilter {

            private TacletWithIfFindAndReplacewith() {
            }

            /**
             * tests if the given taclet consists of an <tt>if</tt>,
             * <tt>find</tt> and <tt>replacewith</tt> part and returns true
             * if the test is positive
             */
            public boolean satisfiesFilterCondition(Taclet taclet) {
                return taclet.ifSequent() != null
                        && taclet.ifSequent() != Sequent.EMPTY_SEQUENT
                        && taclet instanceof FindTaclet
                        && taclet.hasReplaceWith();
            }
        }

        /**
         * This filter selects all Taclet which have an <tt>if</tt>,
         * <tt>find</tt> and <em>no</em> <tt>replacewith</tt> part.
         */
        class TacletWithIfFindAndNoReplacewith implements TacletFilter {

            private TacletWithIfFindAndNoReplacewith() {
            }

            /**
             * tests if the given taclet consists of an <tt>if</tt>,
             * <tt>find</tt> and <em>no</em> <tt>replacewith</tt> part and
             * returns true if the test is positive
             */
            public boolean satisfiesFilterCondition(Taclet taclet) {
                return taclet.ifSequent() != null
                        && taclet.ifSequent() != Sequent.EMPTY_SEQUENT
                        && taclet instanceof FindTaclet
                        && !taclet.hasReplaceWith();
            }
        }

        /**
         * This filter selects all Taclet which have no <tt>if</tt>, but a
         * <tt>find</tt> and at least one <tt>addrule</tt> section.
         */
        class TacletWithNoIfFindAndAddrule implements TacletFilter {

            private TacletWithNoIfFindAndAddrule() {
            }

            /**
             * tests if the goal templates contain at least one addrule section
             * 
             * @param goalDescriptions
             *            the ListOfTacletGoalTemplate to be looked through
             * @return true if an addrule section has been found
             */
            private boolean goalTemplatesContainAddrules(
                    ListOfTacletGoalTemplate goalDescriptions) {
                final IteratorOfTacletGoalTemplate it = goalDescriptions
                        .iterator();
                while (it.hasNext()) {
                    final TacletGoalTemplate tgt = it.next();
                    if (tgt.rules().size() >= 1) {
                        return true;
                    }
                }

                return false;
            }

            /**
             * tests if the given taclet consists of an <tt>if</tt>,
             * <tt>find</tt> and <em>no</em> <tt>replacewith</tt> part and
             * returns true if the test is positive
             */
            public boolean satisfiesFilterCondition(Taclet taclet) {
                // TODO: the null checks should be unneccessary
                return (taclet.ifSequent() == null || taclet.ifSequent() == Sequent.EMPTY_SEQUENT)
                        && taclet instanceof FindTaclet
                        && goalTemplatesContainAddrules(taclet.goalTemplates());
            }
        }

    }
}

