package de.uka.ilkd.key.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.*;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.text.Caret;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.Sequent;
import de.uka.ilkd.key.pp.LogicPrinter;
import de.uka.ilkd.key.pp.NotationInfo;
import de.uka.ilkd.key.pp.ProgramPrinter;
import de.uka.ilkd.key.rule.ListOfTacletGoalTemplate;
import de.uka.ilkd.key.rule.NoFindTaclet;
import de.uka.ilkd.key.rule.Taclet;
import de.uka.ilkd.key.rule.TacletApp;
import de.uka.ilkd.key.util.Debug;

/**
 * This item groups all insert hidden taclets and offers a
 * more convienient user interface to add them.
 *
 */
public class InsertHiddenTacletMenuItem extends JMenu
                                        implements TacletMenuItem {
    
    /** the selected taclet to be applied */
    private TacletApp selectedTaclet;
    
    /** all taclet apps the user can choose from */
    private List insertHiddenTaclets = new LinkedList();

    /** the added action listeners */
    private List listenerList = new LinkedList();

    /** the parent frame of the selection dialog to be displayed */
    private JFrame parent;

    /** the notation info to pretty print the taclet apps */
    private NotationInfo notInfo;
    
    /** the services */
    private Services services;
    
    
    /**
     * creates an instance of the insert hidden menu item
     * @param parent the JFrame with the parent frame
     * @param notInfo the NotationInfo to be used for pretty printing 
     * the apps
     * @param services the Services
     */
    public InsertHiddenTacletMenuItem(JFrame parent, 
            NotationInfo notInfo, Services services) {
        super("Insert Hidden");
        this.parent = parent;
        this.notInfo = notInfo;
        this.services = services;
        
        final JMenuItem menuItem = new JMenuItem("Open Dialog");
        menuItem.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {              
                    openDialog();
            }
            
        });                  
        
        menuItem.setToolTipText("Easy selection of the formula to be added.");
        
        add(menuItem);
        addSeparator();
        
    }
    
    /**
     * tests if this class is responsible for the given taclet
     * @param t the Taclet to be checked
     * @return true if this item implementation shall be used
     */
    public static boolean isResponsible(Taclet t) {
        return checkTaclet(t) != null;
    }
    
    /**
     * determines the sequent with the formulas to be added
     * or null if the given taclet is not thought to be displayed 
     * by this component
     * @param t the Taclet
     * @return the sequent with the formulas to be added
     * or null
     */
    private static Sequent checkTaclet(Taclet t) {
        if (!(t instanceof NoFindTaclet) || 
                !t.displayName().startsWith("insert_hidden")) {
            return null;
        }
        
        final ListOfTacletGoalTemplate goalTemplates = t.goalTemplates(); 
        if (goalTemplates.size() != 1) return null;
        return goalTemplates.head().sequent();
    }
    
    /**
     * adds a new taclet to be displayed by this component
     * it is assumed that the app has been tested before by 
     * @link isResponsible  
     * @param app the TacletApp to be added
     */
    public void add(TacletApp app) {
        insertHiddenTaclets.add(new TacletAppListItem(app));
        final DefaultTacletMenuItem appItem = 
            new DefaultTacletMenuItem(this, app, notInfo);
        appItem.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent e) {
                processTacletSelected(e);
            }
            
        });
        add(appItem);
        
    }
    
    
    /** returns the selected taclet to be applied */
    public TacletApp connectedTo() {        
        return selectedTaclet;
    }

    /**
     * opens the selection dialog displaying all hidden formulas
     * in a list and allowing the user to select the one to be added     
     */
    public void openDialog() {
        final JDialog dialog      = 
            new JDialog(parent, "Insert Hidden Formula", true);
        
        final JList selectionList = 
            new JList(insertHiddenTaclets.toArray());
        
        final JScrollPane scrollPane = 
            new JScrollPane(selectionList, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, 
                    JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        scrollPane.setPreferredSize(new Dimension(300,100));
        scrollPane.setMinimumSize(new Dimension(150,50));
       
        
        selectionList.
            setSelectionMode(ListSelectionModel.SINGLE_SELECTION);  
             
        if (getAppSize()>0) { // should always be true
            selectionList.setSelectedIndex(0);
        }
        
        
        final JTextArea displayHiddenFormula = new JTextArea();
        
        displayHiddenFormula.
          setText(selectionList.getSelectedValue() == null ? "" :
              selectionList.getSelectedValue().toString());

        displayHiddenFormula.setCaretPosition(0);
        
        displayHiddenFormula.setEditable(false);
        
        selectionList.addListSelectionListener(new ListSelectionListener()
                {

                    public void valueChanged(ListSelectionEvent e) {                       
                        if (e.getSource() instanceof JList) {
                            final JList list = (JList)e.getSource();
                            if (list.getSelectedIndex()>=0) {
                                displayHiddenFormula.
                                  setText(list.getSelectedValue().toString());
                            } else {
                                displayHiddenFormula.setText("");
                            }
                            displayHiddenFormula.setCaretPosition(0);
                        }
                    }
                    
                });
        
        final JScrollPane formulaDisplaySP = new JScrollPane(displayHiddenFormula);
        
        final JSplitPane split = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
                     scrollPane, formulaDisplaySP) {
            public void setUI(javax.swing.plaf.SplitPaneUI ui) {
                try{ super.setUI(ui); } catch(NullPointerException e) 
                {Debug.out("Exception thrown by class Main at setUI"); }
            }
        }; // work around bug in 
//      com.togethersoft.util.ui.plaf.metal.OIMetalSplitPaneUI
        selectedTaclet = null;
                               
        final JButton cancel = new JButton("Cancel");
        cancel.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                selectedTaclet = null;
                dialog.setVisible(false);
                dialog.dispose();
            }          
        });
                    
        final JButton apply  = new JButton("Apply");
        apply.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {                
                final TacletAppListItem selectedItem = 
                    ((TacletAppListItem)selectionList.getSelectedValue());
                
                if (selectedItem == null) { // should never be true
                    selectedTaclet = null;
                } else {
                    selectedTaclet = selectedItem.getTacletApp();
                }
                
                dialog.setVisible(false);
                dialog.dispose();
                processTacletSelected(new ActionEvent
                        (InsertHiddenTacletMenuItem.this, 0, ""));
            }           
        });
        
        final JPanel buttonPanel = new JPanel();
        buttonPanel.setLayout(new FlowLayout());
        
        buttonPanel.add(apply);
        buttonPanel.add(cancel);

        dialog.getContentPane().setLayout(new BorderLayout());
        
        dialog.getContentPane().add(split, BorderLayout.CENTER);
        dialog.getContentPane().add(buttonPanel, BorderLayout.SOUTH);                
               
        dialog.setSize(500, 250);
       
        dialog.setVisible(true); 
        
    }

    protected void processTacletSelected(ActionEvent e) {
        final Iterator it = listenerList.iterator();
        while (it.hasNext()) {
            ((ActionListener)it.next()).actionPerformed(e);
        }
    }
    
    public void addActionListener(ActionListener listener) {
        listenerList.add(listener);
    }

    public void removeActionListener(ActionListener listener) {
        listenerList.remove(listener);
    }

    public int getAppSize() {
        return insertHiddenTaclets.size();
    }
    
    
    /**
     * inner class to pretty print the formulas to be added
     */
    class TacletAppListItem {
        private final TacletApp app;
        
        public TacletAppListItem(TacletApp app) {
            this.app = app;
        }
        
        public TacletApp getTacletApp() {
            return app;
        }
                
        public String toString() {
            final LogicPrinter printer = 
                new LogicPrinter(new ProgramPrinter(), 
                        notInfo, true);
            printer.setServices(services);
            printer.setInstantiation(app.instantiations());
            printer.printSequent(checkTaclet(app.taclet()));
            return printer.toString();
        }
    }

}
