// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//


package de.uka.ilkd.key.java;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.HashMap;

import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.expression.literal.NullLiteral;
import de.uka.ilkd.key.java.recoderext.ImplicitFieldAdder;
import de.uka.ilkd.key.java.reference.ExecutionContext;
import de.uka.ilkd.key.java.reference.TypeRef;
import de.uka.ilkd.key.java.reference.TypeReference;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.ldt.LDT;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.ObjectSort;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.util.Debug;

/**
 * an instance serves as representation of a Java model underlying a DL
 * formula. This class provides calls to access the elements of the Java
 * model using the KeY datastructures only. Implementation specific
 * details like the use of Recoder is hidden in the field of type 
 * {@link KeYProgModelInfo}. This class can be extended to provide further 
 * services.
 */
public class JavaInfo {


    private Services services;
    private KeYProgModelInfo  kpmi;

    /**
     * the type of null
     */
    private KeYJavaType nullType = null;

    /**
     * as accessed very often caches: 
     * KeYJavaType of
     *    java.lang.Object, java.lang.Clonable, java.io.Serializable
     * in </em>in this order</em>
     */
    private KeYJavaType[] commonTypes = new KeYJavaType[3];

    //some caches for the getKeYJavaType methods.
    private HashMap sort2KJTCache = null;
    private HashMap type2KJTCache = null;
    private HashMap name2KJTCache = null;
    private HashMap sName2KJTCache = null;
    private int nameCachedSize = 0;
    private int sNameCachedSize = 0;
    private int sortCachedSize = 0;
    
    /**
     * The default execution context is for the case of program statements on
     * the top level. It is equivalent to a static class belonging the default 
     * package. This should only be used when using KeY in academic mode, if
     * the verification conditions are generated they "must" start with a 
     * {@link de.uka.ilkd.key.java.statement.MethodBodyStatement} or a
     * {@link de.uka.ilkd.key.java.statement.MethodFrame}, which contains a 
     * valid execution context.
     */
    private ExecutionContext defaultExecutionContext;
                

    /**
     * a term with the constant 'null'
     */
    private Term nullConst=null;
    private boolean commonTypesCacheValid;
    /** the name of the class used as default execution context */
    static final String DEFAULT_EXECUTION_CONTEXT_CLASS = "<Default>";

	
    /**
     * creates a new JavaInfo object by giving a KeYProgModelInfo to access
     * the Recoder SourceInfo and using the given {@link Services} object.
     */
    JavaInfo(KeYProgModelInfo kpmi, Services s) {
	this.kpmi 	= kpmi;
	services	= s;	  	
    }

    private JavaInfo(JavaInfo proto, Services s) {
	this ( proto.getKeYProgModelInfo().copy(), s );
	nullType=proto.getNullType();
	nullConst=proto.getNullConst();
    }

    /**
     * returns the underlying KeYProgModelInfo providing access to the
     * Recoder structures.
     */
    public KeYProgModelInfo getKeYProgModelInfo(){
        return kpmi;
    }

    void setKeYProgModelInfo(KeYProgModelInfo kpmi){
        this.kpmi = kpmi;
    }

    /**
     * convenience method that returns the Recoder-to-KeY mapping underlying
     * the KeYProgModelInfo of this JavaInfo
     */
    public KeYRecoderMapping rec2key() {
	return getKeYProgModelInfo().rec2key();
    }

    /**
     * copies this JavaInfo and uses the given Services object as the
     * Services object of the copied JavaInfo
     * @param serv
     * @return
     */
    public JavaInfo copy(Services serv) {
 	return new JavaInfo(this, serv);
    }

    /** 
     * Don't make this method public, use <code>Services</code>
     * instead
     *
     * returns the TypeConverter to translate program parts to their
     * logic equivalent
     */
    private TypeConverter getTypeConverter() {
        return services.getTypeConverter();
    }

    /**
     * returns the services associated with this JavaInfo
     */
    public Services getServices(){
	return services;
    }

    //------------------- common services ----------------------


    /**
     * returns the full name of a given {@link
     * de.uka.ilkd.key.java.abstraction.KeYJavaType}. 
     * @param t the KeYJavaType including the package prefix
     * @return the full name
     */
    public String getFullName(KeYJavaType t) {
        return kpmi.getFullName(t);
    }

    /**
     * creates a new TypeReference for the given KeYJavaType
     */
    public TypeReference createTypeReference(KeYJavaType kjt) {
        return new TypeRef(kjt);
    }

    public void resetCaches(){
	sort2KJTCache = null;
	type2KJTCache = null;
	name2KJTCache = null;
	sName2KJTCache = null;
	nameCachedSize = 0;
	sNameCachedSize = 0;
	sortCachedSize = 0;
    }

    /**
     * looks up the fully qualifying name given by a String 
     * in the list of all available
     * KeYJavaTypes in the Java model
     * @param fullName the String
     * @return the KeYJavaType with the name of the String
     */
    public KeYJavaType getTypeByName(String fullName) {
	if(name2KJTCache == null || kpmi.rec2key().size() > nameCachedSize){
	    nameCachedSize = kpmi.rec2key().size();
	    name2KJTCache = new HashMap();
	    Iterator it = (kpmi.allElements()).iterator();
	    while (it.hasNext()) {
		Object o = it.next();
		if (o != null && o instanceof KeYJavaType){
		    name2KJTCache.put(getFullName((KeYJavaType)o), o);
		}
	    }
	}
	return (KeYJavaType) name2KJTCache.get(fullName);
    }


    // WATCHOUT: Woj: shouldn't this also have some sort of caching?
    /**
     * looks up the short name given by a String 
     * in the list of all available
     * KeYJavaTypes in the Java model and returns a list of
     * all possible types matching the name
     * @param name the String
     * @return the ListOfKeYJavaType with the name of the String
     */
    public ListOfKeYJavaType getTypeByShortName(String name) {
      ListOfKeYJavaType result = SLListOfKeYJavaType.EMPTY_LIST;
	Iterator it=(kpmi.allElements()).iterator();
	while (it.hasNext()) {
	    Object o=it.next();
	    if (o instanceof KeYJavaType) {
		if (new String(((KeYJavaType)o).getName()).equals(name)) {
		  result = result.append((KeYJavaType)o);
		}
	    }
	}
	return result;
    }

    /**
     * checks if name refers to a package
     * @param name a String with the name to be checked
     * @return true iff name refers to a package
     */
    public boolean isPackage(String name) {
	return kpmi.isPackage(name);
    }

    /**
     * looks up a KeYJavaType with given name. If the name is a fully
     * qualifying name with package prefix an element with this full name is
     * taken. If the name does not contain a full package prefix some
     * KeYJavaType with this short name is taken.
     * @param className the name to look for (either full or short)
     * @return a class matching the name
     */
    public KeYJavaType getTypeByClassName(String className) {
	KeYJavaType result = getTypeByName(className);
	if(result == null){
	    String shortName;
	    int dotpos = className.lastIndexOf(".");
	    shortName = className.substring(dotpos+1);
	    if(sName2KJTCache == null){
		sName2KJTCache = new HashMap();
		sNameCachedSize = kpmi.rec2key().size();
		Iterator it = (kpmi.allElements()).iterator();
		while (it.hasNext()) {
		    Object o = it.next();
		    if (o != null && o instanceof KeYJavaType){
			KeYJavaType t = (KeYJavaType)o;
			String name=getFullName(t);
			int pos=name.lastIndexOf(".");
			sName2KJTCache.put(name.substring(pos+1), o);
		    }
		}
	    }
	    result = (KeYJavaType) sName2KJTCache.get(shortName);
	}
	if(result != null){
	    Debug.out("javaInfo: type found (className, type):",
		      className, result);
	}else{
	    //this is for the case that the cache has been established to early
	    //(i.e. when not all types were known) 
	    if(kpmi.rec2key().size() > sNameCachedSize){
		sName2KJTCache = null;
		return getTypeByClassName(className);
	    }
	    Debug.out("javaInfo: type not found. Looked for:",className);
	    throw new RuntimeException("TYPE NOT FOUND: "+className);
	}
	return result;
    }


    /**
     * returns a type declaration with the full name of the given String fullName
     */
    public TypeDeclaration getTypeDeclaration(String fullName) {
        return (TypeDeclaration)getTypeByName(fullName).getJavaType();
    }

    
    /**
     * returns all known KeYJavaTypes of the current
     * program type model
     * @return all known KeYJavaTypes of the current
     * program type model
     */
    public Set getAllKeYJavaTypes() {
	Set result = new HashSet();
	Iterator it=(kpmi.allElements()).iterator();
        while (it.hasNext()) {
	    Object o=it.next();     
	    if (o instanceof KeYJavaType) {
		result.add(o);
	    }
	}
	return result;
    }

    /**
     * returns a primitive KeYJavaType matching the given typename making use
     * of the LDTs of the current type converter in the services.
     */
    public KeYJavaType getPrimitiveKeYJavaType(String typename) {
        LDT[] models = getTypeConverter().getModels();
        for (int i = 0; i<models.length ; i++) {
            IteratorOfType it = models[i].javaType().iterator();
            while (it.hasNext()) {
                Type type = it.next();
                if (type.getFullName().equals(typename)) {
                    return models[i].getKeYJavaType(type);
                }
            }
        }
        return null;
    }

    /**
     * returns a KeYJavaType (either primitive of object type) having the
     * full name of the given String fullName 
     */
    public KeYJavaType getKeYJavaType(String fullName) {
        KeYJavaType result = getPrimitiveKeYJavaType(fullName);
        return (result == null ?
            (KeYJavaType)getTypeByName(fullName) :
            result);
    }


    /**
     * this is an alias for getTypeByClassName
     */
    public KeYJavaType getKeYJavaTypeByClassName(String className) {
        return getTypeByClassName(className);
    }


    /**
     * returns true iff the given subType KeYJavaType is a sub type of the
     * given KeYJavaType superType.
     */
    public boolean isSubtype(KeYJavaType subType, KeYJavaType superType) {
        return kpmi.isSubtype(subType, superType);
    }

    /** 
     * returns a KeYJavaType having the given sort
     */
     public KeYJavaType getKeYJavaType(Sort sort) {
	 if(sort2KJTCache == null || kpmi.rec2key().size() > sortCachedSize){
	     sortCachedSize = kpmi.rec2key().size();
	     sort2KJTCache = new HashMap();
	     Iterator it = (kpmi.allElements()).iterator();
	     while (it.hasNext()) {
		 Object o = it.next();
		 if (o != null && o instanceof KeYJavaType){
		     sort2KJTCache.put(((KeYJavaType)o).getSort(), o);
		 }
	     }
	 }
	 KeYJavaType result = (KeYJavaType) sort2KJTCache.get(sort);
	 return result;
     }

    /** returns the KeYJavaType of the expression if it can be
     * determined else null is returned
     */
    public KeYJavaType getPrimitiveKeYJavaType(Expression e) {
        return getTypeConverter().getKeYJavaType(e);
    }

    /**
     * returns all ObjectSorts of the underlying Java model
     * @return a namespace containing the object sorts
     */
    public Namespace getObjectSorts() {
        Iterator it = kpmi.allObjectSorts().iterator();
        Namespace ns = new Namespace();
        while (it.hasNext()) {       
            ns.add((Named)it.next());            
        }
        return ns;
    }    

    /**
     * returns the KeYJavaType belonging to the given Type t
     */
    public KeYJavaType getKeYJavaType(Type t) {
        if (t instanceof PrimitiveType) {
	    return getTypeConverter().getKeYJavaType(t);
        } else {
	    if(type2KJTCache == null){
		type2KJTCache = new HashMap();
		Iterator it = (kpmi.allElements()).iterator();
		while (it.hasNext()) {
		    Object o = it.next();
		    if (o != null && o instanceof KeYJavaType){
			type2KJTCache.put(((KeYJavaType)o).getJavaType(), o);
		    }
		}
	    }
	    return (KeYJavaType) type2KJTCache.get(t);
	}
    }

    /**
     * returns all methods from the given Type
     */
    public ListOfMethod getAllMethods(KeYJavaType kjt) {
        return kpmi.getAllMethods(kjt);
    }

    /**
     * returns all locally declared methods from the given Type
     */
    public ListOfMethod getMethods(KeYJavaType kjt) {
        return kpmi.getMethods(kjt);
    }

    /**
     * returns all methods from the given Type as ProgramMethods
     */
    public ListOfProgramMethod getAllProgramMethods(KeYJavaType kjt) {
        return kpmi.getAllProgramMethods(kjt);
    }

    /**
     * returns the program methods defined in the given KeYJavaType with name
     * m and the list of types as signature of the method 
     * @param classType the KeYJavaType of the class where to look for the 
     *  method 
     * @param methodName the name of the method
     * @param signature a ListOfType with the arguments types 
     * @param context the KeYJavaType of the class context from <em>where</em>
     *  the method is called 
     * @return a matching program method 
     */
    public ProgramMethod getProgramMethod(KeYJavaType classType, 
            String methodName,
            ListOfType signature,
            KeYJavaType context) {
        return kpmi.getProgramMethod(classType, methodName, signature, context);
    }

    /**
     * returns the program methods defined in the given KeYJavaType with name
     * m and the list of types as signature of the method
     * @param classType the KeYJavaType of the class where to look for the 
     *  method 
     * @param methodName the name of the method
     * @param signature a ListOfKeYJavaType with the arguments types 
     * @param context the KeYJavaType of the class context from <em>where</em>
     *  the method is called 
     * @return a matching program method 
     */
    public ProgramMethod getProgramMethod(KeYJavaType ct, String methodName,
					  ListOfKeYJavaType signature,
					  KeYJavaType context) {
        return kpmi.getProgramMethod(ct, methodName, signature, context);
    }


    public Term getProgramMethodTerm(Term prefix, String methodName, 
				     Term[] args,
				     String className) {
	ListOfKeYJavaType sig = SLListOfKeYJavaType.EMPTY_LIST;
	Term[] subs = new Term[args.length+1];
	subs[0] = prefix;
	for (int i=1; i<subs.length; i++) {
              Term t = args[i-1];
              sig=sig.append(getServices().getTypeConverter()
			     .getKeYJavaType(t));
              subs[i] = t;
	}
	KeYJavaType clType = getKeYJavaTypeByClassName(className);
	ProgramMethod pm   = getProgramMethod(clType, methodName, sig, clType);
	if (pm==null) {
	    throw new IllegalArgumentException("Program method "+methodName
					       +" in "+className+" not found.");
	}
	if (pm.isStatic()) {
	    Term[] newSubs = new Term[subs.length-1];
	    System.arraycopy(subs,1,newSubs,0,newSubs.length);
	    subs=newSubs;
	}
	if (pm.getKeYJavaType()==null) {
	    throw new IllegalArgumentException("Program method "+methodName
					       +" in "+className+" must have"
					       +" a non-void type.");
	}
	return TermFactory.DEFAULT.createFunctionTerm(pm, subs);
    }



    /**
     * retrieves the direct extended superclass for the given class
     * @param type the KeYJavaType of the type whose superclass
     * has to be determined 
     * @return KeYJavaType of the extended supertype
     */
    public KeYJavaType getSuperclass(KeYJavaType type) {
	KeYJavaType result = null;
	ClassType javaType = (ClassType) type.getJavaType();

	if (javaType.isInterface()) {
	    return null;
	}

	ListOfKeYJavaType localSupertypes = javaType.getSupertypes();
	IteratorOfKeYJavaType it = localSupertypes.iterator();
	while (result == null && it.hasNext()) {
	    KeYJavaType keYType = it.next();
	    if (!((ClassType)keYType.getJavaType()).isInterface()) {
		result = keYType;
	    }
	}

	if (result == null) {
	    result = getJavaLangObject();
	} 

        return result;	
    }
    

    /**
     * returns the program method defined in the KeYJavaType of the program
     * variable clv, with the name m, and the KeYJavaTypes of the given array
     * of program variables as signatures.
     * @param classType the KeYJavaType of the class where to look for the 
     *  method 
     * @param methodName the name of the method
     * @param args an array of ProgramVariables as the arguments of the 
     * method 
     * @param context the KeYJavaType of the class context from <em>where</em>
     *  the method is called 
     * @return a matching program method
     */
    public ProgramMethod getProgramMethod
	(KeYJavaType classType, String methodName, ProgramVariable[] args,
	        KeYJavaType context){
        ListOfType types = SLListOfType.EMPTY_LIST;
        for (int i = args.length - 1; i>=0; i--) {
            types = types.prepend(args[i].getKeYJavaType());
        }
        return getProgramMethod(classType, methodName, types, context);
    }

    /** gets an array of expression and returns a list of types */
    private ListOfKeYJavaType getKeYJavaTypes(ArrayOfExpression args) {
	ListOfKeYJavaType result = SLListOfKeYJavaType.EMPTY_LIST; 
	if (args != null) {
	    for (int i = args.size()-1; i >= 0 ; i--) {
		Expression argument = args.getExpression(i);
		result = result.prepend
		    (getTypeConverter().getKeYJavaType(argument));
	    }
	}
	return result;
    }


    /**
     * retrieves the signature according to the given expressions
     * @param arguments ArrayOfExpression of which we try to construct a
     * signature 
     * @return the signature 
     */
    public ListOfKeYJavaType createSignature(ArrayOfExpression arguments) {
	return getKeYJavaTypes(arguments);
// 	return arguments == null ? SLListOfKeYJavaType.EMPTY_LIST : 
// 	    
    }

    /**
     * retrieves all attributes locally declared in the given class 
     * (including the implicit ones)
     * @param classDecl the ClassDeclaration where to look for the attributes
     * @return all attributes declared in the given class 
     * (including the implicit ones)
     */
    public ListOfField getAllFields(ClassDeclaration classDecl) {
	ListOfField fields = SLListOfField.EMPTY_LIST;
	ArrayOfMemberDeclaration members = classDecl.getMembers();
	for (int i = members.size()-1; i>=0; i--) {
	    MemberDeclaration member = members.getMemberDeclaration(i);
	    if (member instanceof FieldDeclaration) {
		ArrayOfFieldSpecification specs = 
		    ((FieldDeclaration)member).getFieldSpecifications();
		for (int j = specs.size()-1; j>=0; j--) {
		    fields = fields.prepend(specs.getFieldSpecification(j));
		}
	    }
	}
	return fields;
    }

    /**
     * retrieves all implicit attributes of the given class 
     * @param classDecl the ClassDeclaration where to look for the implicit
     * attributes 
     * @return all implicit attributes of the given class
     */
    public ListOfField getImplicitFields(ClassDeclaration classDecl) {
	ListOfField fields = SLListOfField.EMPTY_LIST;
	ArrayOfMemberDeclaration members = classDecl.getMembers();
	for (int i = members.size()-1; i>=0; i--) {
	    if (members.getMemberDeclaration(i) instanceof FieldDeclaration) {
		ArrayOfFieldSpecification specs = 
		    ((FieldDeclaration)members.getMemberDeclaration(i)).
		    getFieldSpecifications();
		for (int j = specs.size()-1; j>=0; j--) {
		    if (specs.getFieldSpecification(j) instanceof ImplicitFieldSpecification) {
			fields = fields.prepend(specs.getFieldSpecification(j));
		    }
		}
	    }
	}
	return fields;
    }
    

    //----------------- parsing services --------------------------

    /**
     * reads a Java block given as a string java as it was in the given
     * TypeDeclaration asIn.
     */
     public JavaBlock readJavaBlock(String java, TypeDeclaration asIn) {
        ClassDeclaration cd=null;
        if (asIn instanceof ClassDeclaration) {
            cd = (ClassDeclaration)asIn;
        } else {
            Debug.out("Reading Java Block from an InterfaceDeclaration:"
                      +" Not yet implemented.");
        }
        NamespaceSet nss = services.getNamespaces().copy();
        nss.startProtocol();
        final JavaBlock block = kpmi.readBlock(java, cd, nss);
        // if we are here everything is fine nad wen can add the
        // changes (may be new array types)       
        services.getNamespaces().addProtocolled(nss);        
        return block;
    }

    /**
     * reads a Java block given as a String
     */
    public JavaBlock readJavaBlock(String java) {
        NamespaceSet nss = services.getNamespaces().copy();
        nss.startProtocol();
        final JavaBlock block = kpmi.readJavaBlock(java, nss);                
        // if we are here everything is fine nad wen can add the
        // changes (may be new array types)       
        services.getNamespaces().addProtocolled(nss);
        return block;
    }

    /**
     * reads a Java statement not necessarily a block
     */
    public ProgramElement readJava(String java) {
        return ((StatementBlock)readJavaBlock("{"+java+"}")
		.program()).getChildAt(0);
    }

    /**
     * returns a term representing 'null' in logic
     */
    public Term getNullTerm() {
        return getTypeConverter().convertToLogicElement(NullLiteral.NULL);
    }


    /**
     * retrieves a field with the given name out of the list
     * @param name a String with the name of the field to be looked for
     * @param fields the ListOfField where we have to look for the field
     * @return the program variable of the given name or null if not
     * found
     */
    private final ProgramVariable find(String programName, 
                                       ListOfField fields) {
	IteratorOfField it = fields.iterator();
	while (it.hasNext()) {
	    Field field = it.next();
	    if (programName.equals(field.getProgramName())) {
		return (ProgramVariable)
		    ((FieldSpecification)field).getProgramVariable();
	    }
	}
	return null;
    }

    /**
     * extracts all fields out of fielddeclaration
     * @param field the FieldDeclaration of which the field
     * specifications have to be extracted
     * @return a ListOfField the includes all field specifications found
     * int the field declaration of the given list
     */
    private final ListOfField filterField(FieldDeclaration field) {
	ListOfField result = SLListOfField.EMPTY_LIST;
	ArrayOfFieldSpecification spec = field.getFieldSpecifications();
	for (int i = spec.size()-1; i>=0; i--) {
	    result = result.prepend(spec.getFieldSpecification(i));
	} 
	return result;
    }

    /**
     * extracts all field specifications out of the given
     * list. Therefore it descends into field declarations.
     * @param list the ArrayOfMemberDeclaration with the members of a
     * type declaration
     * @return a ListOfField the includes all field specifications found
     * int the field declaration of the given list
     */
    private ListOfField filterField(ArrayOfMemberDeclaration list) {
	ListOfField result = SLListOfField.EMPTY_LIST;
	for (int i = list.size()-1; i >= 0; i--) {
	    MemberDeclaration pe = list.getMemberDeclaration(i);
	    if (pe instanceof FieldDeclaration) {
		result = result.append
		    (filterField((FieldDeclaration)pe));
	    }
	}
	return result;
    }

    /**
     * returns the programvariable for the specified attribute. The attribute 
     * has to be fully qualified, i.e. <tt>declarationType::attributeName</tt> 
     * @param fullyQualifiedName the String with the fully qualified attribute 
     * name
     * @returns an attribute program variable of the given name
     * @throws IllegalArgumentException if the given name is not fully 
     * qualified  
     */
    public ProgramVariable getAttribute(String fullyQualifiedName) {
        final int idx = fullyQualifiedName.indexOf("::");
        
        if (idx == -1) {
            throw new IllegalArgumentException(fullyQualifiedName + 
                    " is not a fully qualified attribute name");
        }
        
        return getAttribute(fullyQualifiedName.substring(idx+2), 
			    fullyQualifiedName.substring(0, idx));
    }


    /**
     * returns the programvariable for the specified attribute declared in 
     * the specified class  
     * @param programName the String with the name of the attribute     
     * @param qualifiedClassName the String with the full (inclusive package) qualified
     * class name
     * @returns the attribute program variable of the given name 
     * @throws IllegalArgumentException if the qualified class name is empty or
     * null
     */
    public ProgramVariable getAttribute(String programName, String qualifiedClassName) {
	if (qualifiedClassName == null | qualifiedClassName.length() == 0) {
            throw new IllegalArgumentException("Missing qualified classname");
        }
        return getAttribute(programName, 
			    getKeYJavaTypeByClassName(qualifiedClassName));
    }

    
    /**
     * returns the program variable representing the attribute of the given 
     * name declared locally in class <tt>classType</tt>
     * @returns the attribute of the given name declared in <tt>classType</tt> 
     */
    public ProgramVariable getAttribute(String name,
					KeYJavaType classType) {       
	if (classType.getJavaType() instanceof ArrayDeclaration) {
	    if (name.length() > 0 && name.charAt(0)=='<' &&
		!(ImplicitFieldAdder.IMPLICIT_NEXT.equals(name) ||
                  ImplicitFieldAdder.IMPLICT_ARRAY_TRA_INITIALIZED.
                  equals(name))) {
		return getAttribute(name, getJavaLangObject());
	    } else {
		final ProgramVariable res = find(name, filterField
		        (((ArrayDeclaration)classType.
		                getJavaType()).getMembers()));
		if (res!=null) {
		    return res;
		}
	    }
	} else {
	    ListOfField list = kpmi.getAllFieldsLocallyDeclaredIn(classType);
	    IteratorOfField it = list.iterator();
	    while(it.hasNext()) {
		Field f = it.next();
		if(f!=null && (f.getName().equals(name) || f.getProgramName().equals(name))){
		    return (ProgramVariable)((VariableSpecification)f).
                    getProgramVariable();
		}
	    }
	}
	return null;
    }

    /**
     * returns an attribute named <tt>attributeName</tt> declared locally 
     * in object type <tt>s</tt>    
     */
    public ProgramVariable getAttribute(String attributeName, ObjectSort s) {
        return getAttribute(attributeName, getKeYJavaType(s));
    }

    /**     
     * returns a list of all attributes with the given program name 
     * declared in one of <tt>type</tt>'s sub- or supertype including 
     * its own attributes
     * <strong>Attention:</strong>
     *   The type must not denote the null type    
     * </ol>
     * 
     * @param programName the String with name of the attribute as declared 
     * in a program 
     * @param type the KeYJavaType specifying the part of the hierarchy 
     * where to look for
     * @return list of found attributes with name <tt>programName</tt> 
     */
    public ListOfProgramVariable getAllAttributes(String programName, 
                                                  KeYJavaType type) {
                                

        ListOfProgramVariable result = 
            SLListOfProgramVariable.EMPTY_LIST;

	if (!(type.getSort() instanceof ObjectSort)) {
	    return result;
	}
               
        if (type.getJavaType() instanceof ArrayType) {
            ProgramVariable var = find(programName, filterField
                        (((ArrayDeclaration)type.getJavaType())
                         .getMembers()));                                          
            if (var != null) { result = result.prepend(var); }
            var = getAttribute(programName, getJavaLangObject());
            if (var != null) { result = result.prepend(var); }
            return result;
        }
                
        
        ListOfKeYJavaType hierarchie = kpmi.getAllSubtypes(type);                
        Debug.assertTrue(!hierarchie.contains(type));
        hierarchie = hierarchie.prepend(kpmi.getAllSupertypes(type));        
        Debug.assertTrue(hierarchie.head() == type);
        
        
        final IteratorOfKeYJavaType it = hierarchie.iterator();
        while (it.hasNext()) {
	    KeYJavaType st = it.next();
	    if(st != null){
		final ProgramVariable var = 
		    getAttribute(programName, st);
		if (var != null) {
		    result = result.prepend(var);
		}            
	    }
        }
        
        return result;        
    }
    
    
    private void fillCommonTypesCache() {
        if (commonTypesCacheValid) return;
        final String[] fullNames = {"java.lang.Object", 
                "java.lang.Cloneable", "java.lang.Serializable"};
        
        for (int i = 0; i<fullNames.length; i++) {
            commonTypes[i] = getKeYJavaTypeByClassName(fullNames[i]);            
        }
        commonTypesCacheValid = true;
    }

    /**
     * returns the KeYJavaType for class <tt>java.lang.Object</tt>
     */
    public KeYJavaType getJavaLangObject() {
        if (commonTypes[0] == null) {
            commonTypes[0] = getKeYJavaTypeByClassName("java.lang.Object");
        }
        return commonTypes[0];
    }


    /**
     * returns the KeYJavaType for class java.lang.Clonable
     */
    public KeYJavaType getJavaLangCloneable() {
        if (commonTypes[1] == null) {
            commonTypes[1] = getKeYJavaTypeByClassName("java.lang.Cloneable");
        }
        return commonTypes[1];
    }

    /**
     * returns the KeYJavaType for class <tt>java.io.Serializable</tt>
     */
    public KeYJavaType getJavaIoSerializable() {
        if (commonTypes[2] == null) {
            commonTypes[2] = getKeYJavaTypeByClassName("java.io.Serializable");
        }
        return commonTypes[2];
    }


    
    /**
     * returns the KeYJavaType for class java.lang.Object
     */
    public Sort getJavaLangObjectAsSort() {
	return getJavaLangObject().getSort();	
    }

    /**
     * returns the KeYJavaType for class java.lang.Cloneable
     */
    public Sort getJavaLangCloneableAsSort() {   
        return getJavaLangCloneable().getSort();        
    }

    /**
     * returns the KeYJavaType for class java.io.Serializable
     */
    public Sort getJavaIoSerializableAsSort() {   
        return getJavaIoSerializable().getSort();        
    }
    
    /**
     * tests if sort represents java.lang.Object, java.lang.Cloneable or 
     * java.io.Serializable
     */
    public boolean isAJavaCommonSort(Sort sort) {
        if (!commonTypesCacheValid) { 
            fillCommonTypesCache();
        }
        for (int i = 0; i<commonTypes.length; i++) {
            if (commonTypes[i].getSort() == sort) {
                return true;
            }
        }        
        return false;
    }

    /**
     * returns the KeYJavaType  representing the type of 'null' 
     */
    public KeYJavaType getNullType() {
	if (nullType==null) {
	    nullType = getKeYJavaTypeByClassName("null");
	    Debug.assertTrue(nullType!=null
			 , "we should already have it in the map");
	}
	return nullType;
    }

    
    /**
     * returns the default execution context. This is equiavlent to executing the program
     * in a static method of a class placed in the default package 
     * @return the default execution context
     */
    public ExecutionContext getDefaultExecutionContext() {
        if (defaultExecutionContext == null) {                                   
            // ensure that default classes are available
            if (!kpmi.rec2key().parsedSpecial()) {
                readJava("{}");                
            }
            final KeYJavaType kjt = 
                getKeYJavaTypeByClassName(DEFAULT_EXECUTION_CONTEXT_CLASS);                     
            defaultExecutionContext = 
                new ExecutionContext(new TypeRef(kjt), null);
        }
        return defaultExecutionContext;
    }
    
    
    /**
     * returns a term representing the null constant in logic
     */
    public Term getNullConst() {
	if (nullConst==null) {
	    nullConst=TermFactory.DEFAULT.createFunctionTerm(Op.NULL);
	}
	return nullConst;
    }
    
    /**
     * returns all proper subtypes of a given type
     * @param type the KeYJavaType whose subtypes are returned
     * @return list of all subtypes
     */
    public ListOfKeYJavaType getAllSubtypes(KeYJavaType type) {
        return kpmi.getAllSubtypes(type);
    }
    
    /**
     * returns all supertypes of a given type
     * @param type the KeYJavaType whose supertypes are returned
     * @return list of all supertypes
     */
    public ListOfKeYJavaType getAllSupertypes(KeYJavaType type) {
        return kpmi.getAllSupertypes(type);
    }

    /**
     * looks up for a field of the given program name
     * visible <em>in</em> the specified class type belonging to the type 
     * or one of its supertypes 
     * 
     * @param programName the String containing the name of the 
     * field to be looked up. The name is in short notation, 
     * i.e. not fully qualified
     * @param classType the KeYJavaType of the class used as context
     * @return the field of the given name  
     */
    public ProgramVariable lookupVisibleAttribute(String programName, 
                                                  KeYJavaType classType) {                             
        return find(programName, kpmi.getAllVisibleFields(classType));
    }        
}
