// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2004 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
package de.uka.ilkd.key.java.recoderext;

import java.util.HashMap;

import org.apache.log4j.Logger;

import recoder.CrossReferenceServiceConfiguration;
import recoder.java.CompilationUnit;
import recoder.java.Identifier;
import recoder.java.StatementBlock;
import recoder.java.declaration.*;
import recoder.java.declaration.modifier.Private;
import recoder.java.declaration.modifier.Static;
import recoder.java.expression.literal.BooleanLiteral;
import recoder.java.expression.operator.CopyAssignment;
import recoder.java.reference.FieldReference;
import recoder.kit.ProblemReport;
import recoder.list.*;
import de.uka.ilkd.key.util.Debug;

/**
 * Each class is prepared before it is initialised. The preparation of
 * a class consists of pre-initialising the class fields with their
 * default values. This class creates the implicit method
 * <code>&lt;clprepare&gt;</code> responsible for the class
 * preparation.
 */
public class ClassPreparationMethodBuilder 
    extends RecoderModelTransformer {

	static Logger logger = Logger.getLogger(ClassPreparationMethodBuilder.class.getName());
	
    public static final String 
	CLASS_PREPARE_IDENTIFIER = "<clprepare>";

    /** maps a class to its static NON CONSTANT fields */
    private HashMap class2staticFields;

    /**  
     * Creates an instance of the class preparation method model
     * transformer. Information about the current recoder model can be
     * accessed via the given service configuration. The implicit
     * preparation method is created and added for all classes,
     * which are declared in one of the given compilation units. 
     * @param services the CrossReferenceServiceConfiguration with the
     * information about the recoder model
     * @param units the CompilationUnitMutableList with the classes to
     * be transformed
     */
    public ClassPreparationMethodBuilder
	(CrossReferenceServiceConfiguration services, 
	 CompilationUnitMutableList units) {	
	super(services, units);
	class2staticFields = new HashMap(10*units.size());
    }

    /** 
     * returns true if the given fieldspecification denotes a constant
     * field. A constant field is declared as final and static and
     * initialised with a time constant, which is not prepared or
     * initialised here.  ATTENTION: this is a derivation from the JLS
     * but the obtained behaviour is equivalent as we only consider
     * completely compiled programs and not partial compilations. The
     * reason for preparation and initialisation of comnpile time
     * constant fields is due to binary compatibility reasons.
     */
    private boolean isConstantField(FieldSpecification spec) {
    	if (logger.isDebugEnabled()) {
    		logger.debug("now at " + spec.getFullName());
    	}
    	boolean result = spec.isStatic() && spec.isFinal();
    	if (!result) {
    		return result;
    	}
    	recoder.service.ConstantEvaluator ce = services.getConstantEvaluator(); 
    	result = ce.isCompileTimeConstant(spec.getInitializer()); 
    	return result;
    }



    /**
     * retrieves all static non constant fields and returns a list of
     * copy assignment pre-initialising them with their default values
     *
     * some special settings for implicit fields are performed here as well
     * @param typeDeclaration the ClassDeclaration whose fields have to be prepared 
     * @return the list of copy assignments 
     */
    private StatementList prepareFields(TypeDeclaration typeDeclaration) {

	StatementMutableList result = new StatementArrayList(20);

	FieldSpecificationList fields = typeDeclaration.getFieldsInScope();

	for (int i = 0; i < fields.size(); i++) {
	    FieldSpecification spec = fields.getFieldSpecification(i);
	    if (spec.isStatic() && !isConstantField(spec)) {
		Identifier ident = spec.getIdentifier();
		if (ident instanceof ImplicitIdentifier) {
		    String identText = ident.getText();
		    if (ImplicitFieldAdder.
			IMPLICIT_NEXT_TO_CREATE.equals(identText)) {
			result.add(new CopyAssignment
			    (new PassiveExpression
			     (new FieldReference
			      ((ImplicitIdentifier)ident.deepClone())),
			     new PassiveExpression
			       (new FieldReference(new ImplicitIdentifier
						   (ImplicitFieldAdder.IMPLICIT_FIRST)))));
		    } else if (!ImplicitFieldAdder.IMPLICIT_FIRST.equals(identText)) {
			result.add(new CopyAssignment
				   (new PassiveExpression
				    (new FieldReference
				     ((ImplicitIdentifier)ident.deepClone())), 
				    getDefaultValue(spec.getType())));
		    }
		} else {
		   result.add(new CopyAssignment
			(new PassiveExpression
			 (new FieldReference
			  ((Identifier)ident.deepClone())), 
			 getDefaultValue(spec.getType())));
		}
	    }
	}

	result.add
	    (new CopyAssignment
	     (new PassiveExpression(new FieldReference
				    (new ImplicitIdentifier
				     (ImplicitFieldAdder.IMPLICIT_CLASS_PREPARED))),
	      new BooleanLiteral(true)));

	return result;
    }
    
    public ProblemReport analyze() {
	for (int unit = 0; unit<units.size(); unit++) {
	    CompilationUnit cu = units.getCompilationUnit(unit);
	    int typeCount = cu.getTypeDeclarationCount();
	    
	    for (int i = 0; i < typeCount; i++) {
		if (cu.getTypeDeclarationAt(i) instanceof ClassDeclaration)
		    { 
			ClassDeclaration cd = (ClassDeclaration)
			    cu.getTypeDeclarationAt(i);
			if (cd.getTypeDeclarationCount()>0) {
			    Debug.out
				("clPrepBuilder: Inner Class detected. " + 
				 "Reject building class initialisation methods.");
			}
									
			// collect initializers for transformation phase
			class2staticFields.put(cd, prepareFields(cd));
		    }
	    }
	}
	setProblemReport(NO_PROBLEM);
	return NO_PROBLEM;
    }

    /** 
     * creates the static method <code>&lt;clprepare&gt;</code> for the
     * given type declaration
     * @param td the TypeDeclaration to which the new created method
     * will be attached
     * @return the created class preparation method
     */
    private MethodDeclaration createPrepareMethod(TypeDeclaration td) {
	ModifierMutableList modifiers = new ModifierArrayList(2);
	modifiers.add(new Static());
	modifiers.add(new Private());
	return new MethodDeclaration(modifiers, 
				     null,  // return type is void
				     new ImplicitIdentifier
				     (CLASS_PREPARE_IDENTIFIER),
				     new ParameterDeclarationArrayList(0), 
				     null, // no throws
				     new StatementBlock
				     ((StatementMutableList)
				      class2staticFields.get(td)));
    }


    /**
     * entry method for the constructor normalform builder
     * @param td the TypeDeclaration 
     */
    protected void makeExplicit(TypeDeclaration td) {
	attach(createPrepareMethod(td), td, 0);
	// for debug 
	// 	    java.io.StringWriter sw = new java.io.StringWriter();
	// 	    //	services.getProgramFactory().getPrettyPrinter(sw).visitMethodDeclaration(nf);
	// 	    services.getProgramFactory().getPrettyPrinter(sw).visitTypeDeclaration((TypeDeclaration)td);
	// 	    System.out.println(sw.toString());
	// 	    try { sw.close(); } catch (Exception e) {}		
    }


}
