// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.logic.op;

import java.lang.ref.WeakReference;
import java.util.WeakHashMap;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.reference.ArrayReference;
import de.uka.ilkd.key.java.reference.ReferencePrefix;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.sort.ArraySort;
import de.uka.ilkd.key.logic.sort.GenericSort;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.rule.MatchConditions;
import de.uka.ilkd.key.rule.inst.GenericSortInstantiations;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.ExtList;

public class ArrayOp extends AccessOp {

    /** 
     * The pool of already created array operators. Ensures
     * 'singleton' property of array operators.
     */
    private static final WeakHashMap operatorPool = new WeakHashMap(50);
    
    /** Sort of the array this operator addresses */
    protected final Sort sort;

    /**
     * returns and if neccessary creates the shadowed array access operator for 
     * the given array sort
     * 
     * @param arraySort the Sort to which this operator is applicable
     *  must be of kind {@link ArraySort} or {@link ProgramSVSort}
     * @return the array access operator for the given array sort
     */
    public static ArrayOp getArrayOp(Sort arraySort) {
        Debug.assertTrue(arraySort instanceof ArraySort ||
                arraySort instanceof ProgramSVSort || 
                arraySort instanceof GenericSort ||
                arraySort == AbstractMetaOperator.METASORT, 
                arraySort + " is no allowed array sort.");
        WeakReference arrayAccessOpReference = (WeakReference) operatorPool
                .get(arraySort);
        if (arrayAccessOpReference == null
                || arrayAccessOpReference.get() == null) {
            // wrapping attributeOp in a weak reference is necessary as
            // it contains a strong reference to its key
            arrayAccessOpReference = 
                   new WeakReference(new ArrayOp(arraySort));
            operatorPool.put(arraySort, arrayAccessOpReference);
        }
        return (ArrayOp) arrayAccessOpReference.get();
    } 
    
    protected ArrayOp(Name name, Sort arraySort) {
        super(name);
        this.sort = arraySort;
    }
    
    /**
     * creates an ArrayOp
     */
    private ArrayOp(Sort arraySort){
     this(new Name("arrayAccess(" + arraySort +")"), arraySort);
    }
    
    /**
     * determines the sort of the {@link Term} if it would be created using this
     * Operator as top level operator and the given terms as sub terms. The
     * assumption that the constructed term would be allowed is not checked.
     * @param term an array of Term containing the subterms of a (potential)
     * term with this operator as top level operator
     * @return sort of the term with this operator as top level operator of the
     * given substerms
     */
    public Sort sort(Term[] term) {	
        if (sort instanceof ProgramSVSort || 
                sort instanceof GenericSort || 
                sort == AbstractMetaOperator.METASORT){ 
            return ProgramSVSort.NONSIMPLEEXPRESSION;
	}  
        return ((ArraySort)sort).elementSort();
    }
    
    /** checks whether the top level structure of the given @link Term
     * is syntactically valid, given the assumption that the top level
     * operator of the term is the same as this Operator. The
     * assumption that the top level operator and the term are equal
     * is NOT checked.
     * @return true iff the top level structure of
     * the @link Term is valid.
     */
    public boolean validTopLevel(Term term) {
        Sort s = term.sub(0).sort();
        if (sort instanceof ArraySort && s instanceof ArraySort) {
            // missing check if term.sub(1) has integer sort 
            return s.extendsTrans(sort); 
        } else if (sort instanceof ProgramSVSort || 
                sort instanceof GenericSort  || 
                sort == AbstractMetaOperator.METASORT ||
                term.sub(0).op() == Op.EVERYTHING) {
	    return true;
	}
	return false;	
    }

    /** 
     * the arity of an array access operator is <code>2</code> where the
     * first sub term denotes the array object itself and the second one
     * the accessed index
     * @return arity of the Operator as int 
     */
    public int arity(){
        return 2;
    }

    /**
     * @return true if the operator is a shadowed 
     */
    public boolean isShadowed() {
	return false;
    }
    
    /**
     * @return the array sort on which this operator is applied
     */
    public Sort arraySort() {
        return sort;
    }
    
    /**
     * @return the sort of this operator
     */
    public Sort sort() {
        return ((ArraySort)sort).elementSort();
    }
    
    public de.uka.ilkd.key.java.Expression convertToProgram(Term t, 
							    ExtList el) {
	ReferencePrefix accessPath = (ReferencePrefix) el.get(0);
	el.remove(accessPath);
	return new ArrayReference(el, accessPath);
    }

    /**
     * tests if the location <code>loc</code> may be an alias of this array 
     * location
     * @param loc the Location to check
     * @return true iff <code>loc</code> may be an alias of the current (this) 
     *  location
     * @see de.uka.ilkd.key.logic.op.Location#mayBeAliasedBy
     *                      (de.uka.ilkd.key.logic.op.Location)
     */
    public boolean mayBeAliasedBy(Location loc) {        
        if ( !( loc instanceof ArrayOp )
             || ( loc instanceof ShadowedOperator ) )
            return false; 
        
        final ArrayOp aOp = (ArrayOp)loc;
        return arraySortAliased ( aOp.arraySort (), arraySort () )
	    || sort instanceof ProgramSVSort
	    || aOp.sort instanceof ProgramSVSort;
    }
    
    protected boolean arraySortAliased(Sort as1, Sort as2) {
        return as1.extendsTrans(as2) || as2.extendsTrans(as1);
    }
    
    /** 
     * implements the default operator matching rule which means 
     * that the compared object have to be equal otherwise
     * matching fails
     */
    public MatchConditions match(Operator op, MatchConditions mc,
            Services services) {
        
	if (arraySort() instanceof ProgramSVSort && 
	    op.getClass() == this.getClass() ) {
	    return mc;
	} else if (arraySort() instanceof GenericSort &&
            op.getClass() == this.getClass()) {
	    final GenericSortInstantiations genSortInst = 
		mc.getInstantiations().getGenericSortInstantiations();
	    if (genSortInst.isInstantiated((GenericSort)arraySort())) {
		final Sort s =  genSortInst.getRealSort(arraySort(), services);
		Debug.assertTrue(s instanceof ArraySort);
		return getArrayOp(s).match(op, mc, services);
	    } else {
	        // generic sort not yet instantiated (will be later)
	        return mc;
	    }
	}
        return super.match(op, mc, services);

    }

    public String toString() {
        return "[]("+arraySort()+")";
    }
}

