// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof;

import java.io.IOException;

import de.uka.ilkd.key.collection.ListOfString;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.StatementContainer;
import de.uka.ilkd.key.java.annotation.Annotation;
import de.uka.ilkd.key.java.annotation.LoopInvariantAnnotation;
import de.uka.ilkd.key.java.statement.LoopStatement;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.IUpdateOperator;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.pp.LogicPrinter;
import de.uka.ilkd.key.pp.NotationInfo;
import de.uka.ilkd.key.pp.ProgramPrinter;
import de.uka.ilkd.key.rule.PosTacletApp;
import de.uka.ilkd.key.rule.TacletApp;

public class LoopInvariantProposer implements InstantiationProposer {

    /**
     * An instance of LoopInvariantProposer
     */
    public static final LoopInvariantProposer DEFAULT = new LoopInvariantProposer();

    
    /**
     * Returns a proposal for the instantiation of <code>var</code> 
     * iff <code>app</code> is a TacletApp for a loop invariant taclet
     * and <code>var</code> is the SchemaVariable representing the invariant 
     * and the loop on which the taclet matches contains a loop invariant
     * annotation. Otherwise null is returned.
     */
    public String getProposal(TacletApp app, 
    			      SchemaVariable var, 
			      Services services, 
			      Node undoAnchor,
			      ListOfString previousProposals){
	
        final Object inst = tryToInstantiate(app, var);
	final LogicPrinter lp = new LogicPrinter(new ProgramPrinter(null), 
						 NotationInfo.createInstance());
	String proposal;
	try {
	    if (inst instanceof Term){
		lp.printTerm((Term) inst);
		proposal = lp.toString();
	    }  else if (inst instanceof ListOfTerm){
		lp.printTerm((ListOfTerm) inst);
		proposal = lp.toString();
	    } else if (var.name().toString().equals("#modifies")) {
		proposal = "{*}";
	    } else { 
		proposal = null;
	    }
	} catch (IOException e){
	    proposal = null;
	}
	return proposal;
    }

    /**
     * Returns an instantiation of <code>var</code> 
     * iff <code>app</code> is a TacletApp for a loop invariant taclet
     * and <code>var</code> is the SchemaVariable representing the invariant 
     * and the loop on which the taclet matches contains a loop invariant
     * annotation. Otherwise null is returned.
     * Depending if the var looked for is a list schemavariable or a normal sv
     * a list of terms or a term is returned
     */
    public Object tryToInstantiate(TacletApp app, SchemaVariable var){	
        Object inst = null;
        if (app instanceof PosTacletApp && 
	    app.taclet().name().toString().indexOf("_inv") != -1){
	    final PosInOccurrence pos = ((PosTacletApp) app).posInOccurrence();
	    final LoopInvariantAnnotation firstLoopInvAnnot = 
                getFirstLoopInvariantAnnotation(pos.subTerm());
            final String varName = var.name().toString();
            if (firstLoopInvAnnot == null) {
                return null;
            } else if (varName.equals("inv")) {
		inst = firstLoopInvAnnot.invariant();
	    } else if(varName.equals("#modifies")) {
		inst = convertToListOfTerm(firstLoopInvAnnot.assignable());
	    } else if(varName.equals("variant")) {
		inst = firstLoopInvAnnot.variant();;
	    } else if(varName.equals("#old")) {
		inst = convertToListOfTerm(firstLoopInvAnnot.olds());
	    }
	}
	return inst;
    }

    
    private LoopInvariantAnnotation getFirstLoopInvariantAnnotation(Term t) {
        final Annotation[] a = getFirstLoopStatement(t).getAnnotations();
        
        for(int i = 0; i<a.length; i++){
            if(a[i] instanceof LoopInvariantAnnotation){
                return (LoopInvariantAnnotation) a[i];
            }
        }
        return null;        
    }
    
    private ListOfTerm convertToListOfTerm(ArrayOfTerm array){
	ListOfTerm result =  SLListOfTerm.EMPTY_LIST;
	for (int i = array.size() - 1; i>=0; i--) {
	    result = result.prepend(array.getTerm(i));    	
	}
	return result;
    }

    private LoopStatement getFirstLoopStatement(Term t){
	while(t.op() instanceof IUpdateOperator){
	    t = ( (IUpdateOperator)t.op () ).target ( t );
	}
	return getLoopHelp(t.javaBlock().program());
    }

    private LoopStatement getLoopHelp(ProgramElement pe){
	if(pe instanceof LoopStatement){
	    return (LoopStatement) pe;
	}
	if(pe instanceof StatementContainer){
	    return getLoopHelp(((StatementContainer) pe).getStatementAt(0));
	}
	//shouldn't happen.
	return null;
    }
}
