// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.proof.init;

import java.util.HashMap;
import java.util.Iterator;

import de.uka.ilkd.key.gui.LDTSettings;
import de.uka.ilkd.key.gui.ProofSettings;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.TypeConverter;
import de.uka.ilkd.key.java.abstraction.ListOfType;
import de.uka.ilkd.key.java.abstraction.PrimitiveType;
import de.uka.ilkd.key.java.abstraction.SLListOfType;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.ldt.*;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.sort.GenericSort;
import de.uka.ilkd.key.logic.sort.PrimitiveSort;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.SortDefiningSymbols;
import de.uka.ilkd.key.ocl.ocldltranslator.ProverFacadeImpl;
import de.uka.ilkd.key.proof.BuiltInRuleIndex;
import de.uka.ilkd.key.proof.TacletIndex;
import de.uka.ilkd.key.proof.mgt.ProofEnvironment;
import de.uka.ilkd.key.proof.ProofAggregate;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.KeYExceptionHandler;

/**
 * an instance of this class describes the initial configuration of the prover.
 * This includes sorts, functions, heuristics, and variables namespaces,
 * information on the underlying java model, and a set of rules.
 */
public class InitConfig {

    private NamespaceSet nss=new NamespaceSet();
    private LDT[] ldtModels = null;
    /** the default Choice of this initial configuration. */
    private Choice defaultChoice;
    private SetOfTaclet taclets = SetAsListOfTaclet.EMPTY_SET;
    private HashMap libraries;
    public static ArithIntLDTIgnoringOF aILDTIOF = new ArithIntLDTIgnoringOF();
    public static IntegerLDT iLDT                = new IntegerLDT();
    public static ArithIntLDT aILDT              = new ArithIntLDT();
    public static BooleanLDT booleanLDT          = new BooleanLDT();
    
    /**
     * maps categories to their default choice (both represented as Strings),
     * which is used if no other choice is specified in the problemfile
     */
    private static HashMap category2DefaultChoice = new HashMap();

    /**
     * maps taclets to their TacletBuilders. This information is needed when
     * a taclet contains GoalTemplates annotated with taclet-options because
     * in this case a new taclet has to be created containing only those
     * GoalTemplates whose options are activated and those who don't belong
     * to any specific option.
     */
    protected HashMap taclet2Builder;

    private boolean noStandardRules = true;

    /**
     * Set of the rule options activated for the current proof. The rule options
     * ({@link Choice}s) allow to use different ruleset modelling or skipping
     * certain features (e.g. nullpointer checks when resolving references)
     */
    private SetOfChoice activatedChoices=SetAsListOfChoice.EMPTY_SET;

    /**
     * the service class allowing to access information about the underlying
     * program model
     */
    private Services serv;

    /**
     * list of built-in-rules like update simplification, external
     * decision procedures etc.
     */
    private ListOfBuiltInRule builtInRules = SLListOfBuiltInRule.EMPTY_LIST;

    /**
     * the proof environment this init config belongs to
     */
    private ProofEnvironment env;
    
    /** HashMap for quick lookups taclet name->taclet */
    private HashMapFromNameToNamed quickTacletMap;
    
    

    public InitConfig(KeYExceptionHandler keh) {
        serv = new Services(keh);
        defaultChoice = new Choice(new Name("Default"));
        nss.setFunctions(defaultChoice.funcNS());
        defaultChoice.funcNS().add(Op.NULL);
        sortNS().add(Sort.NULL);
        category2DefaultChoice = ProofSettings.DEFAULT_SETTINGS.
            getChoiceSettings().getDefaultChoices();
	taclet2Builder = new HashMap();
    }

    public InitConfig(){
        this(null);
    }

    /**
     * sets the NamespaceSet of this initial configuration
     * @param nss the NamespaceSet aggregating available sorts, functions,
     * program variables etc.
     */
    public void setNamespaces(NamespaceSet nss) {
        this.nss = nss;
    }

    /**
     * sets the information of the underlying program model of this
     * initial configuration
     * @param serv the Services providing access to the program model
     */
    public void setServices(Services serv) {
        this.serv = serv;
    }

    public void setDefaultChoice(Choice defaultChoice){
        this.defaultChoice = defaultChoice;
    }

    public Choice getDefaultChoice(){
        return defaultChoice;
    }

    public void setNoStandardRules(boolean no){
        noStandardRules = no;
    }

    public boolean noStandardRules(){
        return noStandardRules;
    }

    /**
     * returns the Services of this initial configuration providing access
     * to the used program model
     * @return the Services of this initial configuration
     */
    public Services getServices() {
        return serv;
    }

    /**
     * {@link Taclet}s are constructed using {@link TacletBuilder}s this map
     * contains the pair of a taclet and its builder which is important as
     * goals of a taclet may depend of the selected choices. Instead of
     * creating all possible combinations in advance this is done by demand
     * @return the map from a taclet to its builder
     */
    public HashMap getTaclet2Builder(){
        return taclet2Builder;
    }

    /** returns the HashMap that maps categories to their default choices */
    public HashMap getCategory2DefaultChoice(){
        return category2DefaultChoice;
    }

    /** sets the HashMap that maps categories to their default choices */
    public void setCategory2DefaultChoice(HashMap category2DefaultChoice){
        InitConfig.category2DefaultChoice = category2DefaultChoice;
    }

    public void setTaclet2Builder(HashMap taclet2Builder){
        this.taclet2Builder = taclet2Builder;
    }

    /**
     * adds entries to the HashMap that maps categories to their
     * default choices.
     * Only entries of <code>init</init> with keys not already contained in
     * category2DefaultChoice are added.
     */
    public void addCategory2DefaultChoices(HashMap init){
        Iterator it = init.keySet().iterator();
        boolean changed=false;
        while(it.hasNext()){
            Object key = it.next();
            if(!category2DefaultChoice.containsKey(key)){
                changed=true;
                category2DefaultChoice.put(key, init.get(key));
            }
        }
        if(changed){
            ProofSettings.DEFAULT_SETTINGS.getChoiceSettings().
                setDefaultChoices((HashMap)category2DefaultChoice.clone());
        }
    }


    /**
     * sets the set of activated choices of this initial configuration.
     * For categories without a specified choice the default choice contained
     * in category2DefaultChoice is added.
     */
    public void setActivatedChoices(SetOfChoice activatedChoices) {
        ProofSettings.DEFAULT_SETTINGS.
	    getChoiceSettings().updateChoices(choiceNS(), false);
        category2DefaultChoice = 
	    ProofSettings.DEFAULT_SETTINGS.getChoiceSettings().
	    getDefaultChoices();

        final IteratorOfChoice it = activatedChoices.iterator();
        HashMap c2DC = (HashMap)category2DefaultChoice.clone();
        Choice c;
        while(it.hasNext()){
            c = it.next();
            c2DC.remove(c.category());
        }
        
        final Iterator itDef = c2DC.values().iterator();
        while(itDef.hasNext()){
            String s = (String) itDef.next();
            c = (Choice) choiceNS().lookup(new Name(s));
            if(c!=null){
                activatedChoices = activatedChoices.add(c);
            }
        }
        this.activatedChoices = activatedChoices;

        // invalidate quick taclet cache
        invalidateCache();
    }

    /** 
     * invalidates the active taclets cache if the active choices have changed. 
     */
    private void invalidateCache() {
        if (quickTacletMap != null) {
            quickTacletMap = null;
        }
    }

    /** Returns the choices which are currently active.
     * For getting the active choices for a specific proof,
     * <code>getChoices</code> in <code>de.uka.ilkd.key.proof.Proof
     * </code> has to be used.
     */
    public SetOfChoice getActivatedChoices(){
        return activatedChoices;
    }

    public HashMap getLibraries(){
        if (libraries ==null){
            return ProofSettings.DEFAULT_SETTINGS.getLibrariesSettings()
                .getLibraries();
        }
        else return libraries;
    }
    
    public void setLibraries(HashMap libs){
        libraries=libs;
    }

    /**
     * sets the list of built in rules of this initial configuration
     */
    public void setBuiltInRules(ListOfBuiltInRule builtIn) {
        this.builtInRules = builtIn;
    }


    /** returns the namespaces of this initial configuration
     */
    public NamespaceSet namespaces() {
        nss.setFunctions(defaultChoice.funcNS());
        return nss;
    }


    /**
     * returns the information on the Java model of this initial configuration
     * @see #getJavaInfo()
     */
    public JavaInfo javaInfo() {
        return serv.getJavaInfo();
    }

    /**
     * Gets the information of the underlying Java model of this
     * initial configuration.
     * @see #javaInfo()
     * @see de.uka.ilkd.key.gui.KeYMediator#getJavaInfo()
     */
    public JavaInfo getJavaInfo() {
        return javaInfo();
    }

    public void setTaclets(SetOfTaclet taclets){
        this.taclets = taclets;
    }
    
    private HashMapFromNameToNamed getQuickTacletMap() {
        if (quickTacletMap == null) {
            quickTacletMap = new HashMapFromNameToNamed();
            IteratorOfTaclet it = activatedTaclets().iterator();
            while (it.hasNext())  {
                Taclet t = it.next();
                quickTacletMap.put(t.name(), t);
            }
        }
        return quickTacletMap;
    }
    
    public Taclet lookupActiveTaclet(Name name) {
        return (Taclet) getQuickTacletMap().get(name);
    }

    public SetOfTaclet getTaclets(){
        return taclets;
    }

    /**
     * returns the activated taclets of this initial configuration
     */
    public SetOfTaclet activatedTaclets() {
        IteratorOfTaclet it = taclets.iterator();
        SetOfTaclet result = SetAsListOfTaclet.EMPTY_SET;
        Taclet t;
        TacletBuilder b;
        while(it.hasNext()){
            t = it.next();
            b = (TacletBuilder) taclet2Builder.get(t);
            if(t.getChoices().subset(activatedChoices)){
                if(b!=null && b.getGoal2Choices()!=null){
                    t = b.getTacletWithoutInactiveGoalTemplates(
                            activatedChoices);
                }
                if(t!=null){
                    result = result.add(t);
                }
            }
        }
        return result;
    }

    public LDT[] ldtModels(){
        return ldtModels();
    }


    private ListOfType getIntTypes(){
        ListOfType type = SLListOfType.EMPTY_LIST;
        type = type.prepend(PrimitiveType.JAVA_INT);
        type = type.prepend(PrimitiveType.JAVA_LONG);
        type = type.prepend(PrimitiveType.JAVA_CHAR);
        type = type.prepend(PrimitiveType.JAVA_BYTE);
        type = type.prepend(PrimitiveType.JAVA_SHORT);
        return type;
    }

    private void initNumberLDT(NumberLDT nLDT, Namespace func_ns){
        Sort sort=null;
        ListOfType type = getIntTypes();
        if(sortNS().lookup(nLDT.getSortName())==null){
            sort = new PrimitiveSort(nLDT.getSortName());
            sortNS().add(sort);
        }else{
            sort = (Sort)sortNS().lookup(nLDT.getSortName());
        }
        nLDT.init(func_ns, sort, type);
        getServices().getTypeConverter().init(nLDT, booleanLDT);
    }

    public ArithIntLDT initArithIntLDT() {
        Choice c = (Choice) choiceNS().lookup(new Name("intRules:arithmeticSemanticsCheckingOF"));
        initNumberLDT(aILDT, c.funcNS().extended(defaultChoice.
                                                 funcNS().allElements()));
        Debug.assertTrue( aILDT!=null,"aILDT==null");
        setUpSorts();
        return aILDT;
    }

    public ArithIntLDTIgnoringOF initArithIntLDTIgnoringOF() {
        Choice c = (Choice) choiceNS().lookup(new Name(
                                      "intRules:arithmeticSemanticsIgnoringOF"));
        initNumberLDT(aILDTIOF, c.funcNS().extended(defaultChoice.
                                                    funcNS().allElements()));
        setUpSorts();
        return aILDTIOF;
    }

    public NumberLDT initIntegerLDT() {
        Choice c = (Choice) choiceNS().lookup(new Name("intRules:javaSemantics"));
        initNumberLDT(iLDT, c.funcNS().extended(defaultChoice.
                                                funcNS().allElements()));
        setUpSorts();
        return iLDT;
    }

    public BooleanLDT initBooleanLDT(){
        ListOfType type = SLListOfType.EMPTY_LIST;
        type = type.prepend(PrimitiveType.JAVA_BOOLEAN);
        Sort sort = (Sort)sortNS().lookup(booleanLDT.getSortName());
        booleanLDT.init(defaultChoice.funcNS(), sort, type);
        return booleanLDT;
    }

    public void initLDTSettings(ProofSettings settings){
        LDTSettings ldtSettings = settings.getLDTSettings();
        IteratorOfChoice it = activatedChoices.iterator();
        Choice c=null;
        LDT integerLDT = null;
        while(it.hasNext()){
            c = it.next();
            if(c.category().equals("intRules")){
                if(c.name().equals(new Name("intRules:javaSemantics"))){
                    integerLDT = iLDT;
                }
                if(c.name().equals(new Name("intRules:arithmeticSemanticsCheckingOF"))){
                    integerLDT = aILDT;
                }
                if(c.name().equals(new Name((
                             "intRules:arithmeticSemanticsIgnoringOF")))){
                    integerLDT = aILDTIOF;
                }
                ldtSettings.setNaturalNumberLDT(integerLDT);
            }
        }
        getServices().getTypeConverter().init(integerLDT, booleanLDT);
        createLDTModels(new LDT[]{integerLDT, booleanLDT});
        setUpSorts();
        settings.getChoiceSettings().updateChoices(choiceNS(),false);
    }

    /** creates the language data type models */
    private void createLDTModels(LDT[] model) {
        TypeConverter converter = getServices().getTypeConverter();
        LDT intLDT     = null;
        LDT localBooleanLDT = null;
        Debug.out("probleminitializer: initialize LDT/Typeconverter");
        for (int i = 0; i < model.length; i++) {
            if(sortNS().lookup(model[i].getSortName())==null)
                sortNS().add(model[i].targetSort());
            add(funcNS(), model[i].functions());
            if (model[i] instanceof NumberLDT) {
                Debug.out("initialize: integerLDT");
                intLDT = model[i];
            } else if (model[i] instanceof BooleanLDT) {
                Debug.out("initialize: booleanLDT");
                localBooleanLDT = model[i];
            }
        }
        converter.init(intLDT, localBooleanLDT);
        ldtModels = model;
        Debug.out("probleminitializer: initialize TypeConverter end.");
    }

    /**
     * Delayed setup of symbols defined by sorts (e.g. functions for
     * collection sorts). This may not have been done for previously
     * defined sorts, as the integer sort was not available.
     */
    protected void setUpSorts () {
        NamespaceSet    nsset     = new NamespaceSet ();
        Namespace       func_ns = nsset.functions ();
        IteratorOfNamed it      =
            sortNS ().allElements ().iterator ();
        Sort            sort;

        while ( it.hasNext () ) {
            sort = (Sort)it.next ();
            if ( sort instanceof SortDefiningSymbols )
                ((SortDefiningSymbols)sort).createSymbols ( func_ns,
                                                            sortNS() );
        }

        add ( nsset, ModStrategy.MOD_ALL );
    }

    /**
     * adds all the object contained in the source namespace to the target
     * @param target the Namespace where to put the objects of
     * <code>source</code>
     * @param source the Namespace from where the objects are taken
     */
    private void add(Namespace target, Namespace source) {
        IteratorOfNamed it=source.elements().iterator();
        while (it.hasNext()) {
            target.add(it.next());
        }

    }

    /** returns the built-in rules of this initial configuration
     */
    public ListOfBuiltInRule builtInRules() {
        return builtInRules;
    }

    /** returns a newly created taclet index for the set of activated
     * taclets contained in this initial configuration
     */
    public TacletIndex createTacletIndex() {
        return new TacletIndex(activatedTaclets());
    }

    /**
     * returns a new created index for built in rules (at the moment immutable
     * list)
     */
    public BuiltInRuleIndex createBuiltInRuleIndex() {
        return new BuiltInRuleIndex(builtInRules);
    }

    /** returns the function namespace of this initial configuration
     */
    public Namespace funcNS() {
        return defaultChoice.funcNS();
    }

    /** returns the sort namespace of this initial configuration
     */
    public Namespace sortNS() {
        return nss.sorts();
    }

    /** returns the heuristics namespace of this initial configuration
     */
    public Namespace ruleSetNS() {
        return nss.ruleSets();
    }

    /** returns the variable namespace of this initial configuration
     */
    public Namespace varNS() {
        return nss.variables();
    }

    /** returns the program variable namespace of this initial configuration
     */
    public Namespace progVarNS() {
        return nss.programVariables();
    }

    /** returns the choice namespace of this initial configuration
     */
    public Namespace choiceNS() {
        return nss.choices();
    }

    public NamespaceSet createNamespacesForActivatedChoices(){
        Namespace func_ns = new Namespace();
        func_ns.add(defaultChoice.funcNS());
        IteratorOfChoice it = activatedChoices.iterator();
        while(it.hasNext()){
            func_ns.add(it.next().funcNS());
        }
        nss.setFunctions(func_ns);
        return nss;
    }

    public ProofSettings mergedProofSettings() {
        ProofSettings defaultSettings = ProofSettings.DEFAULT_SETTINGS;
        ProofAggregate someProof
            = ((ProofAggregate)getProofEnv().getProofs().iterator().next());
        if (someProof!=null) {
            return defaultSettings.setChoiceSettings(
                someProof.getFirstProof().getSettings().getChoiceSettings());
        } else {
            return defaultSettings;
        }
    }

    /** adds namespaces to the namespaces of this initial configuration
     */
    public void add(NamespaceSet ns, ModStrategy mod) {
        if (mod.modifyFunctions()) funcNS().add(ns.functions());
        addSorts ( ns, mod );
        if (mod.modifyVariables()) varNS().add(ns.variables());
        if (mod.modifyProgramVariables()) progVarNS().add(ns.programVariables());
        if (mod.modifyHeuristics()) ruleSetNS().add(ns.ruleSets());
        if (mod.modifyChoices()) choiceNS().add(ns.choices());
    }

    private void addSorts (NamespaceSet ns, ModStrategy mod) {
        final IteratorOfNamed sortsIt = ns.sorts ().elements ().iterator ();
        while ( sortsIt.hasNext () ) {
            final Named named = sortsIt.next ();
            if ( named instanceof GenericSort ) {
                if ( mod.modifyGenericSorts () ) sortNS ().add ( named );
            } else {
                if ( mod.modifySorts () ) sortNS ().add ( named );
            }
        }
    }

    /**
     * creates a facade for the OCL-DL-Translation using the information stored
     * in this initial configuration
     */
    public ProverFacadeImpl createFacadeForTransl() {
        nss.setFunctions(defaultChoice.funcNS());
        return new ProverFacadeImpl(getServices(), nss);
    }

    /** returns a copy of this initial configuration copying the namespaces,
     * the contained JavaInfo while using the immutable set of taclets in the
     * copy
     */
    public InitConfig copy() {
        InitConfig ic=new InitConfig(null);
        ic.setNamespaces(namespaces().copy());
        ic.setDefaultChoice(defaultChoice.copy());
        ic.setActivatedChoices(activatedChoices);
        if (libraries!=null)
            ic.setLibraries((HashMap)libraries.clone());
        ic.setCategory2DefaultChoice((HashMap) getCategory2DefaultChoice().
                                     clone());
        ic.setTaclet2Builder((HashMap) taclet2Builder.clone());
        ic.setTaclets(taclets);
        // we have to initialize the new services
        // with the former ldts
        ic.setServices(serv.copyPreservesLDTInformation());
        ic.setBuiltInRules(builtInRules());
        ic.setProofEnvironment(new ProofEnvironment(ic));
        return ic;
    }

    /**
     * returns the proof environment using this initial configuration
     * @return the ProofEnvironment using this configuration
     */
    public ProofEnvironment getProofEnv() {
        return env;
    }

    /**
     * sets the proof environment using this configuration
     * @param env the ProofEnvironment using this configuration
     */
    public void setProofEnvironment(ProofEnvironment env) {
        this.env=env;
    }

    /**
     * toString
     */
    public String toString() {
        return
            "Namespaces:" + nss +"\n" +
            "Services:" + serv +"\n"+
            "Taclets:" + getTaclets() +"\n"+
            "Built-In:" + builtInRules +"\n";
    }
}
