// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.proof.init;

import java.io.*;
import de.uka.ilkd.key.proof.ProblemLoader;
import de.uka.ilkd.key.gui.KeYFileChooser;
import de.uka.ilkd.key.gui.Main;
import de.uka.ilkd.key.util.ProgressMonitor;

import java.util.Stack;

/** represents an input from a .key user problem file producing
 *  environment as well as a proof obligation. In contrast to its
 *  super class this ProofOblInput has a fixed java path given,
 *  i.e. the java section in a .key file that is loaded with this
 *  ProofOblInput is ignored. Currently the class contains a helper
 *  method to determine a peer directory in which the .key files are
 *  stored when given a .java file.
 */
public class KeYUserJavaProblemFile extends KeYUserProblemFile {

    private File javaPath;

    public KeYUserJavaProblemFile(File f, File javaPath, 
				  ProgressMonitor monitor) {
	super(f.toString(), f, monitor);
	this.javaPath = javaPath;
    }

    public String getJavaPath() throws ProofInputException {
	return javaPath.toString();
    }

    /** returns an overriden version of {@link
     * de.uka.ilkd.key.proof.ProblemLoader}; instead of creating a
     * KeYUserProblemFile the returned ProblemLoader uses this
     * KeYUserJavaProblemFile. 
     */
    public ProblemLoader getProblemLoader(Main main) {
	return new ProblemLoader(getFile(), main) {
		protected KeYUserProblemFile createUserProblemFile(File file) {
		    return KeYUserJavaProblemFile.this;
		}
	    };
    }


    /** for given files specRoot and currentClass, this method returns
     * an appropriate file in the peer specification directory. For a
     * currentClass /a/b/c/d/C.java and a specRoot /a/b_spec a file
     * representing /a/b_spec/c/d/C.key is returned.
     */
    public static File getPeerSpecPath(File specRoot, File currentClass) {
 	specRoot = specRoot.getAbsoluteFile();
 	currentClass = currentClass.getAbsoluteFile();
 	File current = currentClass.getParentFile();	
	Stack stack = new Stack();
 	while (current!=null && !current.equals(specRoot.getParentFile())) {
	    stack.push(current.getName());
 	    current = current.getParentFile();
 	}
 	if (current!=null) {
	    current = specRoot;
	    stack.pop();
	    while (!stack.empty()) {
		current = new File(current, (String)stack.pop());
	    }
	    if (currentClass.getName().endsWith(".java")) {
		String className = currentClass.getName();
		current = new File(current,
				   className.substring(0, className.length()-5)
				   +".key");
	    }
	}
	return current;
    }


    /** @param args the arguments, the first containing the path to
     * the java project, the second containing the path to the file of
     * question, the third containing the path to the root of the
     * specifications (.key files)
     */
    public static void main(String args[]) {
	File jPath = new File(args[0]);
	File currentClass = new File(args[1]);
	File specRoot = new File(args[2]);
	KeYFileChooser kfc = Main.getFileChooser(null);
	kfc.selectFile(getPeerSpecPath(specRoot, currentClass));
	if (!kfc.showOpenDialog(Main.getInstance())) return;
	ProgressMonitor mainMonitor = Main.getInstance().getProgressMonitor();
	KeYUserJavaProblemFile f 
	    = new KeYUserJavaProblemFile(kfc.getSelectedFile(),
					 jPath, mainMonitor);
	f.getProblemLoader(Main.getInstance()).run();
    }
}
