// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.proof.init;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.*;

import de.uka.ilkd.key.collection.ListOfString;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.abstraction.IteratorOfType;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.abstraction.ListOfType;
import de.uka.ilkd.key.java.abstraction.SLListOfType;
import de.uka.ilkd.key.java.declaration.ArrayOfTypeDeclaration;
import de.uka.ilkd.key.java.declaration.ClassDeclaration;
import de.uka.ilkd.key.java.declaration.InterfaceDeclaration;
import de.uka.ilkd.key.java.declaration.TypeDeclaration;
import de.uka.ilkd.key.jml.IteratorOfJMLMethodSpec;
import de.uka.ilkd.key.jml.JMLMethodSpec;
import de.uka.ilkd.key.jml.SetOfJMLMethodSpec;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.parser.jml.JMLSpecBuilder;
import de.uka.ilkd.key.parser.jml.NotSupportedExpressionException;
import de.uka.ilkd.key.proof.CountingBufferedInputStream;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.ProofAggregate;
import de.uka.ilkd.key.proof.mgt.ContractUtil;
import de.uka.ilkd.key.proof.mgt.JMLMethodContract;
import de.uka.ilkd.key.proof.mgt.JavaModelMethod;
import de.uka.ilkd.key.proof.mgt.MethodContract;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.util.ProgressMonitor;

/** represents an input from a .key user problem file producing environment
 *  as well as a proof obligation.
 */
public class KeYUserProblemFile extends KeYFile implements ProofOblInput{

    private Term problemTerm = null;
    private String problemHeader = "";

    private String javaPath;
    private boolean javaPathAlreadyParsed=false;

    // if false only the specifications of Object and Datagroup are read.
    // The parsing of javacode and specifications of nonlibrary classes
    // is not affected by this flag.
    public static boolean parseLibSpecs = false;
    // for disabling the parsing of java classes and their 
    // jmlspecs when running tests
    private boolean parseJMLSpecs;


    /** creates a new representation of a KeYUserFile with the given name,
     * a rule source representing the physical source of the input, and
     * a graphical representation to call back in order to report the progress
     * while reading.
     */
    public KeYUserProblemFile(String name, File file, 
			      ProgressMonitor monitor) {
	this(name, file, monitor, true);
    }

    public KeYUserProblemFile(String name, File file, 
			      ProgressMonitor monitor, boolean parseJMLSpecs) {
	super(name, file, monitor);
	this.parseJMLSpecs = parseJMLSpecs;
    }



    public void readHelp(ModStrategy mod, boolean problemOnly) 
	throws ProofInputException {
	if(file == null) return;
	if (initConfig==null) {
	    throw new IllegalStateException("KeYFile: InitConfig not set.");
	}
	try {
	    CountingBufferedInputStream cinp = 
		new CountingBufferedInputStream
		    (getNewStream(),monitor,getNumberOfChars()/100);
            DeclPicker lexer = new DeclPicker(new KeYLexer(cinp,initConfig.getServices().getExceptionHandler()));

	    /*
	      two namespace sets which share all namespace except the
	      variable namespace 	     	      
	    */
	    NamespaceSet normal = initConfig.namespaces().copy();
	    NamespaceSet schema = 
		new NamespaceSet(normal.variables().copy(), 
				 normal.functions(),
				 normal.sorts(), 
				 normal.ruleSets(),
				 normal.choices(),
				 normal.programVariables());
	    Choice defaultChoice = initConfig.getDefaultChoice().copy();
	    ParserConfig normalConfig 
		= new ParserConfig(initConfig.getServices(),
				   normal, defaultChoice);

	    ParserConfig schemaConfig 
		= new ParserConfig(initConfig.getServices(), 
				   schema, defaultChoice);
	    problemParser = new KeYParser(ParserMode.PROBLEM, lexer, 
					      file.toString(), 
					      schemaConfig, 
					      normalConfig,
					      initConfig.getTaclet2Builder(),
					      initConfig.getTaclets(), initConfig.getActivatedChoices()); 
	    if (problemOnly) {
		problemTerm = problemParser.parseProblem();
	    } else {
		problemTerm = problemParser.problem();
	    }
            problemHeader = lexer.getText();
            if(problemHeader != null && 
	       problemHeader.lastIndexOf("\\problem") != -1){
		problemHeader = problemHeader.substring(
		    0, problemHeader.lastIndexOf("\\problem"));
	    }
	    initConfig.setTaclets(problemParser.getTaclets());
	    initConfig.add(normalConfig.namespaces(), mod);
	    initConfig.setDefaultChoice(normalConfig.getDefaultChoice()); 
	    initConfig.getProofEnv().addContracts(
                problemParser.getContracts(), problemHeader);
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
        } catch (FileNotFoundException fnfe) {
            throw new ProofInputException(fnfe);
        }
    }

    /** reads the whole .key file and modifies the initial configuration
     * assigned to this object according to the given modification strategy. 
     * Throws an exception if no initial configuration has been set yet.
     */
    public void read(ModStrategy mod) throws ProofInputException {
	readHelp(mod, false);
    }

    /** reads the the problem section of the .key file and modifies
     * the initial configuration assigned to this object according to
     * the given modification strategy.  Throws an exception if no
     * initial configuration has been set yet.
     */
    public void readProblem(ModStrategy mod) throws ProofInputException {
	readHelp(mod, true);
    }



    public ProofAggregate getPO() {
        return ProofAggregate.createProofAggregate(
            new Proof(name(), problemTerm, problemHeader,
                        initConfig.createTacletIndex(), 
                        initConfig.createBuiltInRuleIndex(),
                        initConfig.getServices(), settings), name());
    }

    /** returns the <code>java.io.file</code> file encapsulated by
     * the <code>KeYUserProblemFile</code>.
     */
    public File getFile(){
	return file.file();
    }
    
    /**
     * @return Returns the problemHeader.
     */
    protected String getProblemHeader () {
        return problemHeader;
    }

    /** returns true, that is the input asks the user which
     * environment he prefers if there are multiple possibilities
     */
    public boolean askUserForEnvironment() {
	return true;
    }

    public void readSpecs(){
	//	try {
	    Services serv = initConfig.getServices();
	    ProgramMethod meth;
	    Iterator it;
	    if(method2jmlspecs != null && !method2jmlspecs.isEmpty()){
		it = method2jmlspecs.keySet().iterator();
		while(it.hasNext()){
		    meth = (ProgramMethod) it.next();
		    if(method2jmlspecs.get(meth) != null){
			IteratorOfJMLMethodSpec sit = 
			    ((SetOfJMLMethodSpec) 
			     method2jmlspecs.get(meth)).
			    iterator();
			while(sit.hasNext()){
			    JMLMethodSpec jmlspec = sit.next();
			    createJMLMethodContract(meth, jmlspec);
			}
		    }
		}
	    }
	    if(parseJMLSpecs){
		Set kjts = serv.getJavaInfo().getAllKeYJavaTypes();
		it = kjts.iterator();
		while(it.hasNext()){
		    KeYJavaType kjt = (KeYJavaType) it.next();
		    if(!(kjt.getJavaType() instanceof InterfaceDeclaration
			 || kjt.getJavaType() instanceof ClassDeclaration)){
			continue;
		    }
		    ListOfProgramMethod ml = serv.getJavaInfo().getAllProgramMethods(kjt);
		    IteratorOfProgramMethod mit = ml.iterator();
		    IteratorOfJMLMethodSpec sit;
		    JMLMethodSpec jmlspec;
		    while(mit.hasNext()){
			meth = mit.next();
			SetOfJMLMethodSpec specs = 
			    serv.getImplementation2SpecMap().
			    getSpecsForMethod(meth);
			if(specs != null){
			    sit = specs.iterator();
			    while(sit.hasNext()){
				jmlspec = sit.next();
				createJMLMethodContract(meth, jmlspec);
			    }
			}
			sit = serv.getImplementation2SpecMap().
			    getInheritedMethodSpecs(meth, kjt).iterator();
			while(sit.hasNext()){
			    jmlspec = sit.next();
			    createJMLMethodContract(meth, jmlspec);
			}
		    }
		}
	    }
	    // library classes specifications
	    /*	    SpecExtractor spex = null;
	    String jpath = getJavaPath();
	    if (jpath==null) return;
	    spex = new OCLSpecExtractor(initConfig.getServices(),jpath);
	    initConfig.getProofEnv().addContracts(spex.getSpecsFromAllClasses());
	} catch (ProofInputException e) {
	    // do nothing
	    Debug.out("Exception thrown by class KeYFile at SpecExtraction");
	    }*/
    }

    private void createJMLMethodContract(Object meth, JMLMethodSpec jmlspec){
    	Modality[] allMod = new Modality[]{Op.DIA, Op.BOX};
    	String jp = null;
	if(!jmlspec.isValid()) return;
    	try{
	    jp = getJavaPath();
    	}catch(ProofInputException e){
	    e.printStackTrace();
    	}
    	for (int i=0; i<allMod.length; i++) {
	    String suffix = ContractUtil.getPOSuffix(allMod[i]);
	    KeYJavaType returnType;
	    MethodContract mct = null;
	    if(meth instanceof ProgramMethod){
		returnType = 
		    ((ProgramMethod) meth).getKeYJavaType();
		if(returnType == null){
		    returnType = initConfig.getServices().
			getJavaInfo().getKeYJavaType("void");
		}
		mct = new JMLMethodContract(new JavaModelMethod(
						(ProgramMethod) meth, jp), 
					    jmlspec, allMod[i], suffix);
	    }
	    initConfig.getProofEnv().addContract(mct);
    	}
    }
    
    public boolean equals(Object o){
	if(!(o instanceof KeYUserProblemFile)) return false;
	KeYUserProblemFile kf = (KeYUserProblemFile) o;
	if(file != null && kf.file != null){
	    return kf.file.file().getAbsolutePath().
		equals(file.file().getAbsolutePath());
	}
	if(file == null && kf.file == null){
	    try{
		return getJavaPath().equals(kf.getJavaPath());
	    }catch(ProofInputException e){
		return kf == this;
	    }
	}
	return false;
    }

    public int hashCode(){
	if(file == null){
	    try{
		return getJavaPath().hashCode();
	    }catch(ProofInputException e){
		return 0;
	    }
	}
	return file.file().getAbsolutePath().hashCode();
    }

    protected ListOfType getTypesWithJMLSpecs(CompilationUnit[] cUnits){
	ListOfType result = SLListOfType.EMPTY_LIST;
	JavaInfo ji = initConfig.getServices().getJavaInfo();
	if(parseLibSpecs){
	    Set kjts = ji.getAllKeYJavaTypes();
	    Iterator it = kjts.iterator();
	    while(it.hasNext()){
		KeYJavaType kjt = (KeYJavaType) it.next();
		if(kjt.getJavaType() instanceof InterfaceDeclaration
		   || kjt.getJavaType() instanceof ClassDeclaration){
		    result = result.append(kjt.getJavaType());
		}
	    }
	}else{
	    for(int i=0; i<cUnits.length; i++){
		ArrayOfTypeDeclaration tds = cUnits[i].getDeclarations();
		for(int j=0; j<tds.size(); j++){
		    if(tds.getTypeDeclaration(j) instanceof 
		       InterfaceDeclaration ||
		       tds.getTypeDeclaration(j) instanceof ClassDeclaration){
			result = result.append(tds.getTypeDeclaration(j));
		    }
		}
	    }
	    result = result.append(ji.getJavaLangObject().getJavaType());
	    result = result.append(ji.getKeYJavaType(
				       "org.jmlspecs.lang.JMLDataGroup").
				       getJavaType());
	    result = result.append(ji.getKeYJavaType(
				       "java.lang.Exception").
				       getJavaType());
	    result = result.append(ji.getKeYJavaType(
				       "java.lang.Throwable").
				       getJavaType());

	}
	return result;
    }

    protected void parseJMLSpecs(ListOfType types) 
	throws ProofInputException {
	method2jmlspecs = new HashMap();
	Set builders = new HashSet();
	IteratorOfType itt = types.iterator();
	NamespaceSet nss = initConfig.createNamespacesForActivatedChoices();
	while(itt.hasNext()){
	    TypeDeclaration t = (TypeDeclaration) itt.next();
	    JMLSpecBuilder builder = 
		new JMLSpecBuilder(
		    t,
		    initConfig.getServices(),
		    nss,
		    TermFactory.DEFAULT,
		    getJavaPath()
		    );
	    builders.add(builder);
	}
	Iterator it = builders.iterator();
	while(it.hasNext()){
	    JMLSpecBuilder b = (JMLSpecBuilder) it.next();
	    b.parseModelMethodDecls();
	}
	it = builders.iterator();
	while(it.hasNext()){
	    JMLSpecBuilder b = (JMLSpecBuilder) it.next();
	    b.parseModelFieldDecls();
	}
	it = builders.iterator();
	while(it.hasNext()){
	    JMLSpecBuilder b = (JMLSpecBuilder) it.next();
	    b.parseJMLClassSpec();
	}
	it = builders.iterator();
	while(it.hasNext()){
	    JMLSpecBuilder b = (JMLSpecBuilder) it.next();
	    b.parseJMLMethodSpecs();
	    method2jmlspecs.putAll(b.getModelMethod2Specs());
	}
	initConfig.getServices().getImplementation2SpecMap().
	    createLoopAnnotations();
    }

    private Vector getClasses(String f) throws ProofInputException  {
	File cfile = new File(f);
	Vector v=new Vector();
	if (cfile.isDirectory()) {
	    String[] list=cfile.list();
	    for (int i=0; i<list.length; i++) {
		String fullName = cfile.getPath()+File.separator+list[i];
		File n=new File(fullName);
		if (n.isDirectory()) {		    
		    v.addAll(getClasses(fullName));
		} else if (list[i].endsWith(".java")) {
		    v.add(fullName);	
		}
	    }
	    return v;
	} else {
	    throw new ProofInputException("Java model path "+f+" not found.");
	}
	
    }

    private String[] getClassesInPath(String f) throws ProofInputException {
	return (String[]) getClasses(f).toArray(new String[]{});
    }

    public String getJavaPath() throws ProofInputException {
	if (javaPathAlreadyParsed) return javaPath;	  
        try {
            problemParser = new KeYParser(ParserMode.PROBLEM,
                                          new KeYLexer(getNewStream(),null), 
                                          file.toString());
            
            problemParser.preferences(); // skip preferences
            
            ListOfString javaPaths = problemParser.javaSource(); 
	    javaPath = (javaPaths.size() == 0 ? null : javaPaths.head());
	    if(javaPaths.size() > 1)
	       Debug.fail("Don't know what to do with multiple Java paths.");
	    javaPathAlreadyParsed=true;
            if (javaPath==null) return null;
            File cfile = new File(javaPath);
	    if (!cfile.isAbsolute()) { // test relative pathname
	        File parent=file.file().getParentFile();
	        cfile = new File(parent,javaPath).
                    getCanonicalFile().getAbsoluteFile();
                javaPath = cfile.getAbsolutePath();
	    }
            if (!cfile.exists()) {
                throw new ProofInputException("Declared Java source " 
                                              + javaPath + " not found.");
            }
            return javaPath;
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
	} catch (IOException ioe) {
	    throw new ProofInputException(ioe);
        } catch(de.uka.ilkd.key.util.ExceptionHandlerException ehe){
	    ehe.printStackTrace();
            System.out.println(ehe.getCause().getMessage());
            throw new ProofInputException(ehe.getCause().getMessage());
	}
    }

    private boolean checkExceptions(ExtList l){
	Iterator it = l.iterator();
	while(it.hasNext()){
	    if(!(it.next() instanceof NotSupportedExpressionException)){
		return false;
	    }
	}
	return true;
    }

    public String readJava() throws ProofInputException {
	if (initConfig==null) {
	    throw new IllegalStateException("KeYFile: InitConfig not set.");
	}
	// no java is parsed at all just the java path
	javaPath = getJavaPath();
	if (javaPath != null) {
	    Recoder2KeY r2k = new Recoder2KeY(initConfig.getServices(),
					      initConfig.namespaces());
	    String[] cus = getClassesInPath(javaPath);
	    try { 
		if(parseJMLSpecs){
		    parseJMLSpecs(getTypesWithJMLSpecs(
				     r2k.readCompilationUnitsAsFiles(cus)));
		}else{
		    r2k.readCompilationUnitsAsFiles(cus);
		}
	    } catch (ConvertException ce) {
		throw new ProofInputException(ce);
	    } 
	    if(!checkExceptions(initConfig.getServices().getExceptionHandler().
				getExceptions())){
		throw new ProofInputException("Exceptions occured: "+
		  initConfig.getServices().getExceptionHandler().getExceptions());
	    }
	    //	    initConfig.getServices().getImplementation2SpecMap().setProgVarNS(
	    //		initConfig.progVarNS());
	}
        return javaPath;
    }
    
    
    public SetOfChoice readActivatedChoices() throws ProofInputException{
        if (settings==null) getPreferences();
        /* This is a weird contortion: we read a ProofSettings object, 
         * modify it partially based on the SetOfChoice (which the parser 
         * constructs from the \withOptions clause) and then convert it 
         * to a SetOfChoice again, to be used later on (in InitConfig?).
         * Oh, well... -- VK
         */
        settings.setChoiceSettings(settings.getChoiceSettings().
	  updateWith(super.readActivatedChoices()));
        
        //conversion of HashMap to SetOfChoice as other methods rely on this class
        SetOfChoice activatedChoices = SetAsListOfChoice.EMPTY_SET;
        Iterator choiceIter = settings.getChoiceSettings().getDefaultChoices().keySet().iterator();

        //there has to be a better way to accomplish this - hints welcome
        while (choiceIter.hasNext()) {
            Object key = choiceIter.next();
            String item = settings.getChoiceSettings().getDefaultChoices().get(key).toString();
            activatedChoices = activatedChoices.add(new Choice(new Name(item), key.toString()));
        }
        return activatedChoices; 
    }
}
