// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.java.Recoder2KeY;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.casetool.*;
import de.uka.ilkd.key.ocl.tudresden.TUDresdenParserAdapter;
import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.detailedocl.*;
import de.uka.ilkd.key.gui.ProofSettings;
import tudresden.ocl.*;
import de.uka.ilkd.key.proof.mgt.Contract;
import de.uka.ilkd.key.proof.mgt.Contractable;


/** represents an OCL proof obligation input to the prover for proving the
 * behavioral subtyping property.
 */
public class OCLBehavSubPO extends OCLProofOblInput {

    private ReprModelMethod methRepr;
    private ReprModelMethod parentMethRepr;
    private String preText;
    private String postText;
    private String preTextParent;
    private String postTextParent;
    private String contextString;
    private String parentContextString;
    private String xmifile;

    private Term prePO  = null;
    private Term parentPrePO  = null;
    private Term postPO = null;
    private Term parentPostPO = null;

    private ProofAggregate result;


    private boolean oclParseError = false;

    private ProgramVariable[] methodParameters;
    private ProgramVariable[] parentMethodParameters;


    /** creates a new OCL proof obligation input representing a behavioral
     * subtyping property proof obligation using
     * an XMI file name for reading the UML model from, a representation
     * of the model class the PO is invoked from and five strings, (1)
     * describing the contextString, (2) describing the pre condition text
     * of the parent, (3) describing the pre condition text of the parent class,
     * (4) describing the post condtion string of the subclass, and (5)
     * describing the post condition of the parent class.
     */
    public OCLBehavSubPO(ReprModelMethod aMethRepr, ReprModelMethod parentMethRepr, String xmifile,
                         String contextString, String parentContextString,
                         String preTextParent, String preText,
                         String postTextParent, String postText) {
        super("BehavioralSubtyping "+aMethRepr.getName());
        this.methRepr=aMethRepr;
        this.parentMethRepr=parentMethRepr;
        this.xmifile=xmifile;
        this.contextString=contextString;
        this.parentContextString=parentContextString;
        this.preTextParent=preTextParent;
        this.preText=preText;
        this.postTextParent=postTextParent;
        this.postText=postText;
    }


    /** reads the Java model for this PO -- do we need it?
     */
    public String readJava() {
        if (initConfig==null) {
            throw new IllegalStateException("InitConfig not set.");
        }
        Recoder2KeY r2k = new Recoder2KeY(initConfig.getServices(),
                                          initConfig.namespaces());
        String[] cus = methRepr.getContainingClass().getClassesInPackage();
        r2k.readCompilationUnitsAsFiles(cus);
        return methRepr.getContainingClass().getRootDirectory();
    }


    protected OclTree parseOCL(String preconditionHeader,
                               String xmiFile,
                               String precondition,
                               String packageName) {


        TUDresdenParserAdapter parserAdapter = new TUDresdenParserAdapter();

        OclTree parseResult = parserAdapter.doParse1
            (preconditionHeader, getModelManager(),
             xmifile, precondition, packageName);

        oclParseError = oclParseError && parserAdapter.getErrorOccurred();

        return parseResult;
    }


    /** reads this OCL proof obligation, thus invoking the OCL-DL-Translation
     */
    public void read(ModStrategy mod) throws ProofInputException {
        if (initConfig==null) {
            throw new IllegalStateException("InitConfig not set.");
        }
        final String preconditionHeader = contextString + " pre:";
        final String postconditionHeader = contextString + " post:";

        final String parentPreconditionHeader = parentContextString + " pre:";
        final String parentPostconditionHeader = parentContextString + " post:";


        final String packageName = methRepr.getContainingPackage();
        final String parentPackageName = parentMethRepr.getContainingPackage();

        oclParseError = false;

        final OclTree oclPrecondition =
            parseOCL(preconditionHeader, xmifile,
                     "context " + preconditionHeader +
                     FunctionalityOnModel.normalizeConstraint(preText),
                     packageName);

        final OclTree oclParentPrecondition =
            parseOCL(parentPreconditionHeader, xmifile,
                     "context " + parentPreconditionHeader +
                     FunctionalityOnModel.normalizeConstraint(preTextParent),
                     parentPackageName);

        final OclTree oclPostcondition =
            parseOCL(postconditionHeader, xmifile,
                     "context " + postconditionHeader +
                     FunctionalityOnModel.normalizeConstraint(postText),
                     packageName);

        final OclTree oclParentPostcondition =
            parseOCL(parentPostconditionHeader, xmifile,
                     "context " + parentPostconditionHeader +
                     FunctionalityOnModel.normalizeConstraint(postTextParent),
                     parentPackageName);

        final OCLMod[] mods = ProofSettings.DEFAULT_SETTINGS.getOCLRefinementSettings().getOCLMods();

        if ( !oclParseError ) {
            final OCLTreeToDLAggregator oclTranslator = getAggregator
                (initConfig.createFacadeForTransl(), mods, methRepr);
            prePO = oclTranslator.treeToTerm(oclPrecondition, false, true);
            postPO = oclTranslator.treeToTerm(oclPostcondition, false, false);
            methodParameters = oclTranslator.getMethodParameters();

            final OCLTreeToDLAggregator parentOCLTranslator = getAggregator
                (initConfig.createFacadeForTransl(), mods, parentMethRepr);
            parentPrePO = parentOCLTranslator.treeToTerm(oclParentPrecondition, false, true);
            parentPostPO = parentOCLTranslator.treeToTerm(oclParentPostcondition, false, false);
            parentMethodParameters = parentOCLTranslator.getMethodParameters();

        } else {
            throw new ProofInputException(
               "Pre-/Post Conditions are syntactically incorrect");
        }
    }

    public ProofAggregate getPO() {
        if (result!=null) return result;
        if (prePO == null) throw new IllegalStateException("No prePO term");
        if (parentPrePO == null) throw new IllegalStateException("No prePO term for parent method");
        if (postPO == null) throw new IllegalStateException("No postPO term");
        if (parentPostPO == null) throw new IllegalStateException("No postPO term for parent method");
        String s = getProofHeader(methRepr.getContainingClass()
                .getRootDirectory());
        Services services = initConfig.createFacadeForTransl().services();
        JavaInfo javaInfo = services.getJavaInfo();

        Term parameterEqualities = null;
        for (int i = 0; i<methodParameters.length; i++) {
            //      System.out.println(methodParameters[i]+"==>"+parentMethodParameters[i]);
            if (parameterEqualities !=null)
                parameterEqualities =
                    TermFactory.DEFAULT.createJunctorTerm
                    (Op.AND,
                            parameterEqualities,
                            TermFactory.DEFAULT.createEqualityTerm
                            (TermFactory.DEFAULT.createVariableTerm(parentMethodParameters[i]),
                                    TermFactory.DEFAULT.createVariableTerm(methodParameters[i])));
            else
                parameterEqualities = TermFactory.DEFAULT.
                createEqualityTerm
                (TermFactory.DEFAULT.createVariableTerm(parentMethodParameters[i]),
                        TermFactory.DEFAULT.createVariableTerm(methodParameters[i]));
        }

        Term preCondImplication = TermFactory.DEFAULT.
        createJunctorTerm(Op.IMP, parentPrePO, prePO);
        if (parameterEqualities != null)
            preCondImplication = TermFactory.DEFAULT.
            createJunctorTerm(Op.IMP,
                    parameterEqualities,
                    preCondImplication);

        Term postCondImplication = TermFactory.DEFAULT.createJunctorTerm
        (Op.IMP,
                TermFactory.DEFAULT.createJunctorTerm(Op.AND, parentPrePO, prePO),
                TermFactory.DEFAULT.createJunctorTerm(Op.IMP, postPO, parentPostPO));
        if (parameterEqualities != null)
            postCondImplication = TermFactory.DEFAULT.
            createJunctorTerm(Op.IMP,
                    parameterEqualities,
                    postCondImplication);

        return result = ProofAggregate.createProofAggregate(new Proof[] {
                new Proof("pre conditions", preCondImplication, s,
                        initConfig.createTacletIndex(),
                        initConfig.createBuiltInRuleIndex(),
                        initConfig.getServices()),
                        new Proof("post conditions",
                                postCondImplication,
                                s,
                                initConfig.createTacletIndex(),
                                initConfig.createBuiltInRuleIndex(),
                                initConfig.getServices())
        }, name());
    }

    public ReprModelClass getModelClass() {
        return methRepr.getContainingClass();
    }

    public Contractable[] getObjectOfContract() {
        return new Contractable[0];
    }

    public boolean initContract(Contract ct) {
        return false; //%%
    }

}
