// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof.init;

import java.io.*;
import java.net.URL;
import java.util.*;
import org.apache.log4j.Logger;

import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.proof.mgt.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.gui.LibrariesSettings;
import de.uka.ilkd.key.gui.ProofSettings;
import de.uka.ilkd.key.gui.Main;
import de.uka.ilkd.key.logic.ldt.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.KeYExceptionHandler;
import de.uka.ilkd.key.util.KeYResourceManager;
import de.uka.ilkd.key.proof.decproc.JavaDecisionProcedureTranslationFactory;

public class ProblemInitializer {

    private static final RuleSource STANDARD_RULES = 
	RuleSource.initRuleFile("standardRules.key");
    private static final ListOfBuiltInRule DEFAULT_BUILT_IN_RULES;
    public static InitConfig BASE_CONFIG = null;
    private static JavaModel lastModel   = null;

    private Main main = null;

    private ProofOblInput problem = null;
    private HashSet alreadyParsed = new HashSet();
    private KeYExceptionHandler exceptionHandler = null;


    private InitConfig initConfig=null;

    private UpdateSimplifier simplifier;

    private Logger mgtLogger = Logger.getLogger("key.proof.mgt");

    static{
	// update simplifier
	ListOfBuiltInRule list = SLListOfBuiltInRule.EMPTY_LIST.
	    prepend(UpdateSimplificationRule.INSTANCE);
	// check whether "Simplify" is in $PATH. If not,  the 
	// taclet "Run Decision Procedure" won't be available.
	try {
	    
	    File file = File.createTempFile("key-simplify", null);
	    PrintWriter out = new PrintWriter(new FileWriter(file));
	    out.println("TRUE");
	    out.close();
	    Process p = Runtime.getRuntime().exec
		(new String[]{ "Simplify", file.getPath() });
	    list = list.prepend(new SimplifyIntegerRule
				(new JavaDecisionProcedureTranslationFactory()));
	    file.delete();
	}
	catch (IOException e)
	    {Debug.out("Exception thrown by class ProblemInitializer at file-read/write");} 
	// check whether "ICS" is in $PATH. If not,  the 
	// taclet "Run Decision Procedure" won't be available.
	try {
	    
	    File file = File.createTempFile("key-ics", null);
	    PrintWriter out = new PrintWriter(new FileWriter(file));
	    out.println("assert true.");
	    out.close();
	    Process p = Runtime.getRuntime().exec
		(new String[]{ "ics", file.getPath() });
	    list = list.prepend(new ICSIntegerRule
				(new JavaDecisionProcedureTranslationFactory()));
	    file.delete();
	}
	catch (IOException e)
	    {Debug.out("Exception thrown by class ProblemInitializer at file-read/write");} 
	finally {
	    DEFAULT_BUILT_IN_RULES = list;
	}
    }

    
    public ProblemInitializer(Main main) throws ProofInputException {
        this.main = main;
        simplifier = ProofSettings.DEFAULT_SETTINGS
                .getSimultaneousUpdateSimplifierSettings().getSimplifier();
        this.exceptionHandler = main.mediator().getExceptionHandler();
    }
    
    //for tests only
    private ProblemInitializer(ProofOblInput problemFile, boolean loadLDTs)
    throws ProofInputException { 
        this.problem=problemFile;
        KeYFile ldtsForTests =null;
        if (loadLDTs) {
            ldtsForTests = new KeYFile("ldts",
                    RuleSource.initRuleFile("LDTsForTestsOnly.key"),
                    null);
        }
        if (BASE_CONFIG==null) {
            initConfig = new InitConfig(exceptionHandler);
            initConfig.setNoStandardRules(false);
            initConfig.setBuiltInRules(DEFAULT_BUILT_IN_RULES);	
            BASE_CONFIG = initConfig;
            problem.setConfig(initConfig);
            if (loadLDTs) {
                readKeYFile(ldtsForTests);        
            }
        }
    }
    
    //for tests only
    public static ProblemInitializer 
    	problemInitializerForTests(ProofOblInput problemFile, boolean loadLDTs)
                                            throws ProofInputException {
	return new ProblemInitializer(problemFile, loadLDTs);
    }

    public static String startProver(final ProofOblInput po) {
	return startProver(po, null);
    }

    public static String startProver(final ProofOblInput po, 
				     final ProofEnvironment env) {
        String startProverError = "";
        Main prover = Main.getInstance();
        prover.mediator().stopInterface(true);
        try{
	    ProblemInitializer init =
               new ProblemInitializer(prover);
	    prover.setStatusLine("Loading Problem");
	    ProofAggregate pl;
	    if (env!=null) {
		pl = init.setUpProof(po, env);
	    } else {
		pl = init.setUpProof(po);
	    }
        } catch (ProofInputException e) {
           prover.setStatusLine(po.name()+" failed");
           startProverError = e.toString();
        }
	prover.setStandardStatusLine();
        prover.mediator().startInterface(true);
	return startProverError;
    }


    public InitConfig initConfig() {
	return initConfig;
    }

    /** 
     * displays the status report in the status line
     */
    private void reportStatus(String status) {
	if (main != null) {
	    main.setStatusLine(status);	
	}
    }

    /** 
     * displays the status report in the status line 
     * and the maximum used by a progress bar
     * @param status the String to be displayed in the status line
     * @param progressMax an int describing what is 100 per cent
     */
    private void reportStatus(String status, int progressMax) {
	if (main != null) {
	    main.setStatusLine(status, progressMax);	
	}
    }

    /** 
     * displays the standard status line
     */
    private void reportReady() {
	if (main != null) {
	    main.setStandardStatusLine();
	}
    }

    private InitConfig createInitConfig() {
	InitConfig localInitConfig = new InitConfig(exceptionHandler);
	localInitConfig.setBuiltInRules(DEFAULT_BUILT_IN_RULES);
	localInitConfig.setProofEnvironment(new ProofEnvironment(localInitConfig));
	//For OCL simplification
	if (problem instanceof OCLSimplificationPO) {
	    ((OCLSimplificationPO)problem).createNamespaceSet(localInitConfig);
	}
	//
	return localInitConfig;
    }

    /** reads the RuleSources contained in in[0] and in[1]
     * @param in the Include object containing the filenames parsed in
     * the include section of a KeYFile.
     */
    private void readEnvInputs(Includes in) throws ProofInputException {
	if(in.getIncludes().contains("standardRules")){
	    if(initConfig == null){
		if(BASE_CONFIG == null){
		    KeYFile standard = new KeYFile("standard taclets", 
						   in.get("standardRules"),
						   main.getProgressMonitor());
		    initConfig = createInitConfig();
		    readKeYFile(standard);
		    BASE_CONFIG = initConfig;
		}
		initConfig = BASE_CONFIG.copy();
		in.remove("standardRules");
	    }
	    initConfig.setNoStandardRules(false);
	}
	if(!in.getLDTIncludes().isEmpty()) readLDTs(in);
	//For OCL simplification
	if (initConfig != null && problem instanceof OCLSimplificationPO) {
	    ((OCLSimplificationPO)problem).createNamespaceSet(initConfig());
	}
	//
	Iterator it = in.getIncludes().iterator();
	while(it.hasNext()){
	    String fileName = (String) it.next();
	    KeYFile keyFile = new KeYFile(fileName, 
					  in.get(fileName),
					  (main==null) ?
					  null : main.getProgressMonitor());
	    readKeYFile(keyFile);
	}
    }

    private void readKeYFile(KeYFile file) throws ProofInputException {
	if(alreadyParsed.add(file)){
	    Includes includes = file.readIncludes();
	    readEnvInputs(includes);
	    if(initConfig == null){
		initConfig = createInitConfig();
	    }
	    file.addTo(initConfig);
	    reportStatus("Reading "+file.name(), file.getNumberOfChars());
	    file.read(ModStrategy.NO_VARS_FUNCS_GENSORTS);
	}
    }
       
    private void readLDTs(Includes in) throws ProofInputException{
	KeYFile[] keyFile = new KeYFile[in.getLDTIncludes().size()];
	Iterator it = in.getLDTIncludes().iterator();
	int i = 0;
	String name;
	if(initConfig == null){
	    initConfig = new InitConfig(exceptionHandler);
	    initConfig.setBuiltInRules(DEFAULT_BUILT_IN_RULES);
	}
	while(it.hasNext()){
	    name = (String) it.next();
	    keyFile[i] = new KeYFile(name, 
				     (RuleSource)in.get(name), 
				     (main==null) ? null : 
				     main.getProgressMonitor());
	    keyFile[i].addTo(initConfig);
	    Includes includes = null;
	    includes = keyFile[i].readIncludes();
	    if(includes != null){
		readEnvInputs(includes);
	    }
	    i++;
	}
	LDTInput ldtInp = new LDTInput(keyFile, main);
	ldtInp.addTo(initConfig);
	ldtInp.read(ModStrategy.NO_VARS_FUNCS_GENSORTS);
    }


    public ProofAggregate setUpTest()  throws ProofInputException {
        initConfig = BASE_CONFIG.copy();
	problem.setConfig(initConfig);
	initConfig.setActivatedChoices(problem.readActivatedChoices());
	initConfig.initLDTSettings(ProofSettings.DEFAULT_SETTINGS);
	initializeJavaModel();
	problem.read(ModStrategy.NO_FUNCS);
	//	prepareSpecifications();
        if (problem instanceof KeYFile)
        initConfig.setLibraries(
                ((KeYUserProblemFile)problem).readLibrariesSettings().getLibraries());
        readLibraries();
	return problem.getPO();
    }


    // called by setUp<...>
    public void initializeJavaModel() throws ProofInputException {
	reportStatus("Reading Java model");
	problem.readJava();
	reportReady();
    }

    public JavaModel getJavaModel() throws ProofInputException {
	reportStatus("Checking Java model");
        JavaModel jModel = JavaModel.NO_MODEL;
	String modelDir = problem.getJavaPath();
        String modelTag = null;
        if (modelDir != null) { 
	    modelTag = "KeY_"+new Long((new java.util.Date()).getTime());
	    jModel = new JavaModel(modelDir, modelTag);
            if (modelDir.equals(System.getProperty("user.home"))) 
                throw new ProofInputException("You do not want to have "+
                "your home directory as the program model.");
	    CvsRunner cvs = new CvsRunner();
	    try{
		boolean importOK = 
		    cvs.cvsImport(jModel.getCVSModule(), modelDir,
				  System.getProperty("user.name"),
				  modelTag);
		if (importOK && lastModel!=null && 
		    lastModel!=JavaModel.NO_MODEL && 
		    modelDir.equals(lastModel.getModelDir())) {
		    String diff = cvs.cvsDiff(jModel.getCVSModule(),
					      lastModel.getModelTag(), 
					      modelTag);
		    if (diff.length()==0) {
			jModel = lastModel;
		    }
		}
	    }catch(CvsException cvse) {
		// leave already created new Java model
		mgtLogger.error("Dumping Model into CVS failed: "+cvse);
	    }
	}
	lastModel = jModel;
	reportReady();
        return jModel;
    }

    private void prepareSpecifications() throws ProofInputException {
	reportStatus("Preparing specifications" );
	problem.readSpecs();
	MethSpecInput methSpecInp = new MethSpecInput("Method Specification",
						      initConfig);
	methSpecInp.read(ModStrategy.MOD_ALL);
	reportReady();
    }


    /** reads the libraries */
    
    private void readLibraries() throws ProofInputException {
        reportStatus("Loading Libraries");
        HashMap libraries = initConfig.getLibraries();
        if (libraries.size()==0)
            return;
        String path = LibrariesSettings.getLibrariesPath();
        Iterator it = libraries.keySet().iterator();
        while (it.hasNext()){
            String fileName = (String) it.next();
            Boolean  sel =  (Boolean) libraries.get(fileName);
            if (sel.booleanValue()) {
                File file;
                if (!fileName.startsWith(File.separator)) {
                    KeYResourceManager rm = KeYResourceManager.getManager();
                    URL url = rm.getResourceFile(RuleSource.class,
                                                path+fileName);
                    file = new File(url.getFile());
                } else {
                    file = new File(fileName);
                }
                KeYFile keyFile = new KeYFile(fileName, file, 
                            (main == null) ? null : main
                        .getProgressMonitor());
                keyFile.addTo(initConfig);
                Includes includes = keyFile.readIncludes();
                readEnvInputs(includes);
                keyFile.read(ModStrategy.MOD_ALL);
            }
        }
        reportReady();
    }

    /**
     * sets up a new proof environment (initialises rules and model, registers
     * rules) without preparing specifications.
     */
    private ProofEnvironment setUpNewEnvironment(JavaModel jmodel) 
	throws ProofInputException {

	Includes includes = problem.readIncludes();
	if (includes.isEmpty()) {
	    includes.put("standardRules", STANDARD_RULES);
	}
	readEnvInputs(includes);
	alreadyParsed.clear();
	problem.setConfig(initConfig);
	problem.startProtocol();
	initConfig.setActivatedChoices(problem.readActivatedChoices());
        if (problem instanceof KeYFile)
                initConfig.setLibraries(
                        ((KeYFile)problem).readLibrariesSettings().getLibraries());        
        RuleConfig ruleConfig = new RuleConfig(initConfig.getActivatedChoices());
        
	if(!initConfig.noStandardRules()){
	    initConfig.initLDTSettings(ProofSettings.DEFAULT_SETTINGS);
	}
	initializeJavaModel();	
//	System.out.println("getProofEnv: "+initConfig().getProofEnv());
	initConfig().getProofEnv().setJavaModel(jmodel);
	initConfig().getProofEnv().setRuleConfig(ruleConfig);
	initConfig.getProofEnv().registerRules(DEFAULT_BUILT_IN_RULES, 
					       AxiomJustification.INSTANCE);
	return initConfig.getProofEnv();
    }

    /** reads the problem file for an already initialised proof
	environment 
    */ 
    private void readProblem(ProofEnvironment env) throws ProofInputException {
	initConfig = env.getInitConfig();
	initConfig.namespaces().setProgramVariables(new Namespace());	
	problem.setConfig(initConfig);
	problem.startProtocol();

	reportStatus("Loading problem \""+problem.name()+"\"");
	problem.readProblem(ModStrategy.NO_FUNCS);
	reportReady();	
    }
    
    private void fixNamespaces() {
    	NamespaceSet nss = initConfig().createNamespacesForActivatedChoices();
        initConfig().getServices().setNamespaces(nss);	
    }

    /** determines an environment (creates one if necessary) and reads
     * the problem file. If the proof environment must be created it
     * is properly initialised, including prepared specifications.
     */
    private ProofEnvironment getEnvironmentAndRead() throws ProofInputException {       
        JavaModel jmodel = getJavaModel();
	ProofEnvironment env = setUpNewEnvironment(jmodel);
	ProofEnvironment envChosen = 
	    GlobalProofMgt.getInstance().getProofEnvironment
	    (jmodel, env.getRuleConfig(), problem.askUserForEnvironment());
	if (envChosen == null){
	    problem.read(ModStrategy.MOD_ALL);
	    GlobalProofMgt.getInstance().registerProofEnvironment(env);

	    reportStatus("Registering rules");
 	    initConfig.getProofEnv().registerRules(initConfig.getTaclets(), 
						   AxiomJustification.INSTANCE);
 	    
	    reportReady();
   
            if (!(problem instanceof OCLSimplificationPO)) {		
	        readLibraries();
            }
            
            fixNamespaces();
            
            if (!(problem instanceof OCLSimplificationPO)) {
                prepareSpecifications();
            }
            
            return env;
	}else{
	    readProblem(envChosen);
	    fixNamespaces();
	    return envChosen;
	} 
    }


    private ProofAggregate setUpProofHelper(ProofEnvironment env) 
	throws ProofInputException {
	JavaModel jmodel = env.getJavaModel();
	RuleConfig ruleConfig = env.getRuleConfig();
	ProofAggregate pl = problem.getPO();
	Proof[] proofs = pl.getProofs();
	InitConfig pi = initConfig();
	for (int i=0; i < proofs.length; i++) {
	    proofs[i].setSimplifier(simplifier);
	    //proofs[i].setActivatedChoices(pi.getActivatedChoices());
	    //proofs[i].setLDTSettings(settings.getLDTSettings());
            proofs[i].setNamespaces(pi.getServices().getNamespaces());
	}
	initConfig().getProofEnv().registerProof(problem, pl);
	main.addProblem(pl);
	GlobalProofMgt.getInstance().tryReuse(pl);
	return pl;
    }

    /** determines an appropriate proof environment (creates one if
     * necessary) and sets up a proof in the environment. 
     */ 
    public ProofAggregate setUpProof(ProofOblInput problemFile) 
	throws ProofInputException {

	this.problem = problemFile;
	return setUpProofHelper(getEnvironmentAndRead());
    }

    /** sets up a proof in a given proof environment */
    public ProofAggregate setUpProof(ProofOblInput problemFile,
				     ProofEnvironment env) 
	throws ProofInputException {
	this.problem = problemFile;
	readProblem(env);
	fixNamespaces();
	return setUpProofHelper(env);
    }


    public void tryReadProof(ProblemLoader prl) throws ProofInputException {
	if (problem instanceof KeYFile) {
	    KeYFile proof = (KeYFile)problem;
	    reportStatus("Loading proof", proof.getNumberOfChars());
	    proof.readProof(prl);
	}
    }
    
    public LDT[] ldtModels() {
	return initConfig.ldtModels();
    }

}
