// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof.mgt;

import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.java.ContextStatementBlock;
import de.uka.ilkd.key.java.declaration.MethodDeclaration;
import de.uka.ilkd.key.java.statement.MethodBodyStatement;
import de.uka.ilkd.key.java.statement.CatchAllStatement;
import de.uka.ilkd.key.java.statement.Throw;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.util.*;
import java.util.*;

/** 
 * This abstract class implements the transformation function used to
 * generate a method lemma taclet from the method's specification. The
 * transformation is described in detail in the minor thesis of Bastian
 * Katz for OCL specifications.
 * The generated taclet is of similar to :
 *  <code>
 *     find(==> <{t::mbs(params);}> #postcond 
 *     add(==> pre);
 *     add(==>\forall modifies_schema_variables 
 *                 (\forall transformation_vars (pi1 -> pi2_with_updates) 
 *                     -> #postcond_with_updates))
 *  </code>
 * where <ul>
 *    <li><code>pre</code> is the specified precondition </li>
 *    <li><code>postcond</code> is the specified postcondition </li>
 *    <li><code>pi1, pi2, modifies_*, </code> are connected with the
 *    modifies clause and the elimination of <code>@pre</code> like operators</li>
 * </ul>
 */
public abstract class AbstractMethodLemmaFactory {
    

    private final TermFactory tf = TermFactory.DEFAULT;
 
    private Term updMethPost;
    private Statement transformedStatement;
    private Term preCondition;
    private Services services;
    private int cnt;
    private int lvCnt;
    private Term origPostSVTerm;
  
    /** 
     * precondition of the method whose specification is processed
     */
    protected abstract Term getPreCond(); 
    
    /** 
     * quantifiable variables used to store the values of non-rigid
     * functions in order to access them in the postcondition of the
     */
    protected abstract ListOfQuantifierPrefixEntry getExtraVars(); 

    /**
     * A further list of quantifiers that have to be added to pre- or
     * postcondition.
     */
    protected abstract ListOfQuantifierPrefixEntry getQuantifierPrefix();
    
    /** 
     * additional precondition necessary to eliminate
     * <code>@pre</code> like operators 
     */
    protected abstract Term getExtraPre();
    
    /** 
     * post condition of the method specification
     */
    protected abstract Term getPostCond();

    /** modifies clause as an array of modifies elements */
    protected abstract ModifiesElement[] getModifies(); 
    
    /** 
     * the statement, usually the method body statement, this taclet is about. 
     * It contains concrete code, that is no schema variables.
     */
    protected abstract ProgramElement getProofStatement();

    /** the modality to be used in the created taclet. This may 
     * be a SchemaModelOperator
     */
    protected abstract Operator getModality();
    
    /** 
     * string which can be element of the taclet name and the taclet 
     * display name
     */
    protected abstract String getSuffix();

    /** rule set the taclet belongs to */
    protected abstract RuleSet getRuleSet();

    
    
    protected AbstractMethodLemmaFactory() {}

    protected AbstractMethodLemmaFactory(Services services) {
	this.services = services;
    }

    /**
     * creates and returns the generated taclet for the given method
     * specification
     * @return the taclet that can be used to insert the given method
     * specification rather than the concrete method implementation
     */
    public Taclet getTaclet() {	
	if ( getModifies() == null ) {
	    return null;
	}        
        
	Map schemaVariablesMap = new HashMap();
	createSpecialSVMappings(schemaVariablesMap);
	// special treatment to specify thrown exceptions
	ProgramVariable catchExceptionsAs = null;
	Statement unwrappedProofStatement 
	    = unwrapStatementBlocks((Statement)getProofStatement());
	if (unwrappedProofStatement instanceof CatchAllStatement) {
	    CatchAllStatement catchAll 
		= (CatchAllStatement) unwrappedProofStatement;
	    unwrappedProofStatement = catchAll.getBody();
	    catchExceptionsAs 
		= (ProgramVariable) catchAll.getParameterDeclaration()
		      .getVariableSpecification().getProgramVariable();	
	    SchemaVariable catchExcAsSV = SchemaVariableFactory.
		createProgramSV(new ProgramElementName("#thrownException"), 
				ProgramSVSort.VARIABLE, false);
	    schemaVariablesMap.put(catchExceptionsAs, catchExcAsSV);
	}

         
        
	// substitute program variables by schema variables
	final TermSVReplacer tSVR = new TermSVReplacer(schemaVariablesMap);
	transformedStatement = tSVR.inst(unwrappedProofStatement);
	preCondition = tSVR.inst(getPreCond()); // PRE
	Term extraPreCondition = tSVR.inst(getExtraPre()); // pi1
	updMethPost = tSVR.inst(createUpdMethPost()); // pi2
		
	//now start to build the taclet...
	SuccTacletBuilder stb = new SuccTacletBuilder();

	// sanity check: no variables are bound both in pre- and postcondition
	assertDisjointExtraFreeVars ( updMethPost, preCondition );
	assertDisjointExtraFreeVars ( extraPreCondition, preCondition );
    
	Map lV2SV = new HashMap();
	// this is the second add clause for the postcondition
	Term addFormula = createNormalCaseFormula(extraPreCondition,
						  catchExceptionsAs, 
						  schemaVariablesMap, lV2SV);

        stb = stb.setFind(createFindTerm());	
	
	// add one add part to the taclet builder
	stb.addTacletGoalTemplate(createTGTfromTerm(createPreFormula(preCondition, lV2SV), 
						    "Precondition",
						    null));
	SortedSchemaVariable excSV 
	    = (SortedSchemaVariable)schemaVariablesMap.get(catchExceptionsAs);
	// add the other add part to the taclet builder
	stb.addTacletGoalTemplate(createTGTfromTerm(addFormula, 
						    "Postcondition",
						    excSV));
	// add third (exceptional) case if necessary
	if (catchExceptionsAs!=null) {
	    Term excFormula = createExceptionalCaseFormula
		(extraPreCondition, catchExceptionsAs, 
		 schemaVariablesMap, lV2SV);
	    stb.addTacletGoalTemplate(createTGTfromTerm(excFormula,
							"Exception thrown",
							excSV));
	    stb.addVarsNew(excSV, catchExceptionsAs.sort());
	}
	//add `not free in' variable conditions
	Iterator it = lV2SV.values().iterator();
	while (it.hasNext()) {
	    stb.addVarsNotFreeIn((SchemaVariable)it.next(), 
				 (SchemaVariable)origPostSVTerm.op());
	}

	//add `new' variable conditions
	SetOfProgramVariable newPVs = SetAsListOfProgramVariable.EMPTY_SET;
	Iterator pvIt = schemaVariablesMap.keySet().iterator();
	while (pvIt.hasNext()) {
	    ProgramVariable pv = (ProgramVariable)pvIt.next();
	    if (!tSVR.getPVsInProg().contains(pv)) {
		newPVs = newPVs.add(pv);
	    }
	}
	
	IteratorOfProgramVariable newPVit 
	    = tSVR.getLocallyDeclaredVars().union(newPVs).iterator(); 
	while (newPVit.hasNext()) {
	    ProgramVariable pv = newPVit.next();
	    stb.addVarsNew((SchemaVariable)schemaVariablesMap.get(pv), 
			   pv.sort());	    
	}
	//add names and rule sets
	stb.setDisplayName(getTacletDisplayName(unwrappedProofStatement, 
						((catchExceptionsAs==null) ? "" 
						 : "exceptional_")+getSuffix()));
	stb.setName(getTacletName(unwrappedProofStatement));
	if (getRuleSet()!=null) {
	    stb.addRuleSet(getRuleSet());
	}
	Taclet t = stb.getSuccTaclet();
	return t;
    }

    private Term createUpdMethPost() {
        if (getModifies().length==0) return getPostCond();
        Term[] updMethPostLHS = new Term[getModifies().length];
        Term[] updMethPostRHS = new Term[getModifies().length];
        Term[] updMethPostGuard = new Term[getModifies().length];
        ArrayOfQuantifiableVariable[] updMethPostVars 
           = new ArrayOfQuantifiableVariable[getModifies().length];
        for (int i=0; i<getModifies().length; i++) {
            UpdatePairWithGuard up = getModifies()[i].createUpdateToRigid(services,
                    "c"+cnt++, "i"+lvCnt++);
            updMethPostLHS[i] = up.getLHS();
            updMethPostRHS[i] = up.getRHS();
            updMethPostGuard[i] = up.getGuard();
            updMethPostVars[i] = up.createArrayOfQuantifiedVars();
        }
        return tf.createQuanUpdateTerm(updMethPostVars, updMethPostGuard, 
                updMethPostLHS, updMethPostRHS, getPostCond());        
    }
    
    private void assertDisjointExtraFreeVars (Term t1, Term t2) {
        final IteratorOfQuantifierPrefixEntry it = getQuantifierPrefix().iterator();
        while ( it.hasNext () ) {
            final QuantifiableVariable var = it.next().getVariable();
            Debug.assertFalse ( t1.freeVars ().contains ( var )
                                && t2.freeVars ().contains ( var ) );
        }
    }

    private StringBuffer getNameBody(Statement transformedStmnt) {
	final StringBuffer stbName = new StringBuffer();
	if (transformedStmnt instanceof MethodBodyStatement) {
	    final MethodBodyStatement mbs = (MethodBodyStatement) transformedStmnt;            
            stbName.append
		(mbs.getProgramMethod(services).name().toString());	    
	}
	return stbName;
    }

    /**
     * returns the name to be displayed when the user is offered this taclet
     * @param sta the statement this taclet talks about
     * @param suffix the suffix added to the name
     * @return a name body relative to the given statement, underscore, suffix
     */
    protected String getTacletDisplayName(Statement sta,
					String suffix) {
	return getNameBody(sta) + "_"+suffix;
    }

    /** 
     * calculates and returns the name of the created taclet
     * @param sta the statement this taclet talks about
     * @return the name of the created method lemma taclet
     */
    protected Name getTacletName(Statement sta) {
	final StringBuffer stbName = getNameBody(sta);
	if (sta instanceof MethodBodyStatement) { 
	    final MethodDeclaration meth = 
		((MethodBodyStatement) sta).
		getProgramMethod(services).getMethodDeclaration();

	    for (int i = 0; i<meth.getParameterDeclarationCount(); i++) {
		stbName.append("_");
		stbName.append(meth.getParameterDeclarationAt(i).
			       getVariableSpecification().getType().getName());
	    }
	} 
	stbName.append("_");
	stbName.append(getSuffix());
	return new Name(stbName.toString());
    }
	    
    private Term createUpdOrigPost(Statement[] st) {
        if (origPostSVTerm==null) {
            origPostSVTerm = tf.createVariableTerm
                              (SchemaVariableFactory.createFormulaSV
                                (new ProgramElementName("#origPost"), false));
        }	    

        ExtList statementList = new ExtList();
	if (st!=null) {
	    for (int i=0; i<st.length; i++) {
		statementList.add(st[i]);
	    }
	} 

	Term origPost = tf.createProgramTerm
	    (getModality(),
	     JavaBlock.createJavaBlock(new ContextStatementBlock(statementList)), 
	     origPostSVTerm);
        
	
	if (getModifies().length==0) {
	    return origPost;
	} else {
	    Term[] subs = new Term[updMethPost.arity()];
	    ArrayOfQuantifiableVariable[] bdVars 
		= new ArrayOfQuantifiableVariable[updMethPost.arity()];
	    for (int i=0; i<subs.length-1; i++) {
		subs[i]   = updMethPost.sub(i);
		bdVars[i] = updMethPost.varsBoundHere(i);
	    }
	    subs[subs.length-1] = origPost;
	    bdVars[bdVars.length-1] = updMethPost.varsBoundHere(bdVars.length-1);
	    
	    return tf.createQuanUpdateTerm((QuanUpdateOperator)updMethPost.op(), subs, bdVars); 
	  }
    }
    
    private static SortedSchemaVariable getSV(QuantifiableVariable lv, 
					      Map logic2Schema) {
	if (!logic2Schema.containsKey(lv)) {
	    logic2Schema.put(lv, 
			     SchemaVariableFactory.
			     createVariableSV(lv.name(), lv.sort(), false));
	}
	return (SortedSchemaVariable)logic2Schema.get(lv);
    }

    /**
     * replaces all logic variables in t with schema variables. The used
     * mapping logic variable to schema variable is stored in the
     * logic2Schema map 
     * @param t the Term whose logic variables have to be replaced
     * @param logic2Schema a Map associating each logic variable to a
     * schema variable so that the same logic variable is replaced with
     * the same term
     * @return a term with all logic variables replaced by schema
     * variables 
     */
    private static Term bindLogicVarsAsSchemaVars(Term t, Map logic2Schema) {
	final Term[] subs = new Term[t.arity()];
	final ArrayOfQuantifiableVariable[] a
	    = new ArrayOfQuantifiableVariable[t.arity()];

	boolean changed = false;
	for (int i=0; i<subs.length; i++) {

	    subs[i] = bindLogicVarsAsSchemaVars(t.sub(i), logic2Schema);
	    if (subs[i]!=t.sub(i)) changed = true;

	    final ArrayOfQuantifiableVariable boundVars = t.varsBoundHere(i);
	    final int boundVarsSize = boundVars.size();
	    final QuantifiableVariable[] qs = 
		new QuantifiableVariable[boundVarsSize];

	    changed = (changed || boundVarsSize > 0);
	    for (int j=0; j<boundVarsSize; j++) {
		qs[j] = getSV(boundVars.getQuantifiableVariable(j), 
			      logic2Schema);
	    }
	    a[i] = new ArrayOfQuantifiableVariable(qs);		    
	}

	if (t.op() instanceof LogicVariable) {
	    return TermFactory.DEFAULT.createVariableTerm
		(getSV((LogicVariable)t.op(), logic2Schema));
	} else if (changed){
	    return TermFactory.DEFAULT.createTerm(t.op(), subs, a, 
						  t.javaBlock(), null);
	} else {
	    return t;
	}
    }

    private Term createPreFormula(Term formula, Map lV2SV) {
        formula = bindLogicVarsAsSchemaVars(formula, lV2SV);
        formula = addVars ( getQuantifierPrefix(), formula, lV2SV, true );
        return formula;
    }
    
    private Term createAddFormula(Term psiUpd, 
				  Term extraPreCondition,
				  Map lV2SV) {
	Term addFormula = tf.createJunctorTermAndSimplify(Op.AND, 
							  extraPreCondition, 
							  updMethPost);
	
	addFormula = bindLogicVarsAsSchemaVars(addFormula, lV2SV);
	psiUpd     = bindLogicVarsAsSchemaVars(psiUpd, lV2SV);

	addFormula = addVars ( getExtraVars(), addFormula, lV2SV, true );
	addFormula = addVars ( getQuantifierPrefix(), addFormula, lV2SV, false );
    
	addFormula = tf.createJunctorTermAndSimplify(Op.IMP, 
						     addFormula, 
						     psiUpd);
	return addFormula;
    }  

    private Term addVars (ListOfQuantifierPrefixEntry prefix,
                          Term addFormula,
                          Map lV2SV,
                          boolean invert) {
        if ( prefix.isEmpty () ) return addFormula;

        addFormula = addVars ( prefix.tail (), addFormula, lV2SV, invert );

        final QuantifierPrefixEntry entry = prefix.head ();
        final SortedSchemaVariable sv = getSV ( entry.getVariable (), lV2SV );
        if ( !addFormula.freeVars ().contains ( sv ) ) return addFormula;

        final Quantifier quantifier =
            ( entry.isUniversal () ^ invert ) ? Op.ALL : Op.EX;
        return tf.createQuantifierTerm ( quantifier, sv, addFormula );
    }

    private Term createNormalCaseFormula(Term extraPreCondition,
					 ProgramVariable catchExceptionsAs,
					 Map schemaVariablesMap,
					 Map lV2SV) {
	Term extraPre = extraPreCondition;
	Statement[] sta = null;
	if (catchExceptionsAs!=null) {
	    SortedSchemaVariable excSV 
		= (SortedSchemaVariable)
		schemaVariablesMap.get(catchExceptionsAs);
	    extraPre = tf.createJunctorTermAndSimplify
		(Op.AND, extraPreCondition,
		 tf.createEqualityTerm
		 (tf.createFunctionTerm(excSV), 
		  services.getJavaInfo().getNullTerm()));
	}	
	return createAddFormula(createUpdOrigPost(sta), 
				extraPre, lV2SV);
    }

    private Term createExceptionalCaseFormula(Term extraPreCondition,
					      ProgramVariable catchExceptionsAs,
					      Map schemaVariablesMap,
					      Map lV2SV) {
	SortedSchemaVariable catchExcAsSV 
	    = (SortedSchemaVariable)schemaVariablesMap.get(catchExceptionsAs);
	Term preexc = tf.createJunctorTermAndSimplify
	    (Op.AND, extraPreCondition, 
	     tf.createJunctorTerm
	     (Op.NOT,
	      tf.createEqualityTerm(tf.createFunctionTerm(catchExcAsSV), 
				    services.getJavaInfo().getNullTerm())));
	ExtList args = new ExtList();
	args.add(catchExcAsSV);
	Term psiUpd = createUpdOrigPost(new Statement[]{new Throw(args)});
	return createAddFormula(psiUpd, preexc, lV2SV);	
    }

    private static Statement unwrapStatementBlocks(Statement s) {
	while (s instanceof StatementBlock 
	       && ((StatementBlock)s).getChildCount()==1) {
	    s = (Statement)((StatementBlock)s).getChildAt(0);
	}
	return s;
    }

    private Term createFindTerm() {
	transformedStatement = unwrapStatementBlocks(transformedStatement);
	ExtList children = new ExtList();
	if (transformedStatement instanceof StatementBlock 
	    && ((StatementBlock)transformedStatement).getChildCount()>1) {
	    for (int i=0; i<((StatementBlock)transformedStatement).getChildCount();
		 i++) {
		children.add(((StatementBlock)transformedStatement).getChildAt(i));
	    }
	} else {
	    children.add(transformedStatement);
	}
	return tf.createProgramTerm
	    (getModality(),
	     JavaBlock.createJavaBlock(new ContextStatementBlock(children)), 
	     origPostSVTerm);
    }

    private TacletGoalTemplate createTGTfromTerm(Term add, String name, 
						 SortedSchemaVariable addPV) {
	final Semisequent semSeq = Semisequent.EMPTY_SEMISEQUENT.
	    insertFirst(new ConstrainedFormula(add)).semisequent();
	SetOfSchemaVariable addPVs = SetAsListOfSchemaVariable.EMPTY_SET;
	if (addPV!=null) {
	    addPVs=addPVs.add(addPV);
	}
	final TacletGoalTemplate tgt =
	    new AntecSuccTacletGoalTemplate
	    (Sequent.EMPTY_SEQUENT,
	     SLListOfTaclet.EMPTY_LIST, 
	     SLListOfRulevariable.EMPTY_LIST,
	     Sequent.createSuccSequent(semSeq), 
	     addPVs);
	tgt.setName(name);
	return tgt;
    }

    private void createSpecialSVMappings(Map schemaVariablesMap) {
	//the next case is to allow non-simple expressions like self.a
	//as callee in a mbs. nothing wrong can happen since mbs exist
	//only if there is no null pointer exception thrown
	Statement sta 
	    = unwrapStatementBlocks((Statement)getProofStatement());
	if (sta instanceof MethodBodyStatement) {
	    MethodBodyStatement mbs = (MethodBodyStatement) sta;
	    SchemaVariable methodPrefixSV = null;
	    if (!mbs.isStatic(services)) {
 		methodPrefixSV = SchemaVariableFactory.
		    createProgramSV(new ProgramElementName("#self"), 
				    ProgramSVSort.EXPRESSION, false);
 		schemaVariablesMap.put(mbs.getDesignatedContext(), 
 				       methodPrefixSV);
 	    }
	}
    }
    
    /**
     * Abstract representation of a modifies clause element. The 
     * only requirement is that the modifies element can create an 
     * update to new rigid symbols as used in method lemma rules 
     * and the loop invariant rule
     *
     */
    interface ModifiesElement {
        /**
         * creates an update which updates the locations of the modifies 
         * to new rigid (anonymising) symbols as used in method lemma rules 
         * and the loop invariant rule
         * @param services the services for looking up model information
         * @param base the base name of the rigid symbols
         * @param lvBase the base name of introduced logic variables
         * @return the created update
         */
        UpdatePairWithGuard createUpdateToRigid(Services services, 
                                                String base, String lvBase);
    }
    
    /**
     * a standard implementation of ModifiesElement. Modifies elements are 
     * terms including QuantifiedArrayTerms.
     *
     */
    static class TermModifiesElement implements ModifiesElement{
        private Term term;
        private int lvCnt;
        private final static TermFactory tf = TermFactory.DEFAULT;
        private Map op2Sk = new HashMap();
        
        public TermModifiesElement(Term t) {
            this.term = t;
        }
        
        /**
         * creates an array of modifies elements for a given set of 
         * terms which may include QuantifiedArrayTerms
         */
        public static ModifiesElement[] create(SetOfTerm modifies) {
            if (modifies==null) return null; //NOT the empty array
            ModifiesElement[] res = 
                new ModifiesElement[modifies.size()];
            IteratorOfTerm it = modifies.iterator();
            for (int i=0; it.hasNext(); i++) {
                res[i] = new TermModifiesElement(it.next());
            }
            return res;
        }
        
        public UpdatePairWithGuard createUpdateToRigid(Services services, 
                                                       String base,
                                                       String lvBase) {            
            Term lhs;
            Term rhs;
            Term guard;
            int cnt=0;
            lvCnt = 0;
            QuantifiableVariable[] quantVars;           
            List qva = new LinkedList();
            guard = getGuard(term, qva, lvBase, services);
            quantVars = ((QuantifiableVariable[])qva.toArray(new QuantifiableVariable[0]));
            lhs = replaceQuanIndex(term, qva);               
            rhs = createSkTerm(lhs, base, cnt, services);
            return new UpdatePairWithGuard(lhs, rhs, guard, quantVars);
    }
    
    private Term createSkTerm(Term t, String base, int cnt, Services services) {
        if (!op2Sk.containsKey(t.op())) {
            Sort[] argsorts = new Sort[t.arity()];
	
	    int offset = 0;
	    // %%HACK: but this class is disappearing soon anyways
	    if (t.op() instanceof AttributeOp) {
		
                final ProgramVariable attribute = (ProgramVariable)
			       ((AttributeOp)t.op()).attribute();
                if (t.sub(0).sort() instanceof ArraySort &&
                        "length".equals(attribute.name().toString())) {
                    argsorts[0] = services.getJavaInfo().getJavaLangObjectAsSort();
                } else {
                    argsorts[0] = attribute.getContainedInKeYJavaType().getSort();
                }
		offset = 1;
	    } 

            for (int i=offset; i<t.arity(); i++) {
                argsorts[i]=t.sub(i).sort();
            }

            Function skf = new RigidFunction(getSkFctName(t, base, cnt), 
                                             t.sort(), argsorts);
            op2Sk.put(t.op(), skf);
        }
        Function sk = (Function) op2Sk.get(t.op());
        Term[] subs = new Term[t.arity()];
        for (int i=0; i<t.arity(); i++) {
            subs[i] = t.sub(i);
        }
        return tf.createFunctionTerm(sk, subs);
    }
        
        private Name getSkFctName(Term t, String base, int cnt) {
            return new Name(base+(cnt++==0 ? "" : "_"+ --cnt));
        }
        
        private Term getGuard(Term t, List qva, String lvBase, 
                                     Services serv) {            
            Term guard = TermFactory.DEFAULT.createJunctorTerm(Op.TRUE);    
            if(t.op() instanceof AccessOp){
                if(t.op() instanceof QuantifiedArrayOp){
                    LogicVariable lv = 
                        new LogicVariable(new Name(lvBase
                                                   +(lvCnt++==0 ? "" : "_"+ --lvCnt)), 
                                          t.sub(1).sort());
                    qva.add(lv);
                    Term vt = tf.createVariableTerm(lv);
                    Function leq = (Function) serv.getNamespaces().
                    functions().lookup(new Name("leq"));
                    guard = tf.createJunctorTermAndSimplify(Op.AND, guard,
                            tf.createFunctionTerm(leq, new Term[]{t.sub(1), vt}));
                    guard = tf.createJunctorTermAndSimplify(Op.AND, guard,
                            tf.createFunctionTerm(leq, new Term[]{vt, t.sub(2)}));
                }
                for(int i=0; i<t.arity(); i++){
                    guard = tf.createJunctorTermAndSimplify(Op.AND, guard,
                            getGuard(t.sub(i), qva, lvBase, serv));
                }
            }
            return guard;
        }
        
        private static Term replaceQuanIndex(Term t, List vars){
            Term[] subTerms;
            if (t.op() instanceof AccessOp){
                if (t.op() instanceof QuantifiedArrayOp){
                    LogicVariable lv = (LogicVariable) vars.remove(0);
                    Term vt = tf.createVariableTerm(lv);
                    Term sub0 = replaceQuanIndex(t.sub(0), vars);
                    ArrayOp op;                    
                    if (t.op() instanceof ShadowedQuantifiedArrayOp) {
                        subTerms = new Term[3];
                        subTerms[2] = replaceQuanIndex(t.sub(3), vars);
                        op = ShadowArrayOp.getShadowArrayOp(sub0.sort());
                    } else {
                        subTerms = new Term[2];
                        op = ArrayOp.getArrayOp(sub0.sort());
                    }                    
                    subTerms[0] = sub0;
                    subTerms[1] = vt;
                    t = tf.createArrayTerm(op, subTerms);                    
                } else {
                    subTerms = new Term[t.op().arity()];
                    for(int i=0; i<t.arity(); i++){
                        subTerms[i] = replaceQuanIndex(t.sub(i), vars);
                    }
                    t = tf.createTerm(t.op(), subTerms, 
                            (ArrayOfQuantifiableVariable[]) null, 
                            null, null);
                }
            }
            return t;
        }
    }
    
    /**
     * represents an update pair (lhs, rhs) with a guard and quantified 
     * variables <code>v1,...,vn</code>, which would be written as
     *  <code>\for v1,...,vn \if guard lhs := rhs</code>
     */
    static class UpdatePairWithGuard {
        private Term lhs;
        private Term rhs;
        private Term guard;
        private QuantifiableVariable[] qvars;
        
        
        public UpdatePairWithGuard(Term lhs, Term rhs, 
                                   Term guard, QuantifiableVariable[] qvars) {
            this.lhs = lhs;
            this.rhs = rhs;
            this.guard = guard;
            this.qvars = qvars;
        }
        
        public Term getLHS() {
            return lhs;
        }

        public Term getRHS() {
            return rhs;
        }

        public Term getGuard() {
            return guard;
        }

        public QuantifiableVariable[] getQuantifiedVars() {
            return qvars;
        }
        
        public ArrayOfQuantifiableVariable createArrayOfQuantifiedVars() {
            return new ArrayOfQuantifiableVariable(qvars);
        }
    }
}
