// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.proof.mgt;

import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.SetOfTerm;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.op.Operator;
import de.uka.ilkd.key.rule.RuleSet;

/**
 * This class builds a method lemma taclet for a given dynamic logic
 * specification. A detailed description of this transformation function
 * can be found in the minor thesis of Bastian Katz.
 */
public class DLMethodLemmaFactory extends AbstractMethodLemmaFactory {

    /** the termfactory to use */
    private final TermFactory tf = TermFactory.DEFAULT;

    /** precondition of the method */
    private final Term preCondition;
    /** postcondition of the method */
    private final Term postCondition;
    /** heuristic to which the created lemma taclet will belong */
    private final RuleSet ruleSet;
    /** set of term describing the locations change by the method */
    private final SetOfTerm modifies;
    /** the statement whose specification is transformed 
     * to a taclet 
     */
    private final Statement statement;

    private final Operator  modality;

    private final String tacletName;
    
    private final String displayName;
    
    private AbstractMethodLemmaFactory.ModifiesElement[] modElems;


    /**
     * creates a method lemma factory for specifications given in
     * dynamic logics
     * @param preCondition the Term representing the method's precondition
     * @param postCondition the Term representing the method's
     * postcondition 
     * @param statement the statement whose
     * specification is transformed to a taclet
     * @param ruleSet the Heuristic to which the lemma belongs
     * @param modifies the SetOfTerm describing the locations change by the method 
     * @param suffix a piece of text attached to the name of the lemma
     */
    public DLMethodLemmaFactory(Term preCondition,
				Term postCondition,
				Statement statement,
				RuleSet ruleSet,
				SetOfTerm modifies,
				Operator modality,
				String suffix,
				String displayName,
				Services services) {
	super(services);
	this.preCondition = preCondition;
	this.postCondition = postCondition;
	this.statement = statement;
	this.ruleSet = ruleSet;
	this.modifies = modifies;	       
	this.modality = modality;
	this.displayName = displayName;
	this.tacletName = (suffix==null) ? "DL" : suffix;
    }
    

    /**
     * Usually the extra precondition is used to eliminate 
     * <code> @pre </code>-like operators of OCL/JML. In principal the
     * precondition stores the value of non-rigid terms before method
     * execution. As we do not support an <code> @pre </code> operator
     * in dynamic logic itself, this is not necessary at the
     * moment. As the created lemma is of shape 
     * <code> pre and extrapre -> <m> extraPos -> post </code> , 
     * we return the neutral element of the conjunction.
     * @return precondition storing values of <code>@pre</code>
     * postfixed functions
     */
    protected Term getExtraPre() {
	return tf.createJunctorTerm(Op.TRUE);
    }

    /**
     * these variables are used to eliminate <code>@pre</code> like
     * operators. In fact they are the \"stores\" for the values of
     * non-rigid terms whose initial values have to be accessed after
     * the execution of the method.
     * @return the variables used to store the function values of
     * <code>@pre</code> postfixed functions
     */
    protected ListOfQuantifierPrefixEntry getExtraVars() {
	return SLListOfQuantifierPrefixEntry.EMPTY_LIST;
    }

    protected ListOfQuantifierPrefixEntry getQuantifierPrefix () {
        return SLListOfQuantifierPrefixEntry.EMPTY_LIST;
    }

    /**
     * returns the heuristic the taclet belongs to
     */
    protected RuleSet getRuleSet() {
	return ruleSet;
    }

    /**
     * returns the terms describing the locations to be modified by the
     * method
     * @return set of terms representing changed locations 
     */
    protected AbstractMethodLemmaFactory.ModifiesElement[] getModifies() {
        if (modElems==null) {
            modElems = AbstractMethodLemmaFactory.TermModifiesElement.create(modifies);
        }
        return modElems;
    }

    /**
     * returns the postcondition of the method
     * @return the postconditions of the method
     */
    protected Term getPostCond() {
	return postCondition;
    }

    /**
     * returns the precondition of the method
     * @return the preconditions of the method
     */
    protected Term getPreCond() {
	return preCondition;
    }

    /** returns the method body statement that is specified here */
    protected ProgramElement getProofStatement() {
	return statement;
    }
  
    /** returns a unique suffix */
    protected String getSuffix() {
	return tacletName;
    }
        
    protected String getTacletDisplayName(Statement transformedStatement,
		                          String suffix) {
        return displayName!=null ? displayName 
        		         : getTacletName(transformedStatement).toString();
    }
    
    protected Name getTacletName(Statement transformedStatement) {
        return new Name(tacletName);
    }

    protected Operator getModality() {
	return modality;
    }


}
