// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.rule.conditions;


import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.logic.Constraint;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.TermOrdering;
import de.uka.ilkd.key.logic.op.SVSubstitute;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.rule.SyntacticalReplaceVisitor;
import de.uka.ilkd.key.rule.TacletApp;
import de.uka.ilkd.key.rule.VariableConditionAdapter;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.util.Debug;


/**
* Condition for heuristic application: Compare the instantiations
* of two termSV or formulaSV, using a term ordering
*/
public abstract class AbstractOrderingCondition extends VariableConditionAdapter
                                                implements OrderingCondition {
    
    /**
     * The schema variables that are to be be compared
     */
    private final SchemaVariable var0, var1;

    protected AbstractOrderingCondition ( SchemaVariable p_var0,
                                          SchemaVariable p_var1 ) {
	var0 = p_var0;
	var1 = p_var1;

	if ( !( var0.isFormulaSV () && var1.isFormulaSV () ||
		var0.isTermSV ()    && var1.isTermSV () ) )
	    throw new IllegalArgumentException ( "Illegal schema variables" );
    }

    /**
     * @return true iff the instantiation of the taclet application satisfies
     *         some ordering condition (concerning variables <code>var0</code>,
     *         <code>var1</code>) regarding the given ordering
     */
    public boolean checkOrdering ( TacletApp    p_app,
                                   TermOrdering p_ordering ) {
	final SVInstantiations svi = p_app.instantiations ();
	final Constraint       c   = p_app.constraint     ();

	Term t0  = (Term)svi.getInstantiation ( var0 );
	Term t1  = (Term)svi.getInstantiation ( var1 );

	Debug.assertTrue ( t0 != null && t1 != null,
			   "Trying to compare uninstantiated schema variables" );

	if ( !c.isBottom () ) {
	    // Substitute metavariables of the two terms
	    SyntacticalReplaceVisitor srv = new SyntacticalReplaceVisitor ( c );
	    t0.execPostOrder ( srv );
	    t0 = srv.getTerm ();

	    srv = new SyntacticalReplaceVisitor ( c );
	    t1.execPostOrder ( srv );
	    t1 = srv.getTerm ();
	}

	return checkInstantiations ( t0, t1, p_ordering );
    }

    /**
     * Decide whether the instantiations of <code>var0</code>,
     * <code>var1</code> satisfy the condition represented by this class
     */
    protected abstract boolean checkInstantiations(Term t0, Term t1,
                                                   TermOrdering p_ordering);

    /**
     * checks if the condition for a correct instantiation is fulfilled
     * @param var the template Variable to be instantiated
     * @param term the Term to be mapped to var
     * @param svInst the SVInstantiations that are already known to be needed 
     * @return true iff condition is fulfilled
     */
    public boolean check(SchemaVariable var, SVSubstitute term, 
			 SVInstantiations svInst, Services services) {
	// This condition does not have any meaning regarding
	// soundness
	return true;
    }

    protected SchemaVariable getVar0() {
        return var0;
    }
    protected SchemaVariable getVar1() {
        return var1;
    }
}