// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.rule.conditions;


import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.reference.ExecutionContext;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.op.SVSubstitute;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.rule.VariableConditionAdapter;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.util.Debug;


/**
 * ensures that the types where the variables are declared are not the same
 */
public class TypeComparisionCondition extends VariableConditionAdapter {
    
    /** checks if sorts are not same */
    public final static int NOT_SAME = 0;
    /** checks if sorts are not compatible */
    public final static int NOT_COMPATIBLE = 1;
    /** checks subtype relationship */
    public final static int IS_SUBTYPE = 2;
    /** checks subtype relationship */
    public final static int NOT_IS_SUBTYPE = 3;

    private final SchemaVariable fstType;
    private final SchemaVariable sndType;
    private final int mode;


    /**
     * creates a condition that checks if the declaration types of the
     * schemavariable's instantiations are unequal 
     * @param fstType one of the SchemaVariable whose type is checked
     * @param sndType one of the SchemaVariable whose type is checked
     * @param mode an int encoding if testing of not same or not compatible
     */
    public TypeComparisionCondition (SchemaVariable fstType, 
				     SchemaVariable sndType,
				     int mode) {
	this.fstType = fstType;
	this.sndType = sndType;
	this.mode = mode;
    }

    /**
     * returns the sort of the given svsubstitute
     */
    private Sort getSort(SVSubstitute inst, ExecutionContext ec, 
            Services services) {
	final Sort sort;
	if (inst instanceof Expression) {
	    sort = ((Expression)inst).getKeYJavaType(services, ec).getSort();
	} else if (inst instanceof Term) {
	    sort = ((Term)inst).sort(); 
	} else {
	    sort = null;
	    Debug.fail("TypeComparisionCondition: " + 
		       "Unexpected instantiation of schemavariable.");
	}	
	return sort;
    }


    /**
     * tests if the instantiations for both schema variables have a
     * different type
     * @param var the template Variable to be instantiated
     * @param subst the SVSubstitute to be mapped to var
     * @param svInst the SVInstantiations that are already known to be
     * needed
     * @return true iff condition is fulfilled
     */
    public boolean check(SchemaVariable var, 
			 SVSubstitute subst, 
			 SVInstantiations svInst,
			 Services services) {

	SVSubstitute newInst;
	final SVSubstitute inst;
	// is instantiation of observed schemavariables complete
	if (var == fstType) {
	    inst = subst;
	    newInst = (SVSubstitute)svInst.getInstantiation(sndType);
	} else if (var == sndType) {
	    inst = (SVSubstitute)svInst.getInstantiation(fstType);	    
	    newInst = subst;
	} else {
	    inst = (SVSubstitute)svInst.getInstantiation(fstType);
	    newInst = (SVSubstitute)svInst.getInstantiation(sndType);
	}

	if (inst == null || newInst == null) {	    
	    // incomplete instantiation 
	    return true;
	}

	final ExecutionContext ec = svInst.getExecutionContext();
	final Sort fstSort = getSort(inst, ec, services);
	final Sort sndSort = getSort(newInst, ec, services);
	
	switch (mode) {
	case NOT_SAME:
	    return fstSort != sndSort;
	case NOT_COMPATIBLE:
	    return !(fstSort.extendsTrans(sndSort) ||
		     sndSort.extendsTrans(fstSort));
	case IS_SUBTYPE:        
	    return fstSort.extendsTrans(sndSort);
	case NOT_IS_SUBTYPE:	    
	    return !fstSort.extendsTrans(sndSort);
	default:
	    Debug.fail("TypeComparisionCondition: " + 
		       "Unknown mode.");
	    return false;
	}
    }

    public String toString () {
	switch (mode) {
	case NOT_COMPATIBLE:
	    return "\\notCompatibleTypes("+fstType+", "+sndType+")";
	case NOT_SAME:
	    return "\\notSameTypes("+fstType+", "+sndType+")";
	case IS_SUBTYPE:
	    return "\\isSubtype("+fstType+", "+sndType+")";
	case NOT_IS_SUBTYPE:
	    return "\\notIsSubtype("+fstType+", "+sndType+")";
	default:
	    return "invalid type copmparision mode";         	    
	}
    }

}
