// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.rule.metaconstruct.arith;

import java.math.BigInteger;

import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.abstraction.PrimitiveType;
import de.uka.ilkd.key.java.expression.literal.IntLiteral;
import de.uka.ilkd.key.logic.IteratorOfTerm;
import de.uka.ilkd.key.logic.ListOfTerm;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.SLListOfTerm;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.ldt.LDT;
import de.uka.ilkd.key.logic.op.AbstractMetaOperator;
import de.uka.ilkd.key.logic.op.Operator;
import de.uka.ilkd.key.logic.op.TermSymbol;

/**
 * Class for analysing and modifying monomial expressions over the integers
 */
public class Monomial {
    
    private final ListOfTerm parts;
    private final BigInteger coefficient;
    
    private Monomial(final ListOfTerm parts, final BigInteger coefficient) {
        this.parts = parts;
        this.coefficient = coefficient;
    }
    
    public static Monomial create(Term monomial, Services services) {
        final Analyser a = new Analyser ( services );
        a.analyse ( monomial );
        return new Monomial ( a.parts, a.coeff );
    }
    
    /**
     * @return true iff the monomial <code>this</code> divides the monomial
     *         <code>m</code>
     */
    public boolean divides (Monomial m) {
        if ( m.coefficient.signum () == 0 ) return true;
        if ( this.coefficient.signum () == 0 ) return false;
        if ( m.coefficient.remainder ( this.coefficient ).signum () != 0 )
                return false;
        
        return difference ( this.parts, m.parts ).isEmpty ();
    }
    
    /**
     * @precondition this.divides(m)
     * @return the result of dividing the monomial <code>m</code> by the
     *         monomial <code>this</code>
     */
    public Monomial divide(Monomial m) {
        if ( m.coefficient.signum () == 0 )
                return new Monomial ( SLListOfTerm.EMPTY_LIST, BigInteger.ZERO );
        return new Monomial ( difference ( m.parts, this.parts ),
                              m.coefficient.divide ( this.coefficient ) );
    }
    
    public Term toTerm (Services services) {
        final LDT intModel = services.getTypeConverter().
    		getModelFor(PrimitiveType.JAVA_INT);
        final Namespace intFunctions = intModel.functions();
        final TermSymbol mul = (TermSymbol)intFunctions.lookup(new Name("mul"));

        Term res = null;
        
        final IteratorOfTerm it = parts.iterator ();
        if ( it.hasNext () ) {
            res = it.next ();
            while ( it.hasNext () )
                res = TermFactory.DEFAULT.createFunctionTerm ( mul, res,
                                                               it.next () );
        }
        
	final IntLiteral lit = new IntLiteral ( coefficient.toString () );
        final Term cTerm = services.getTypeConverter ().convertToLogicElement ( lit );

        if ( res == null )
            res = cTerm;
        else	
            res = TermFactory.DEFAULT.createFunctionTerm ( mul, res, cTerm );
        
        return res;        
    }
    
    private static class Analyser {
        public BigInteger coeff = BigInteger.ONE;
        public ListOfTerm parts = SLListOfTerm.EMPTY_LIST;
        private final Services services;
        private final Operator numbers, mul;
        	
        public Analyser(final Services services) {
            this.services = services;

            final LDT intModel = services.getTypeConverter().
	    	getModelFor(PrimitiveType.JAVA_INT);
            final Namespace intFunctions = intModel.functions();
            
            numbers = (Operator)intFunctions.lookup(new Name("Z"));
            mul = (Operator)intFunctions.lookup(new Name("mul"));
        }
        
        public void analyse(Term monomial) {
            if ( monomial.op () == mul ) {
                analyse ( monomial.sub ( 0 ) );
                analyse ( monomial.sub ( 1 ) );
            } else if ( monomial.op () == numbers ) {
                final BigInteger c =
                    new BigInteger ( AbstractMetaOperator
                                     .convertToDecimalString ( monomial, services ) );
                coeff = coeff.multiply ( c );
            } else {
                parts = parts.prepend ( monomial );
            }
        }
    }
    
    /**
     * @return the list of all terms that occur in <code>a</code> but not in
     *         <code>b</code>. multiplicity is treated as well here, so this
     *         is really difference of multisets
     */
    private static ListOfTerm difference(ListOfTerm a, ListOfTerm b) {
        ListOfTerm res = a;
        final IteratorOfTerm it = b.iterator ();
        while ( it.hasNext () && !res.isEmpty () )
            res = res.removeFirst ( it.next () );
        return res;
    }
    
}
