// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.rule.updatesimplifier;

import java.util.ArrayList;

import de.uka.ilkd.key.logic.BoundVariableTools;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.op.ArrayOfQuantifiableVariable;
import de.uka.ilkd.key.logic.op.IUpdateOperator;
import de.uka.ilkd.key.logic.op.Location;
import de.uka.ilkd.key.logic.op.Operator;
import de.uka.ilkd.key.logic.op.QuanUpdateOperator;
import de.uka.ilkd.key.rule.AbstractUpdateRule;
import de.uka.ilkd.key.rule.UpdateSimplifier;
import de.uka.ilkd.key.util.Debug;


/**
 * Applies an update on an array using rule ...
 */
public class ApplyOnUpdate extends AbstractUpdateRule {
       
    /**
     * creates an instance of this rule used by the given update
     * simplifier
     */
    public ApplyOnUpdate(UpdateSimplifier us) {
	super(us);
    }

    /**
     * determines whether this rule is applicable for the 
     * update x target (the term on which the update will be
     * applied) pair 
     * @param update the Update to be applied on target
     * @param target the Term on which the update is applied
     * @return true if the top level operator of target is an
     * attribute operator
     */
    public boolean isApplicable(Update update, Term target) {
	return target.op() instanceof IUpdateOperator;
    }

    /**
     * determines whether this rule is applicable for the pair of update and
     * target (the term on which the update will be applied) term
     * 
     * @param update
     *            the Update to be applied on the given term
     * @param target
     *            the Term on which the update is applied
     * @return true if the rule can be applied on the update, target pair
     */
    public Term apply (Update update, Term target) {
        final Term targetRes = BoundVariableTools.DEFAULT
                            .resolveCollisions ( target, update.freeVars () );
        return applyHelp ( update, targetRes );
    }
    
    /**
     * Here it is assumed that collisions have already been resolved, i.e. no
     * variables that are bound by the top-level operator of <code>target</code>
     * turn up free in <code>update</code>
     */
    public Term applyHelp(Update update, Term target) {
        final IUpdateOperator targetOp = (IUpdateOperator) target.op();
        final ArrayList asgnPairs = new ArrayList();

        final ArrayOfAssignmentPair allUpdatePairs = 
	    update.getAllAssignmentPairs();
        
        for (int i = 0; i < allUpdatePairs.size(); i++)
            asgnPairs.add(allUpdatePairs.getAssignmentPair(i));

        for (int accessOpPos = 0, targetLocCount = targetOp.locationCount(); 
             accessOpPos < targetLocCount;
             accessOpPos++) {
            final Location accessOp = targetOp.location(accessOpPos);
            // apply update on location
            final Term[] locSubs =
                applyOnLocation ( update,
                                  target,
                                  accessOp,
                                  targetOp.locationSubtermsBegin ( accessOpPos ),
                                  targetOp.locationSubtermsEnd ( accessOpPos ) );
            // apply update on value
            final Term value =
                updateSimplifier ()
                .simplify ( update, targetOp.value ( target, accessOpPos ) );
            
            if ( targetOp instanceof QuanUpdateOperator ) {
                final QuanUpdateOperator quanTOp = (QuanUpdateOperator)targetOp;
                // apply update on guard
                final Term guard = applyUpdateOnGuard ( update,
                                                        target,
                                                        accessOpPos,
                                                        quanTOp );
                // nothing needs to be done for quantified variables
                final ArrayOfQuantifiableVariable boundVars =
                    quanTOp.boundVars ( target, accessOpPos );
                asgnPairs.add ( new AssignmentPairImpl
                                ( boundVars, guard, accessOp, locSubs, value ) );
            } else {
                asgnPairs.add ( new AssignmentPairImpl ( accessOp, locSubs, value ) );
            }
        }

//        final AssignmentPair[] result = optimize ( asgnPairs );
        final AssignmentPair[] result =
            (AssignmentPair[])asgnPairs.toArray ( new AssignmentPair [asgnPairs.size ()] );

        return updateSimplifier ().simplify ( Update.createUpdate ( result ),
                                              targetOp.target(target) );
    }

    /**
     * TODO comment
     * @param update
     * @param target
     * @param accessOp
     * @param valuePos
     * @return
     */
    private Term[] applyOnLocation (Update update,
                                    Term target,
                                    final Operator accessOp,
                                    int beginLocationSubterms,
                                    int endLocationSubterms) {
        final Term[] locSubs = new Term[accessOp.arity()];
        // applies update on location
        for ( int i = beginLocationSubterms, j = 0;
              i != endLocationSubterms;
              ++i, ++j )
            locSubs[j] = updateSimplifier ().simplify ( update, target.sub ( i ) );
        return locSubs;
    }
    
    private Term applyUpdateOnGuard (Update update,
                                     Term target,
                                     int accessOpPos,
                                     final QuanUpdateOperator quanTOp) {
        if ( !quanTOp.guardExists ( accessOpPos ) )
            return UpdateSimplifierTermFactory.DEFAULT.getValidGuard ();

        final Term oldGuard = quanTOp.guard ( target, accessOpPos );
        return updateSimplifier ().simplify ( update, oldGuard );
    }
    
    public Term matchingCondition (Update update, Term target) {
        // an update is not a location
        Debug.fail ( "matchingCondition(...) must not be called for target "
                     + target );
        return null; // unreachable
    }
}
