// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.strategy;

import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.PosInOccurrence;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.rule.RuleApp;
import de.uka.ilkd.key.strategy.feature.*;


/**
 *
 */
public class SimpleFOLStrategy extends AbstractFeatureStrategy {
    
    public static final Name NAME = new Name ( "Simple FOL" );
    
    private final Feature completeF;
    
    private final Feature approvalF;

    protected SimpleFOLStrategy ( Proof p_proof ) {
        super ( p_proof );
        
        bindRuleSet ( "closure",  -9000 );
        bindRuleSet ( "concrete", -10000 );
        bindRuleSet ( "alpha",    -7000 );
        bindRuleSet ( "delta",    -6000 );
        bindRuleSet ( "pull_out_quantifier", 5000 );

        bindRuleSet ( "beta",
          add ( longConst ( -2500 ),
                ifZero ( SimplifyBetaCandidateFeature.INSTANCE,
                  SumFeature.createSum ( new Feature[] {
                    longConst ( -1070 ),
                    ifZero ( PurePosDPathFeature.INSTANCE,
                             longConst ( -200 ) ),
                    //	   ifZero ( ContainsQuantifierFeature.INSTANCE,
                    //		    longConst ( -2000 ) ),
                    ScaleFeature.createScaled ( CountPosDPathFeature.INSTANCE, -3.0 ),
                    ScaleFeature.createScaled ( CountMaxDPathFeature.INSTANCE, 10.0 )
                  }) ) ) );

        final Feature introducedByGammaF = FormulaAddedByRuleFeature.create (
               getFilterFor ( new String[] { "gamma", "gamma_destructive" } ) );

        bindRuleSet ( "gamma",
                      ifZero ( introducedByGammaF,
                               longConst ( 0 ),
                               add ( ifZero ( NonDuplicateAppFeature.INSTANCE,
                                              longConst ( -200 ) ),
                                     longConst ( -3250 ) ) ) );

        bindRuleSet ( "gamma_destructive",
                      ifZero ( introducedByGammaF, longConst ( -5000 ) ) );

        bindRuleSet ( "replace_known",
           SumFeature.createSum ( new Feature[] {
               SimplifyReplaceKnownCandidateFeature.INSTANCE,
               ifZero ( ConstraintStrengthenFeature.INSTANCE,
                        SumFeature.createSum ( new Feature[] {
			    ifZero ( SimplifyBetaCandidateFeature.INSTANCE,
				     inftyConst () ),
			    NotBelowQuantifierFeature.INSTANCE,
			    LeftmostNegAtomFeature.INSTANCE } ),
                        longConst ( -800 ) ),
               longConst ( -4000 ),
               ScaleFeature.createScaled ( CountMaxDPathFeature.INSTANCE,
                                           10.0 ) } ) );
        
        bindRuleSet ( "apply_equations",
         SumFeature.createSum ( new Feature [] {
                 ifZero ( MatchedIfFeature.INSTANCE,
                          add ( CheckApplyEqFeature.INSTANCE,
                                OrderingFeature.INSTANCE ) ),
                 ifZero ( ConstraintStrengthenFeature.INSTANCE,
                          SumFeature.createSum ( new Feature [] {
                                  ifZero ( SimplifyBetaCandidateFeature.INSTANCE,
                                           inftyConst () ),
                                  NotBelowQuantifierFeature.INSTANCE,
                                  ifZero ( ContainsQuantifierFeature.INSTANCE,
                                           inftyConst () ) } ) ),
                 longConst ( -4000 ) } ) );
        
        bindRuleSet ( "order_terms",
                      add ( OrderingFeature.INSTANCE, longConst ( -8000 ) ) );
        
        bindRuleSet ( "simplify_literals",
                      ifZero ( ConstraintStrengthenFeature.INSTANCE,
                               longConst ( -4000 ),
                               longConst ( -8000 ) ) );
        
        bindRuleSet ( "try_apply_subst",
                      add ( EqNonDuplicateAppFeature.INSTANCE,
                            longConst ( -10000 ) ) );

        final Feature simplifierF = selectSimplifier ( -10000 );

        final Feature duplicateF = ifHeuristics ( new String[] { "gamma", "gamma_destructive" },
                                    longConst ( 0 ),
                                    NonDuplicateAppFeature.INSTANCE );

        final Feature ifMatchedF = ifZero ( MatchedIfFeature.INSTANCE, longConst ( +1 ) );
        
        completeF = SumFeature.createSum ( new Feature[] { AgeFeature.INSTANCE,
            AutomatedRuleFeature.INSTANCE, simplifierF, duplicateF,
            ifMatchedF, getDispatcher() } );
            
        approvalF = duplicateF;
    }
    
    public Name name () {
        return NAME;
    }


    /**
     * Evaluate the cost of a <code>RuleApp</code>.
     * 
     * @return the cost of the rule application expressed as a
     *         <code>RuleAppCost</code> object.
     *         <code>TopRuleAppCost.INSTANCE</code> indicates that the rule
     *         shall not be applied at all (it is discarded by the strategy).
     */
    public RuleAppCost computeCost (RuleApp app, PosInOccurrence pio, Goal goal) {
        return completeF.compute ( app, pio, goal );
    }

    /**
     * Re-Evaluate a <code>RuleApp</code>. This method is
     * called immediately before a rule is really applied
     * @return true iff the rule should be applied, false otherwise
     */
    public boolean isApprovedApp (RuleApp app, PosInOccurrence pio, Goal goal) {
        return !( approvalF.compute ( app, pio, goal ) instanceof TopRuleAppCost );
    }
    
    public static class Factory extends StrategyFactory {
        public Strategy create ( Proof p_proof ) {
            return new SimpleFOLStrategy ( p_proof );
        }
        
        public Name name () {
            return NAME;
        }
    }
}
