// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2005 Universitaet Karlsruhe, Germany
//                         Universitaet Koblenz-Landau, Germany
//                         Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
//

package de.uka.ilkd.key.strategy;

import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.PosInOccurrence;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.rule.RuleApp;
import de.uka.ilkd.key.strategy.feature.AgeFeature;
import de.uka.ilkd.key.strategy.feature.AutomatedRuleFeature;
import de.uka.ilkd.key.strategy.feature.CheckApplyEqFeature;
import de.uka.ilkd.key.strategy.feature.ConstraintStrengthenFeature;
import de.uka.ilkd.key.strategy.feature.ConstraintStrengthenFeatureUC;
import de.uka.ilkd.key.strategy.feature.CountMaxDPathFeature;
import de.uka.ilkd.key.strategy.feature.CountPosDPathFeature;
import de.uka.ilkd.key.strategy.feature.DiffFindAndIfFeature;
import de.uka.ilkd.key.strategy.feature.EqNonDuplicateAppFeature;
import de.uka.ilkd.key.strategy.feature.ExcludeTopSymbolFeature;
import de.uka.ilkd.key.strategy.feature.Feature;
import de.uka.ilkd.key.strategy.feature.FindDepthFeature;
import de.uka.ilkd.key.strategy.feature.MatchedIfFeature;
import de.uka.ilkd.key.strategy.feature.NonDuplicateAppFeature;
import de.uka.ilkd.key.strategy.feature.NotDirectlyBelowSymbolFeature;
import de.uka.ilkd.key.strategy.feature.OrderingFeature;
import de.uka.ilkd.key.strategy.feature.PurePosDPathFeature;
import de.uka.ilkd.key.strategy.feature.ScaleFeature;
import de.uka.ilkd.key.strategy.feature.SumFeature;
import de.uka.ilkd.key.strategy.feature.UCIncompatibleFeature;


/**
 *
 */
public class SimpleJavaCardDLStrategy extends AbstractFeatureStrategy {

    private final SimpleJavaCardDLOptions options;
    
    protected Feature completeF;
    
    private final Feature approvalF;

    protected SimpleJavaCardDLStrategy ( Proof p_proof,
                                         SimpleJavaCardDLOptions options) {
        super ( p_proof );
        
        this.options = options; 
        
        setupDispatcher ( options );
        
        final Feature simplifierF = selectSimplifier ( -10000 );

        final Feature ifMatchedF = ifZero ( MatchedIfFeature.INSTANCE,
                                            longConst ( +1 ) );

        final Feature strengthenConstraints =
            ifHeuristics ( new String[] { "concrete", "closure" },
                           longConst ( 1000 ),
                           ifZero ( ConstraintStrengthenFeatureUC.create ( p_proof ),
                                    inftyConst () ) );
        
        completeF = SumFeature.createSum ( new Feature [] {
	    AutomatedRuleFeature.INSTANCE, strengthenConstraints,
	    NonDuplicateAppFeature.INSTANCE, AgeFeature.INSTANCE,
            simplifierF, ifMatchedF, getDispatcher () } );
            
        approvalF = add ( NonDuplicateAppFeature.INSTANCE,
                          UCIncompatibleFeature.create ( p_proof ) );

    }
    
    private void setupDispatcher(SimpleJavaCardDLOptions options) {
        bindRuleSet ( "closure", -9000 );
        bindRuleSet ( "concrete", -10000 );
        bindRuleSet ( "alpha", -7000 );
        bindRuleSet ( "delta", -6000 );
        
        bindRuleSet ( "pull_out_quantifier", inftyConst () );
        bindRuleSet ( "simplify_boolean", -200 );
        bindRuleSet ( "simplify", -200 );
        bindRuleSet ( "simplify_int", -100 );
        
        bindRuleSet ( "beta",
           SumFeature.createSum ( new Feature [] {
             longConst ( -2500 ),
             ifZero ( PurePosDPathFeature.INSTANCE,
                      longConst ( -200 ) ),
             ScaleFeature.createScaled ( CountPosDPathFeature.INSTANCE,
                                          -3.0 ),
             ScaleFeature.createScaled ( CountMaxDPathFeature.INSTANCE,
                                          10.0 ),
           } ) );

        // currently gamma rules are not applied automatically at all by this
        // strategy
        bindRuleSet ( "gamma", inftyConst () );
        bindRuleSet ( "gamma_destructive", inftyConst () );
        	
        bindRuleSet ( "replace_known",
         add ( ifZero ( MatchedIfFeature.INSTANCE,
                        DiffFindAndIfFeature.INSTANCE ),
               longConst ( -4000 ),
               ScaleFeature.createScaled ( CountMaxDPathFeature.INSTANCE, 10.0 ) ) );

        bindRuleSet ( "confluence_restricted",
                      ifZero ( MatchedIfFeature.INSTANCE,
                               DiffFindAndIfFeature.INSTANCE ) );

        bindRuleSet ( "apply_equations",
           add ( ifZero ( MatchedIfFeature.INSTANCE,
                          add ( CheckApplyEqFeature.INSTANCE,
                                OrderingFeature.INSTANCE ) ),
                 longConst ( -4000 ) ) );
        
        bindRuleSet ( "order_terms",
                      add ( OrderingFeature.INSTANCE,
                            longConst ( -8000 ) ) );

        bindRuleSet ( "simplify_literals",
                      ifZero ( ConstraintStrengthenFeature.INSTANCE,
                               longConst ( 0 ),
                               longConst ( -8000 ) ) );

        bindRuleSet ( "simplify_instanceof_static",
                      add ( EqNonDuplicateAppFeature.INSTANCE,
                            longConst ( -500 ) ) );

        bindRuleSet ( "beta_lemma", -10 );

        bindRuleSet ( "try_apply_subst",
                      add ( EqNonDuplicateAppFeature.INSTANCE,
                            longConst ( -10000 ) ) );

        bindRuleSet ( "split_cond", -50 );

        // features influenced by the strategy options

        bindRuleSet ( "use_method_specs",
                      options.useMethodContracts () ? longConst ( -20 )
                                                    : inftyConst () );

        bindRuleSet ( "loop_expand",
                      options.unwindLoops () ? longConst ( 0 )
                                             : inftyConst () );

        bindRuleSet ( "method_expand",
                      options.expandMethods () ? longConst ( 100 )
                                               : inftyConst () );

        bindRuleSet ( "polySimp", -200 );
        setupPolySimp ();
    }

    
    private void setupPolySimp() {
        bindRuleSet ( "polySimp", FindDepthFeature.INSTANCE );
        
        bindRuleSet ( "polySimp_dist", -80 );

        bindRuleSet ( "polySimp_elimSubNeg",
                      add ( ExcludeTopSymbolFeature.create ( "intNotZ", "Z" ),
                            longConst ( -120 ) ) );

        bindRuleSet ( "polySimp_assoc", -120 );

        bindRuleSet ( "polySimp_addOrder",
                      add ( OrderingFeature.INSTANCE,
                            ExcludeTopSymbolFeature.create ( "intNotAdd", "add" ),
                            longConst ( -20 ) ) );

        bindRuleSet ( "polySimp_normalise",
                      add ( ExcludeTopSymbolFeature.create ( "numNotNegLit", "neglit" ),
                            longConst ( -15 ) ) );

        bindRuleSet ( "polySimp_mulOrder",
                      add ( OrderingFeature.INSTANCE,
                            ExcludeTopSymbolFeature.create ( "intNotMul", "mul" ),
                            longConst ( -100 ) ) );

        bindRuleSet ( "polySimp_mulOne",
           SumFeature.createSum( new Feature[] {
             ExcludeTopSymbolFeature.create ( "intNotZ", "Z" ),
             ExcludeTopSymbolFeature.create ( "intNotAddMulZ", "add" ),
             ExcludeTopSymbolFeature.create ( "intNotAddMulZ", "mul" ),
             ExcludeTopSymbolFeature.create ( "intNotAddMulZ", "Z" ),
             longConst ( -35 ) } ) );

        bindRuleSet ( "polySimp_balanceEq", 40 );

        bindRuleSet ( "polySimp_elimOne",
                      add ( NotDirectlyBelowSymbolFeature.create ( "add" ),
                            NotDirectlyBelowSymbolFeature.create ( "equals" ) ) );

        bindRuleSet ( "polySimp_applyEq",
                      add ( ExcludeTopSymbolFeature.create ( "intNotZ", "Z" ),
                            longConst ( 100 ) ) );
    }
    
    public Name name () {
        return options.name ();
    }


    /**
	 * Evaluate the cost of a <code>RuleApp</code>.
	 * 
	 * @return the cost of the rule application expressed as a
	 *         <code>RuleAppCost</code> object.
	 *         <code>TopRuleAppCost.INSTANCE</code> indicates that the rule
	 *         shall not be applied at all (it is discarded by the strategy).
	 */
    public RuleAppCost computeCost (RuleApp app, PosInOccurrence pio, Goal goal) {
        return completeF.compute ( app, pio, goal );
    }

    /**
	 * Re-Evaluate a <code>RuleApp</code>. This method is called immediately
	 * before a rule is really applied
	 * 
	 * @return true iff the rule should be applied, false otherwise
	 */
    public boolean isApprovedApp (RuleApp app, PosInOccurrence pio, Goal goal) {
        return !( approvalF.compute ( app, pio, goal ) instanceof TopRuleAppCost );
    }
    
    

    public static class Factory extends StrategyFactory {
        private final SimpleJavaCardDLOptions options;

        public Factory (SimpleJavaCardDLOptions options) {
            this.options = options;
        }

        public Strategy create ( Proof p_proof ) {
            return new SimpleJavaCardDLStrategy ( p_proof, options );
        }
        
        public Name name () {
            return options.name ();
        }
    }
}
